
#if !defined(_NativeInterpreterMacros__)
#define _NativeInterpreterMacros__

#include "NativeInterpreterTypes.h"
#include "NativeInterpreterVm.h"

#include "core/nerd_memory.h"

#if defined(CEM_ARCH_IS_X86_64)
#define NANOJIT_TERP_ALLOW_UNALIGNED_ACCESS
#endif

namespace nanojit
{
namespace interpreter
{


template<typename T>
inline T readValue(const uint8_t*& stream)
{
	T value;
#if defined(NANOJIT_TERP_ALLOW_UNALIGNED_ACCESS)
	value = *reinterpret_cast<const T*>(stream);
#else
	nerd::MemCpy(&value, stream, sizeof(T));
#endif

	stream += sizeof(T);
	return value;
}


template<typename T>
inline void writeValue(uint8_t*& stream, T value)
{
	stream -= sizeof(T);
#if defined(NANOJIT_TERP_ALLOW_UNALIGNED_ACCESS)
	*reinterpret_cast<T*>(stream) = value;
#else
	nerd::MemCpy(stream, &value, sizeof(T));
#endif
}

// specialization for float / double types (assume unaligned access is not allowed)

template<>
inline float readValue<float>(const uint8_t*& stream)
{
	float value;
	nerd::MemCpy(&value, stream, sizeof(float));
	stream += sizeof(float);
	return value;
}

template<>
inline void writeValue<float>(uint8_t*& stream, float value)
{
	stream -= sizeof(float);
	nerd::MemCpy(stream, &value, sizeof(float));
}


template<>
inline double readValue<double>(const uint8_t*& stream)
{
	float value;
	nerd::MemCpy(&value, stream, sizeof(double));
	stream += sizeof(double);
	return value;
}

template<>
inline void writeValue<double>(uint8_t*& stream, double value)
{
	stream -= sizeof(double);
	nerd::MemCpy(stream, &value, sizeof(double));
}


} // namespace interpreter
} // namespace nanojit


// read macros
#define NANOJIT_TERP_READ_OPCODE(_v, stream)				\
	do { _v = ::nanojit::interpreter::readValue<uint16_t>(stream); } while (false)
#define NANOJIT_TERP_READ_REGISTER_INDEX(_v, stream)		\
	do { _v = ::nanojit::interpreter::readValue<uint8_t>(stream); } while (false)
#define NANOJIT_TERP_READ_IMMI8(_v, stream)					\
	do { _v = ::nanojit::interpreter::readValue<uint8_t>(stream); } while (false)
#define NANOJIT_TERP_READ_IMMI16(_v, stream)				\
	do { _v = ::nanojit::interpreter::readValue<uint16_t>(stream); } while (false)
#define NANOJIT_TERP_READ_IMMI32(_v, stream)				\
	do { _v = ::nanojit::interpreter::readValue<uint32_t>(stream); } while (false)
#define NANOJIT_TERP_READ_IMMI64(_v, stream)				\
	do { _v = ::nanojit::interpreter::readValue<uint64_t>(stream); } while (false)
#define NANOJIT_TERP_READ_IMMP(_v, stream)					\
	do { _v = ::nanojit::interpreter::readValue<uintptr_t>(stream); } while (false)
#define NANOJIT_TERP_READ_IMMF32(_v, stream)				\
	do { _v = ::nanojit::interpreter::readValue<float>(stream); } while (false)
#define NANOJIT_TERP_READ_IMMF64(_v, stream)				\
	do { _v = ::nanojit::interpreter::readValue<double>(stream); } while (false)
#define NANOJIT_TERP_READ_JUMP_OFFSET(_v, stream)			\
	do { _v = ::nanojit::interpreter::readValue<ptrdiff_t>(stream); } while (false)

// write macros
#define NANOJIT_TERP_WRITE_OPCODE(_v, stream, msg, ...)												\
	do {																							\
		::nanojit::interpreter::writeValue<uint16_t>(stream, nerd::NumCast<uint16_t>(_v));									\
		verbose_only(																				\
			char buffer[256];																		\
			VMPI_snprintf(buffer, 256, msg, __VA_ARGS__);											\
			asm_output("%s %s", nanojit::interpreter::getInterpreterOpcodeString(_v), buffer);		\
		)																							\
	} while (false)
#define NANOJIT_TERP_WRITE_REGISTER_INDEX(_v, stream)		\
	do { ::nanojit::interpreter::writeValue<uint8_t>(stream, nerd::NumCast<uint8_t>(::nanojit::REGNUM(_v))); } while (false)
#define NANOJIT_TERP_WRITE_IMMI8(_v, stream)				\
	do { ::nanojit::interpreter::writeValue<uint8_t>(stream, _v); } while (false)
#define NANOJIT_TERP_WRITE_IMMI16(_v, stream)				\
	do { ::nanojit::interpreter::writeValue<uint16_t>(stream, _v); } while (false)
#define NANOJIT_TERP_WRITE_IMMI32(_v, stream)				\
	do { ::nanojit::interpreter::writeValue<uint32_t>(stream, _v); } while (false)
#define NANOJIT_TERP_WRITE_IMMI64(_v, stream)				\
	do { ::nanojit::interpreter::writeValue<uint64_t>(stream, _v); } while (false)
#define NANOJIT_TERP_WRITE_IMMP(_v, stream)					\
	do { ::nanojit::interpreter::writeValue<uintptr_t>(stream, reinterpret_cast<uintptr_t>(_v)); } while (false)
#define NANOJIT_TERP_WRITE_IMMF32(_v, stream)				\
	do { ::nanojit::interpreter::writeValue<float>(stream, _v); } while (false)
#define NANOJIT_TERP_WRITE_IMMF64(_v, stream)				\
	do { ::nanojit::interpreter::writeValue<double>(stream, _v); } while (false)
#define NANOJIT_TERP_WRITE_JUMP_OFFSET(_v, stream)			\
	do { ::nanojit::interpreter::writeValue<ptrdiff_t>(stream, _v); } while (false)

#endif // !defined(_NativeInterpreterMacros__)
