
#if !defined(_NativeInterpreterTypes__)
#define _NativeInterpreterTypes__

#include "nanojit.h"
#include "NativeInterpreterFloat.h"

namespace nanojit
{

// additional opcodes for interpreter
enum InterpreterAdditionnalOpcodes
{
	LIR_genflgsint		= LIR_sentinel + 1,
	LIR_genflgsflt,
	LIR_branch,
	LIR_branch_cond,

	LIR_spilli,
#if defined(NANOJIT_64BIT)
	LIR_spillq,
#endif
	LIR_spillf,
	LIR_spilld,

	LIR_restorei,
#if defined(NANOJIT_64BIT)
	LIR_restoreq,
#endif
	LIR_restoref,
	LIR_restored,

	LIR_pushi,
#if defined(NANOJIT_64BIT)
	LIR_pushq,
#endif
	LIR_pushf,
	LIR_pushd,

	LIR_popi,
#if defined(NANOJIT_64BIT)
	LIR_popq,
#endif
	LIR_popf,
	LIR_popd,

	LIR_copyreg,

	LIR_addimm32i,
#if defined(NANOJIT_64BIT)
	LIR_addimm32q,
#endif

	LIR_clearfpflags,
};



enum MemoryByteOrder 
{
	MemoryByteOrder_Host,
	MemoryByteOrder_Reversed
};



union RegisterValue 
{
	uint32_t		asI;
	uint64_t		asQ;
	float32_t		asF;
	float64_t		asD;

	template<typename T> 
	inline T get() const;

	template<typename T> 
	void set(T value);
};


template<>
inline uint32_t RegisterValue::get<uint32_t>() const {
	return asI;
}

template<>
inline int32_t RegisterValue::get<int32_t>() const {
	return static_cast<int32_t>(asI);
}

template<>
inline uint64_t RegisterValue::get<uint64_t>() const {
	return asQ;
}

template<>
inline int64_t RegisterValue::get<int64_t>() const {
	return static_cast<int64_t>(asQ);
}

template<>
inline float32_t RegisterValue::get<float32_t>() const {
	return asF;
}

template<>
inline float64_t RegisterValue::get<float64_t>() const {
	return asD;
}

template<>
inline void RegisterValue::set<uint32_t>(uint32_t value) {
	asI = value;
}

template<>
inline void RegisterValue::set<int32_t>(int32_t value) {
	asI = value;
}

template<>
inline void RegisterValue::set<uint64_t>(uint64_t value) {
	asQ = value;
}

template<>
inline void RegisterValue::set<int64_t>(int64_t value) {
	asQ = value;
}

template<>
inline void RegisterValue::set<float32_t>(float32_t value) {
	asF = value;
}

template<>
inline void RegisterValue::set<float64_t>(float64_t value) {
	asD = value;
}

template<typename T, size_t DEPTH_MAX>
struct Stack 
{
	size_t				depth;
	T					stack[DEPTH_MAX];

	inline void set(unsigned int i, T v) {
		NanoAssert(i < DEPTH_MAX);
		stack[i] = v;
	}

	inline T get(unsigned int i) const {
		NanoAssert(i < DEPTH_MAX);
		return stack[i];
	}

	inline void push(T v) {
		NanoAssert(depth < DEPTH_MAX);
		stack[depth] = v;
		depth++;
	}

	inline T pop() {
		NanoAssert(depth > 0);
		T res = stack[depth - 1];
		 depth--;
		 return res;
	}
};

struct IntepreterContext 
{
	RegisterValue		registerValue[nanojit::LastRegNum + 1];
	unsigned int		flagMask;

	static const size_t STACK_DEPTH = 16;

	Stack<uint32_t, STACK_DEPTH>	stackI;
	Stack<uint64_t, STACK_DEPTH>	stackQ;
	Stack<float32_t, STACK_DEPTH>	stackF;
	Stack<float64_t, STACK_DEPTH>	stackD;

	template<typename T> 
	inline T get(unsigned int reg) const {
		NanoAssert(reg <= nanojit::LastRegNum);
		return registerValue[reg].get<T>();
	}

	template<typename T> 
	inline void set(unsigned int reg, T value) {
		NanoAssert(reg <= nanojit::LastRegNum);
		return registerValue[reg].set<T>(value);
	}

	inline void clear(unsigned int reg) {
		memset(&registerValue[reg], 0, sizeof(RegisterValue));
	}

	inline void copy(unsigned int regDst, unsigned int regSrc) {
		NanoAssert(regDst <= nanojit::LastRegNum);
		NanoAssert(regSrc <= nanojit::LastRegNum);
		registerValue[regDst] = registerValue[regSrc];
	}


	template<typename T> 
	inline void setStack(unsigned int stackIndex, T v);

	template<typename T> 
	inline T getStack(unsigned int stackIndex) const;

	inline bool getFlag(ArithFlag flag) const { return (flagMask & flag) != 0; }
	inline void setFlag(ArithFlag flag) { flagMask |= flag; }
	inline void clearFlag(ArithFlag flag) { flagMask &= ~flag; }
};

template<> 
inline void IntepreterContext::setStack<uint32_t>(unsigned int stackIndex, uint32_t v) {
	stackI.set(stackIndex, v);
}

template<> 
inline void IntepreterContext::setStack<uint64_t>(unsigned int stackIndex, uint64_t v) {
	stackQ.set(stackIndex, v);
}

template<> 
inline void IntepreterContext::setStack<float32_t>(unsigned int stackIndex, float32_t v) {
	stackF.set(stackIndex, v);
}

template<> 
inline void IntepreterContext::setStack<float64_t>(unsigned int stackIndex, float64_t v) {
	stackD.set(stackIndex, v);
}



template<> 
inline uint32_t IntepreterContext::getStack<uint32_t>(unsigned int stackIndex) const {
	return stackI.get(stackIndex);
}

template<> 
inline uint64_t IntepreterContext::getStack<uint64_t>(unsigned int stackIndex) const {
	return stackQ.get(stackIndex);
}

template<> 
inline float32_t IntepreterContext::getStack<float32_t>(unsigned int stackIndex) const {
	return stackF.get(stackIndex);
}

template<> 
inline float64_t IntepreterContext::getStack<float64_t>(unsigned int stackIndex) const {
	return stackD.get(stackIndex);
}


typedef uint8_t* (*CallbackFunc)(IntepreterContext* a_pContext, uint8_t* a_pStream);


} // namespace nanojit



#endif // !defined(_NativeInterpreterTypes__)
