
#include "nanojit.h"
#include <limits.h>
#include "MemoryAccess.h"
#include "NativeInterpreterVmCallbacks.h"

using namespace nanojit;
using namespace nanojit::interpreter;

namespace
{

NANOJIT_FORCEINLINE const uint8_t* dispatch(IntepreterContext* context, const uint8_t* stream)
{
	unsigned int		op			= LIR_sentinel;

	NANOJIT_TERP_READ_OPCODE(op, stream);

	switch (op)
	{
	case LIR_immi:			stream = callBackImm<uint32_t>(context, stream); break;
#if defined(NANOJIT_64BIT)
	case LIR_immq:			stream = callBackImm<uint64_t>(context, stream); break;
#endif
	case LIR_immf:			stream = callBackImm<float32_t>(context, stream); break;
	case LIR_immd:			stream = callBackImm<float64_t>(context, stream); break;

	case LIR_addi:			stream = callbackAdd<uint32_t>(context, stream); break;
#if defined(NANOJIT_64BIT)
	case LIR_addq:			stream = callbackAdd<uint64_t>(context, stream); break;
#endif
	case LIR_addf:			stream = callbackAdd<float32_t>(context, stream); break;
	case LIR_addd:			stream = callbackAdd<float64_t>(context, stream); break;

	case LIR_subi:			stream = callbackSub<uint32_t>(context, stream); break;
#if defined(NANOJIT_64BIT)
	case LIR_subq:			stream = callbackSub<uint64_t>(context, stream); break;
#endif
	case LIR_subf:			stream = callbackSub<float32_t>(context, stream); break;
	case LIR_subd:			stream = callbackSub<float64_t>(context, stream); break;

	case LIR_negi:			stream = callbackNeg<int32_t>(context, stream); break;
#if defined(NANOJIT_64BIT)
	case LIR_negq:			stream = callbackNeg<int64_t>(context, stream); break;
#endif
	case LIR_negf:			stream = callbackNeg<float32_t>(context, stream); break;
	case LIR_negd:			stream = callbackNeg<float64_t>(context, stream); break;

	case LIR_muli:			stream = callbackMul<uint32_t>(context, stream); break;
#if defined(NANOJIT_64BIT)
	case LIR_mulq:			stream = callbackMul<uint64_t>(context, stream); break;
#endif
	case LIR_mulf:			stream = callbackMul<float32_t>(context, stream); break;
	case LIR_muld:			stream = callbackMul<float64_t>(context, stream); break;

	case LIR_mullsi:		stream = callbackMulLong<int32_t, int64_t>(context, stream); break;
	case LIR_mullui:		stream = callbackMulLong<uint32_t, uint64_t>(context, stream); break;
#if defined(NANOJIT_64BIT)
	case LIR_mullsq:		NanoAssert(0); break; // to be implemented - 128 mul
	case LIR_mulluq:		NanoAssert(0); break; // to be implemented - 128 mul
#endif

	case LIR_addci:			stream = callbackAddc<uint32_t>(context, stream); break;
#if defined(NANOJIT_64BIT)
	case LIR_addcq:			stream = callbackAddc<uint64_t>(context, stream); break;
#endif


	case LIR_subbi:			stream = callbackSubb<uint32_t>(context, stream); break;
#if defined(NANOJIT_64BIT)
	case LIR_subbq:			stream = callbackSubb<uint64_t>(context, stream); break;
#endif

	case LIR_noti:			stream = callbackNot<uint32_t>(context, stream); break;
#if defined(NANOJIT_64BIT)
	case LIR_notq:			stream = callbackNot<uint64_t>(context, stream); break;
#endif


	case LIR_ori:			stream = callbackOr<uint32_t>(context, stream); break;
#if defined(NANOJIT_64BIT)
	case LIR_orq:			stream = callbackOr<uint64_t>(context, stream); break;
#endif

	case LIR_xori:			stream = callbackXor<uint32_t>(context, stream); break;
#if defined(NANOJIT_64BIT)
	case LIR_xorq:			stream = callbackXor<uint64_t>(context, stream); break;
#endif

	case LIR_andi:			stream = callbackAnd<uint32_t>(context, stream); break;
#if defined(NANOJIT_64BIT)
	case LIR_andq:			stream = callbackAnd<uint64_t>(context, stream); break;
#endif

	case LIR_andci:			stream = callbackAndc<uint32_t>(context, stream); break;
#if defined(NANOJIT_64BIT)
	case LIR_andcq:			stream = callbackAndc<uint64_t>(context, stream); break;
#endif

	case LIR_lshi:			stream = callbackLsh<uint32_t>(context, stream); break;
#if defined(NANOJIT_64BIT)
	case LIR_lshq:			stream = callbackLsh<uint64_t>(context, stream); break;
#endif

	case LIR_rshi:			stream = callbackRsh<int32_t>(context, stream); break;
#if defined(NANOJIT_64BIT)
	case LIR_rshq:			stream = callbackRsh<int64_t>(context, stream); break;
#endif

	case LIR_rshui:			stream = callbackRsh<uint32_t>(context, stream); break;
#if defined(NANOJIT_64BIT)
	case LIR_rshuq:			stream = callbackRsh<uint64_t>(context, stream); break;
#endif

	case LIR_rori:			stream = callbackRor<uint32_t>(context, stream); break;
#if defined(NANOJIT_64BIT)
	case LIR_rorq:			stream = callbackRor<uint64_t>(context, stream); break;
#endif

	case LIR_rcri:			stream = callbackRcr<uint32_t>(context, stream); break;
#if defined(NANOJIT_64BIT)
	case LIR_rcrq:			stream = callbackRcr<uint64_t>(context, stream); break;
#endif

	case LIR_absf:			stream = callbackAbs<float32_t>(context, stream); break;
	case LIR_absd:			stream = callbackAbs<float64_t>(context, stream); break;

	case LIR_sqrtf:			stream = callbackSqrt<float32_t>(context, stream); break;
	case LIR_sqrtd:			stream = callbackSqrt<float64_t>(context, stream); break;

	case LIR_divf:			stream = callbackDiv<float32_t>(context, stream); break;
	case LIR_divd:			stream = callbackDiv<float64_t>(context, stream); break;

	case LIR_recipf:		stream = callbackRecip<float32_t>(context, stream); break;
	case LIR_recipd:		stream = callbackRecip<float64_t>(context, stream); break;

	case LIR_rsqrtf:		stream = callbackRsqrt<float32_t>(context, stream); break;
	case LIR_rsqrtd:		stream = callbackRsqrt<float64_t>(context, stream); break;

	case LIR_sti2c:			stream = callbackStore<MemoryByteOrder_Host, uint8_t, uint32_t>(context, stream); break;
	case LIR_sti2s:			stream = callbackStore<MemoryByteOrder_Host, uint16_t, uint32_t>(context, stream); break;
	case LIR_stri2s:		stream = callbackStore<MemoryByteOrder_Reversed, uint16_t, uint32_t>(context, stream); break;
	case LIR_sti:			stream = callbackStore<MemoryByteOrder_Host, uint32_t, uint32_t>(context, stream); break;
	case LIR_stri:			stream = callbackStore<MemoryByteOrder_Reversed, uint32_t, uint32_t>(context, stream); break;

#if defined(NANOJIT_64BIT)
	case LIR_stq:			stream = callbackStore<MemoryByteOrder_Reversed, uint64_t, uint64_t>(context, stream); break;
#endif

	case LIR_stf:			stream = callbackStore<MemoryByteOrder_Host, float32_t, float32_t>(context, stream); break;
	case LIR_std:			stream = callbackStore<MemoryByteOrder_Host, float64_t, float64_t>(context, stream); break;

	case LIR_lduc2ui:		stream = callbackLoad<MemoryByteOrder_Host, uint8_t, uint32_t>(context, stream); break;
	case LIR_ldus2ui:		stream = callbackLoad<MemoryByteOrder_Host, uint16_t, uint32_t>(context, stream); break;
	case LIR_ldrus2ui:		stream = callbackLoad<MemoryByteOrder_Reversed, uint16_t, uint32_t>(context, stream); break;
	case LIR_ldc2i:			stream = callbackLoad<MemoryByteOrder_Host, int8_t, int32_t>(context, stream); break;
	case LIR_lds2i:			stream = callbackLoad<MemoryByteOrder_Host, int16_t, int32_t>(context, stream); break;
	case LIR_ldrs2i:		stream = callbackLoad<MemoryByteOrder_Reversed, int16_t, int32_t>(context, stream); break;
	case LIR_ldi:			stream = callbackLoad<MemoryByteOrder_Host, uint32_t, uint32_t>(context, stream); break;
	case LIR_ldri:			stream = callbackLoad<MemoryByteOrder_Reversed, uint32_t, uint32_t>(context, stream); break;

#if defined(NANOJIT_64BIT)
	case LIR_ldq:			stream = callbackLoad<MemoryByteOrder_Reversed, uint64_t, uint64_t>(context, stream); break;
#endif

	case LIR_ldf:			stream = callbackLoad<MemoryByteOrder_Host, float32_t, float32_t>(context, stream); break;
	case LIR_ldd:			stream = callbackLoad<MemoryByteOrder_Host, float64_t, float64_t>(context, stream); break;

	case LIR_getflgz:		stream = callbackGetFlg(context, stream, ArithFlag_Zero); break;
	case LIR_getflgs:		stream = callbackGetFlg(context, stream, ArithFlag_Sign); break;
	case LIR_getflgc:		stream = callbackGetFlg(context, stream, ArithFlag_Carry); break;
	case LIR_getflgb:		stream = callbackGetFlg(context, stream, ArithFlag_Borrow); break;
	case LIR_getflgv:		stream = callbackGetFlg(context, stream, ArithFlag_Overflow); break;
	case LIR_getflgnz:		stream = callbackGetNotFlg(context, stream, ArithFlag_Zero); break;
	case LIR_getflgns:		stream = callbackGetNotFlg(context, stream, ArithFlag_Sign); break;
	case LIR_getflgnc:		stream = callbackGetNotFlg(context, stream, ArithFlag_Carry); break;
	case LIR_getflgnb:		stream = callbackGetNotFlg(context, stream, ArithFlag_Borrow); break;
	case LIR_getflgnv:		stream = callbackGetNotFlg(context, stream, ArithFlag_Overflow); break;
	case LIR_getflgfov:		stream = callbackGetFlg(context, stream, ArithFlag_FloatOverflow); break;
	case LIR_getflgfuf:		stream = callbackGetFlg(context, stream, ArithFlag_FloatUnderflow); break;
	case LIR_getflgfix:		stream = callbackGetFlg(context, stream, ArithFlag_FloatInvalid); break;


	case LIR_setflgc:		stream = callbackSetFlg(context, stream, ArithFlag_Carry); break;
	case LIR_setflgnc:		stream = callbackSetNotFlg(context, stream, ArithFlag_Carry); break;
	case LIR_setflgb:		stream = callbackSetFlg(context, stream, ArithFlag_Borrow); break;
	case LIR_setflgnb:		stream = callbackSetNotFlg(context, stream, ArithFlag_Borrow); break;

	case LIR_copyi:			stream = callbackCopy(context, stream); break;
#if defined(NANOJIT_64BIT)
	case LIR_copyq:			stream = callbackCopy(context, stream); break;
#endif
	case LIR_copyf:			stream = callbackCopy(context, stream); break;
	case LIR_copyd:			stream = callbackCopy(context, stream); break;

	case LIR_cmovi:			stream = callbackCmov<uint32_t, int32_t>(context, stream); break;
#if defined(NANOJIT_64BIT)
	case LIR_cmovq:			stream = callbackCmov<uint64_t, int64_t>(context, stream); break;
#endif
	case LIR_cmovf:			stream = callbackCmov<float32_t, float32_t>(context, stream); break;
	case LIR_cmovd:			stream = callbackCmov<float64_t, float64_t>(context, stream); break;

	case LIR_branch:		stream = callbackBranch(stream); break;
	case LIR_branch_cond:	stream = callbackBranchCond(context, stream); break;

	case LIR_callv:			stream = callbackExternalCall(context, LTy_V, stream); break;
	case LIR_calli:			stream = callbackExternalCall(context, LTy_I, stream); break;
#if defined(NANOJIT_64BIT)
	case LIR_callq:			stream = callbackExternalCall(context, LTy_Q, stream); break;
#endif
	case LIR_callf:			stream = callbackExternalCall(context, LTy_F, stream); break;
	case LIR_calld:			stream = callbackExternalCall(context, LTy_D, stream); break;

	case LIR_genflgsint:	stream = callbackGenFlagsInt(context, stream); break;
	case LIR_genflgsflt:	stream = callbackGenFlagsFlt(context, stream); break;

#if defined(NANOJIT_64BIT)
	case LIR_movsi2sq:		stream = callbackMov<int32_t, int64_t>(context, stream); break;
	case LIR_movui2uq:		stream = callbackMov<uint32_t, uint64_t>(context, stream); break;
	case LIR_movq2i:		stream = callbackMov<uint64_t, uint32_t>(context, stream); break;
#endif
	case LIR_movi2f:		stream = callbackMov<uint32_t, float32_t>(context, stream); break;
	case LIR_movf2i:		stream = callbackMov<float32_t, uint32_t>(context, stream); break;
	case LIR_movi2d:		stream = callbackMov<uint32_t, float64_t>(context, stream); break;
	case LIR_movd2i:		stream = callbackMov<float64_t, uint32_t>(context, stream); break;
#if defined(NANOJIT_64BIT)
	case LIR_movq2d:		stream = callbackMov<uint64_t, float64_t>(context, stream); break;
	case LIR_movd2q:		stream = callbackMov<float64_t, uint64_t>(context, stream); break;
#endif
	case LIR_movf2d:		stream = callbackMov<float32_t, float64_t>(context, stream); break;
	case LIR_movd2f:		stream = callbackMov<float64_t, float32_t>(context, stream); break;

	case LIR_cvtif:			stream = callbackCvt<float32_t, int32_t>(context, stream); break;
	case LIR_cvtuif:		stream = callbackCvtu<float32_t, uint32_t>(context, stream); break;
	case LIR_cvtitf:		stream = callbackCvtt<float32_t, int32_t>(context, stream); break;
	case LIR_cvtuitf:		stream = callbackCvtut<float32_t, uint32_t>(context, stream); break;
	case LIR_cvfif:			stream = callbackCvf<float32_t, int32_t>(context, stream); break;
	case LIR_cvfuif:		stream = callbackCvf<float32_t, uint32_t>(context, stream); break;
	case LIR_cvtid:			stream = callbackCvt<float64_t, int32_t>(context, stream); break;
	case LIR_cvtuid:		stream = callbackCvtu<float64_t, uint32_t>(context, stream); break;
	case LIR_cvtitd:		stream = callbackCvtt<float64_t, int32_t>(context, stream); break;
	case LIR_cvtuitd:		stream = callbackCvtut<float64_t, uint32_t>(context, stream); break;
	case LIR_cvfid:			stream = callbackCvf<float64_t, int32_t>(context, stream); break;
	case LIR_cvfuid:		stream = callbackCvf<float64_t, uint32_t>(context, stream); break;
#if defined(NANOJIT_64BIT)
	case LIR_cvtqd:			stream = callbackCvt<float64_t, int64_t>(context, stream); break;
	case LIR_cvtuqd:		stream = callbackCvtu<float64_t, uint64_t>(context, stream); break;
	case LIR_cvtqtd:		stream = callbackCvtt<float64_t, int64_t>(context, stream); break;
	case LIR_cvtuqtd:		stream = callbackCvtut<float64_t, uint64_t>(context, stream); break;
	case LIR_cvfqd:			stream = callbackCvf<float64_t, int64_t>(context, stream); break;
	case LIR_cvfuqd:		stream = callbackCvf<float64_t, uint64_t>(context, stream); break;
#endif
	case LIR_cvffd:			stream = callbackCvf<float64_t, float32_t>(context, stream); break;
	case LIR_cvtfd:			stream = callbackCvt<float64_t, float32_t>(context, stream); break;

	case LIR_spilli:		stream = callbackSpill<uint32_t>(context, stream); break;
#if defined(NANOJIT_64BIT)	
	case LIR_spillq:		stream = callbackSpill<uint64_t>(context, stream); break;
#endif	
	case LIR_spillf:		stream = callbackSpill<float32_t>(context, stream); break;
	case LIR_spilld:		stream = callbackSpill<float64_t>(context, stream); break;

	case LIR_restorei:		stream = callbackRestore<uint32_t>(context, stream); break;
#if defined(NANOJIT_64BIT)
	case LIR_restoreq:		stream = callbackRestore<uint64_t>(context, stream); break;
#endif
	case LIR_restoref:		stream = callbackRestore<float32_t>(context, stream); break;
	case LIR_restored:		stream = callbackRestore<float64_t>(context, stream); break;

	case LIR_pushi:			stream = callbackPush<uint32_t>(context, stream); break;
#if defined(NANOJIT_64BIT)
	case LIR_pushq:			stream = callbackPush<uint64_t>(context, stream); break;
#endif
	case LIR_pushd:			stream = callbackPush<float32_t>(context, stream); break;
	case LIR_pushf:			stream = callbackPush<float64_t>(context, stream); break;

	case LIR_popi:			stream = callbackPop<uint32_t>(context, stream); break;
#if defined(NANOJIT_64BIT)
	case LIR_popq:			stream = callbackPop<uint64_t>(context, stream); break;
#endif
	case LIR_popd:			stream = callbackPop<float32_t>(context, stream); break;
	case LIR_popf:			stream = callbackPop<float64_t>(context, stream); break;

	case LIR_copyreg:		stream = callbackCopy(context, stream); break;

	case LIR_addimm32i:		stream = callbackAddImm32<uint32_t>(context, stream); break;
#if defined(NANOJIT_64BIT)
	case LIR_addimm32q:		stream = callbackAddImm32<uint64_t>(context, stream); break;
#endif
	
	case LIR_clearfpflags:	stream = callbackClearFpFlags(context, stream); break;

	case LIR_exit:			stream = NULL; break;
	default:				NanoAssertMsgf(0, "unsupported opcode %d\n", op); stream = NULL; break;
	}

	return stream;
}

}


namespace nanojit
{
namespace interpreter
{

void runInterpreter(IntepreterContext* context, const void* stream, void* framePointer) {

	context->set<uintptr_t>(nanojit::REGNUM(nanojit::FP), reinterpret_cast<uintptr_t>(framePointer));

	const uint8_t* instructionStream = reinterpret_cast<const uint8_t*>(stream);

	while (instructionStream != NULL) {
		instructionStream = dispatch(context, instructionStream);
	}
}

#if defined(NANOJIT_VERBOSE)

const char* getInterpreterOpcodeString(unsigned int op) {

	if (op <= LIR_sentinel) {
		return getLirName(static_cast<LOpcode>(op));
	} else {

		switch (op) {
		case LIR_genflgsint:		return "genflgsint";
		case LIR_genflgsflt:		return "genflgsflt";
		case LIR_branch:			return "branch";
		case LIR_branch_cond:		return "branch_cond";
		case LIR_spilli:			return "spilli";
#if defined(NANOJIT_64BIT)
		case LIR_spillq:			return "spillq";
#endif
		case LIR_spillf:			return "spillf";
		case LIR_spilld:			return "spilld";
		case LIR_restorei:			return "restorei";
#if defined(NANOJIT_64BIT)
		case LIR_restoreq:			return "restoreq";
#endif
		case LIR_restoref:			return "restoref";
		case LIR_restored:			return "restored";
		case LIR_pushi:				return "pushi";
#if defined(NANOJIT_64BIT)
		case LIR_pushq:				return "pushq";
#endif
		case LIR_pushf:				return "pushf";
		case LIR_pushd:				return "pushd";
		case LIR_popi:				return "popi";
#if defined(NANOJIT_64BIT)
		case LIR_popq:				return "popq";
#endif
		case LIR_popf:				return "popf";
		case LIR_popd:				return "popd";
		case LIR_copyreg:			return "copyreg";
		case LIR_addimm32i:			return "adddimm32i";
#if defined(NANOJIT_64BIT)
		case LIR_addimm32q:			return "adddimm32q";
#endif
		case LIR_clearfpflags:		return "clearfpflags";

		default:					NanoAssert(0); return "!invalid opcode!";
		}
	}
}

#endif

} // namespace interpreter
} // namespace nanojit

