/* -*- Mode: C++; c-basic-offset: 4; indent-tabs-mode: nil; tab-width: 4 -*- */
/* vi: set ts=4 sw=4 expandtab: (add to ~/.vimrc: set modeline modelines=5) */
/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

#include "nanojit.h"

#if defined(NANOJIT_VERBOSE)
#include <stdarg.h>
#endif


#ifdef _MSC_VER
	// disable some specific warnings which are normally useful, but pervasive in the code-gen macros
	#pragma warning(disable:4310) // cast truncates constant value
#endif


#if defined(NANOJIT_64BIT)
NanoStaticAssert(sizeof(void*) == 8);
NanoStaticAssert(sizeof(uintptr_t) == 8);
#else // assume 32bit
NanoStaticAssert(sizeof(void*) == 4);
NanoStaticAssert(sizeof(uintptr_t) == 4);
#endif

namespace nanojit
{
	/**
	 * Need the following:
	 *
	 * - merging paths (build a graph?), possibly use external rep to drive codegen
	 */
	Assembler::Assembler(
		CodeAlloc* codeAlloc, Allocator* internalAlloc, 
		const Config& config
		verbose_only(, LogControl* logc)
		verbose_only(, LInsPrinter* printer))
		: verbose_only(_logc(logc),)
		  verbose_only(_printer(printer),)
		  _internalAlloc(internalAlloc)
		, _codeAlloc(codeAlloc)
		, _relativeBranchPatchList(internalAlloc)
		, _relativeBranchRelocationList(internalAlloc)
		, _entryPointList(internalAlloc)
		, _relativeOffsetPatchList(internalAlloc)
		, _exitPointPatchList(internalAlloc)
		, _noise(NULL)
	#if NANOJIT_USES_IMMD_POOL
		, _immDPool(internalAlloc)
	#endif
		, _codeBlockList(NULL)
		, _err(None)
	#ifdef VMCFG_VTUNEl
		, vtuneHandle(NULL)
	#endif
		, _config(config) {

		verbose_only( _outputCache = 0; )
		verbose_only( outline[0] = '\0'; )
		verbose_only( outlineEOL[0] = '\0'; )

		reset(0);

		nCreate();
	}

	Assembler::~Assembler() {
		nDestroy();
	}

	void Assembler::arReset(unsigned int arGroupCount) {
		_activation.reset(arGroupCount);

	#if NANOJIT_USES_IMMD_POOL
		_immDPool.clear();
	#endif
	}

	void Assembler::setBlockMargins(CodeBlock* block, NIns* start, NIns* end) {

		NanoAssert(block->isInBlock(start));
		NanoAssert(start >= block->internalBufferStart());
		NanoAssert(end <= block->internalBufferEnd());
		NanoAssert(start <= end);

		size_t marginStart = uintptr_t(start) - uintptr_t(block->internalBufferStart());
		size_t marginEnd = uintptr_t(block->internalBufferEnd()) - uintptr_t(end);

		block->setCodeStartOffset(marginStart);
		block->setCodeEndOffset(marginEnd);

		NanoAssert(block->codeSize() == (uintptr_t(end) - uintptr_t(start)));
	}

	void Assembler::codeAlloc() {
		// CodeAlloc contract: allocations never fail
		if (currentCodeBlock != NULL) {
			setBlockMargins(currentCodeBlock, _nIns, _nLastIns);
			_codeBlockList.addBlockToList(currentCodeBlock);
		}

		CodeBlock* newBlock = _codeAlloc->alloc();
		verbose_only( codeBytes += newBlock->codeSize(); )
		NanoAssert(newBlock->internalBufferSize() >= CodeAlloc::minAllocSize);
		_nIns = newBlock->internalBufferEnd();
		verbose_only( _nInsAfter = _nIns; )

		currentCodeBlock = newBlock;
	}

	void Assembler::finalizeCodeAlloc() {
		verbose_only( codeBytes -= (_nIns - currentCodeBlock->internalBufferStart()) * sizeof(NIns); )
		CodeBlock* remainderCodeBlock = _codeAlloc->addRemainder(currentCodeBlock, _nIns);

		if (remainderCodeBlock != NULL) {
			setBlockMargins(remainderCodeBlock, _nIns, _nLastIns);
			_codeBlockList.addBlockToList(remainderCodeBlock);
		}

		currentCodeBlock = NULL;

		if (!_codeBlockList.isEmpty()) {
			// note: the code pages are no longer writable from this point onwards
			_codeAlloc->markListExec(&_codeBlockList);

			// at this point all our new code is in the d-cache and not the i-cache,
			// so flush the i-cache on cpu's that need it.
			_codeAlloc->flushListIcache(&_codeBlockList);
		}

		clearNInsPtrs();
	}

	void Assembler::clearNInsPtrs() {
		_nIns = NULL;
		_nLastIns = NULL;
		currentCodeBlock = NULL;
	}

	void Assembler::reset(unsigned int arGroupCount) {
		clearNInsPtrs();
		nativePageReset();
		_allocator.initialize(&_config);
		arReset(arGroupCount);
	}

	#ifdef NANOJIT_DEBUG
	void Assembler::pageValidate() {
		if (hasError()) return;
		// This may be a normal code chunk or an exit code chunk.
		NanoAssertMsg(currentCodeBlock->internalBufferStart() <= _nIns && _nIns <= currentCodeBlock->internalBufferEnd(),
					 "Native instruction pointer overstep paging bounds; check overrideProtect for last instruction");
	}
	#endif

	#ifdef NANOJIT_DEBUG

	void Assembler::resourceConsistencyCheck() {
		NanoAssert(!hasError());
#ifdef NANOJIT_IA32
		// Within the expansion of a single LIR instruction, we may use the x87
		// stack for unmanaged temporaries. Otherwise, we do not use the x87 stack
		// as such, but use the top element alone as a single allocatable FP register.
		// Compensation code must be inserted to keep the stack balanced and avoid
		// overflow, and the mechanisms for this are rather fragile and IA32-specific.
		// The predicate below should hold between any pair of instructions within
		// a basic block, at labels, and just after a conditional branch. Currently,
		// we enforce this condition between all pairs of instructions, but this is
		// overly restrictive, and would fail if we did not generate unreachable x87
		// stack pops following unconditional branches.
		NanoAssert((_allocator.getActive(FST0) && _fpuStkDepth == -1) ||
				   (!_allocator.getActive(FST0) && _fpuStkDepth == 0));
#endif
		_activation.checkForResourceConsistency(_allocator);
		registerConsistencyCheck();
	}

	void Assembler::registerConsistencyCheck() {
		RegisterMask managed = _allocator.getManagedSet();
		for (Register r = lsReg(managed); managed; r = nextLsReg(managed, r)) {
			// A register managed by register allocation must be either
			// free or active, but not both.
			if (_allocator.isFree(r)) {
				NanoAssertMsgf(_allocator.getActive(r) == 0, "register %s is free but assigned to ins", gpn(r));
			} else {
				// An LIns defining a register must have that register in
				// its reservation.
				LIns* ins = _allocator.getActive(r);
				NanoAssert(ins);
				NanoAssert(ins->isInReg());
				/* note: registers may overlap; but if they do, the primary reg. defined by the instruction must completely contain
				   all the overlapping registers. */
				NanoAssertMsg( (rmask(r) & rmask(ins->getReg()) ) == rmask(r), "Register record mismatch");
			}
		}

		RegisterMask not_managed = ~_allocator.getManagedSet();
		for (Register r = lsReg(not_managed); not_managed; r = nextLsReg(not_managed, r)) {
			// A register not managed by register allocation must be
			// neither free nor active.
			if (REGNUM(r) <= LastRegNum) {
				NanoAssert(!_allocator.isFree(r));
				NanoAssert(!_allocator.getActive(r));
			}
		}
	}
	#endif /* NANOJIT_DEBUG */

	void Assembler::findRegFor2(RegisterMask allowa, LIns* ia, Register& ra,
								RegisterMask allowb, LIns* ib, Register& rb) {
		// There should be some overlap between 'allowa' and 'allowb', else
		// there's no point calling this function.
		NanoAssert(allowa & allowb);

		if (ia == ib) {
			ra = rb = findRegFor(ia, allowa & allowb); // use intersection(allowa, allowb)

		} else if (ib->isInRegMask(allowb)) {
			// 'ib' is already in an allowable reg -- don't let it get evicted
			// when finding 'ra'.
			rb = ib->getReg();
			ra = findRegFor(ia, allowa & ~rmask(rb));

		} else {
			ra = findRegFor(ia, allowa);
			rb = findRegFor(ib, allowb & ~rmask(ra));
		}
	}


	Register Assembler::allocReg(LIns* ins, RegisterMask allow) {
		NanoAssert(ins->isLive());
		NanoAssert(!ins->isInReg());

		// prevent copy ops to become extant if not necessary
		if (
				ins->isCopy()
			&&	(!ins->isExtant())
			&&	(!ins->getGenFlags())
			&&	ins->canOptimize()
			&&	ins->oprnd1()->isInReg()) {

			Register r = ins->oprnd1()->getReg();
			if ((allow & rmask(r)) != 0)
				return r;
		}

		// 'ins' isn't in a register (must be in a spill slot or nowhere).
		Register r = _allocator.allocReg(ins, allow);

		if (r == UnspecifiedReg) {
			LIns* vic = _allocator.findVictim(ins->retType(), allow);
			NanoAssert(vic->isInReg());
			r = vic->getReg();

			evict(vic);

			r = _allocator.allocSpecificReg(ins, r);
		}

		return r;
	}

	Register Assembler::allocTempReg(LTy insType, RegisterMask allow) {
		Register r;

		r = _allocator.findFreeReg(insType, allow, 0);

		if (r == UnspecifiedReg) {
			LIns* vic = _allocator.findVictim(insType, allow);
			NanoAssert(vic->isInReg());
			r = vic->getReg();
			evict(vic);
		}

		r = _allocator.allocTempRegIfAvailable(insType, rmask(r));
		NanoAssert(r != UnspecifiedReg);
		return r;
	}

	Register Assembler::moveToReg(LIns* ins, RegisterMask allow) {
		Register        r       = ins->getReg();
		RegisterMask    rm      = rmask(r);

		if ((rm & allow) == rm) {
			// 'ins' is in an allowed register.
			_allocator.useActive(r);
			return r;
		}

		// 'ins' is in a register (r) that's not in 'allow'.
		RegisterMask copyCandidates = _allocator.nRegCopyCandidates(r, allow);

		RegisterMask rmallow = allow & copyCandidates;

		RegisterMask rmFree = _allocator.freeMask();

		Register rTestAlloc = firstAvailableReg(ins->retType(), rmFree & rmallow);
		bool canMoveToFreeReg = (rTestAlloc != UnspecifiedReg);

		if (canMoveToFreeReg) {

			Register s = r;
			_allocator.retire(r);
			ins->clearReg();

			r = allocReg(ins, rmallow);

			if ((rmask(s) & GpRegs) && (rmask(r) & GpRegs)) {
				nGpCopy(s, r);
			} else {
				nNonGpCopy(s, r);
			}

		} else {

			LIns* vic = _allocator.findVictim(ins->retType(), rmallow);
			Register s = r;
			r = vic->getReg();
			evict(vic);
			_allocator.retire(s);
			_allocator.allocSpecificReg(ins, r);

			if ((rmask(s) & GpRegs) && (rmask(r) & GpRegs)) {
				nGpCopy(s, r);
			} else {
				nNonGpCopy(s, r);
			}

		};

		return r;
	}

	// Finds a register in 'allow' to hold the result of 'ins'.  Used when we
	// encounter a use of 'ins'.  The actions depend on the prior regstate of
	// 'ins':
	// - If the result of 'ins' is not in any register, we find an allowed
	//   one, evicting one if necessary.
	// - If the result of 'ins' is already in an allowed register, we use that.
	// - If the result of 'ins' is already in a not-allowed register, we find an
	//   allowed one and move it.
	//
	Register Assembler::findRegFor(LIns* ins, RegisterMask allow) {
		 /* Make sure we'll get a register from a proper class */
		 NanoAssert(allow);

		if (!ins->isInReg()) {
			// 'ins' isn't in a register (must be in a spill slot or nowhere).
			return allocReg(ins, allow);
		} else if ((rmask(ins->getReg()) & allow) == 0) {
			return moveToReg(ins, allow);
		}

		return ins->getReg();
	}

	// Like findSpecificRegFor(), but only for when 'r' is known to be free
	// and 'ins' is known to not already have a register allocated.  Updates
	// the regstate (maintaining the invariants) but does not generate any
	// code.  The return value is redundant, always being 'r', but it's
	// sometimes useful to have it there for assignments.
	Register Assembler::findSpecificRegForUnallocated(LIns* ins, Register r) {
		NanoAssert(!ins->isInReg());
		_allocator.allocSpecificReg(ins,r);
		return r;
	}

#if NANOJIT_USES_IMMD_POOL
	const uint64_t* Assembler::findImmDFromPool(uint64_t q) {
		uint64_t* p = _immDPool.get(q);
		if (!p) {
			p = new (_dataAlloc) uint64_t;
			*p = q;
			_immDPool.put(q, p);
		}
		return p;
	}
#endif

	int Assembler::findMemFor(LIns *ins) {
		NanoAssert(!NANOJIT_USES_IMMD_POOL || !(ins->isImmD() || ins->isImmF()));
		NanoAssert(!RegAlloc::nCanRemat(ins));

		if (!ins->isInAr()) {
			unsigned int arGroup = ins->getArGroup();

			if (arGroup != LINS_DEFAULT_AR_GROUP) {
				unsigned int arIndex = arGroup;
				reallocateArEntry(ins, arIndex);
			} else {
				uint32_t arIndex = arReserve(ins);
				ins->setArIndex(arIndex);
			}

#if defined(NANOJIT_DEBUG)
			NanoAssert(_activation.isValidEntry(ins->getArIndex(), ins));
#endif

		} else {
			NanoAssert(ins->isInAr());
		}

		NanoAssert((ins->getArGroup() == LINS_DEFAULT_AR_GROUP) || (ins->getArIndex() == ins->getArGroup()));

		return arDisp(ins->getArIndex());
	}

	void Assembler::reallocateArEntry(LIns* ins, uint32_t entry) {
		NanoAssert(!ins->isInAr());
		NanoAssert(!RegAlloc::nCanRemat(ins));
		_activation.reallocateEntry(ins, entry);
		ins->setArIndex(entry);
	}

	// Finds a register in 'allow' to hold the result of 'ins'.  Also
	// generates code to spill the result if necessary.  Called just prior to
	// generating the code for 'ins' (because we generate code backwards).
	//
	// An example where no spill is necessary.  Lines marked '*' are those
	// done by this function.
	//
	//   regstate:  R
	//   asm:       define res into r
	// * regstate:  R + r(res)
	//              ...
	//   asm:       use res in r
	//
	// An example where a spill is necessary.
	//
	//   regstate:  R
	//   asm:       define res into r
	// * regstate:  R + r(res)
	// * asm:       spill res from r
	//   regstate:  R
	//              ...
	//   asm:       restore res into r2
	//   regstate:  R + r2(res) + other changes from "..."
	//   asm:       use res in r2
	//
	Register Assembler::prepareResultReg(LIns *ins, RegisterMask allow) {
		// At this point, we know the result of 'ins' is used later in the
		// code, unless it is a call to an impure function that must be
		// included for effect even though its result is ignored.  It may have
		// had to be evicted, in which case the restore will have already been
		// generated, so we now generate the spill.  QUERY: Is there any attempt
		// to elide the spill if we know that all restores can be rematerialized?
#ifdef NANOJIT_IA32
		const bool notInFST0 = (!ins->isInReg() || ins->getReg() != FST0);
		Register r = findRegFor(ins, allow);
		// If the result register is FST0, but FST0 is not in the post-regstate,
		// then we must pop the x87 stack.  This may occur because the result is
		// unused, or because it has been stored to a spill slot or an XMM register.
		const bool needPop = notInFST0 && (r == FST0);
		const bool didSpill = asm_maybe_spill(ins, needPop);
		if (!didSpill && needPop) {
			// If the instruction is spilled, then the pop will have already
			// been performed by the store to the stack slot.  Otherwise, we
			// must pop now.  This may occur when the result of a LIR_calld
			// to an impure (side-effecting) function is not used.
			FSTP(FST0);
		}
#else
		Register r = UnspecifiedReg;
		r = findRegFor(ins, allow);
		asm_maybe_spill(ins, false);
#endif
		return r;
	}

	bool Assembler::asm_maybe_spill(LIns* ins, bool pop) {
		NanoUnusedVariable(pop);

		if (ins->isInAr()) {

			NanoAssert(!RegAlloc::nCanRemat(ins));

			verbose_only(
				RefBuf b;
				if (_logc->getLcBits() & LC_Native) {
					setOutputForEOL(" <= spill %s", _printer->formatRef(&b, ins));
				}
				compilationStats.spillCount++;
			)

			nSpill(ins);
			return true;
		}
		return false;
	}


	// Frees all record of registers and spill slots used by 'ins'.
	void Assembler::freeResourcesOf(LIns *ins) {

		if (ins->isInReg()) {
			_allocator.retire(ins->getReg());   // free any register associated with entry
			ins->clearReg();
		}

		if (ins->isInAr()) {
			arRelease(ins);                     // free any AR space associated with entry
			ins->clearAr();
		}

		ins->setResultDead();
	}

	// Frees 'r' in the RegAlloc regstate, if it's not already free.
	void Assembler::evictIfActive(Register r) {
		if (LIns* vic = _allocator.getActive(r)) {
			NanoAssert(vic->getReg() == r);
			evict(vic);
		}
	}

	// Frees 'r' (which currently holds the result of 'vic') in the regstate.
	// An example:
	//
	//   pre-regstate:  eax(ld1)
	//   instruction:   mov ebx,-4(ebp) <= restore add1   # %ebx is dest
	//   post-regstate: eax(ld1) ebx(add1)
	//
	// At run-time we are *restoring* 'add1' into %ebx, hence the call to
	// nRestore().  But at regalloc-time we are moving backwards through
	// the code, so in that sense we are *evicting* 'add1' from %ebx.
	//
	void Assembler::evict(LIns* ins) {
		LIns* vic = ins;
		Register r = vic->getReg();

		// Not free, need to steal.
		NanoAssert(!_allocator.isFree(r));
		NanoAssert(vic == _allocator.getActive(r));

		verbose_only(
			RefBuf b;
			if (_logc->getLcBits() & LC_Native) {
				setOutputForEOL(" <= restore %s", _printer->formatRef(&b, vic));
			}
			compilationStats.restoreCount++;
		)


		// will call findMemFor()
		nRestore(vic, r);


		_allocator.retire(r);
		vic->clearReg();

		// At this point 'vic' is unused (if rematerializable), or in a spill
		// slot (if not).
	}

	void Assembler::compile(LirChainWriter* writer, unsigned int arGroupCount) {
		LirStorage* storage     = writer->getStorage();

		verbose_only(
		bool anyVerb        = (_logc->getLcBits() & 0xFFFF & ~LC_FragProfile) > 0;
		bool liveVerb       = (_logc->getLcBits() & 0xFFFF & LC_Liveness) > 0;
		bool readLirVerb    = (_logc->getLcBits() & 0xFFFF & LC_ReadLIR) > 0;
		)

		_relativeBranchPatchList.clear();
		_relativeBranchPatchList.reserve(256);

		_relativeBranchRelocationList.clear();
		_relativeBranchRelocationList.reserve(256);

		_entryPointList.clear();
		_entryPointList.reserve(256);

		_relativeOffsetPatchList.clear();
		_relativeOffsetPatchList.reserve(256);

		_exitPointPatchList.clear();
		_exitPointPatchList.reserve(256);

		/* BEGIN decorative preamble */
		verbose_only(
		if (anyVerb) {
			_logc->printf("========================================"
						  "========================================\n");
			_logc->printf("=== BEGIN LIR::compile\n");
			_logc->printf("===\n");
		})
		/* END decorative preamble */

		verbose_only( if (liveVerb) {
			_logc->printf("\n");
			_logc->printf("=== Results of liveness analysis:\n");
			_logc->printf("===\n");
			LirReader br(storage);
			LirFilter* lir = &br;

			live(lir, _internalAlloc, _printer, _logc);

			_logc->flush();
		})

		verbose_only(if (readLirVerb) {
			LirReader br(storage);
			ReverseLister* pInitialLister = new (_internalAlloc) ReverseLister(
				&br, _internalAlloc, _printer, 
				_logc, "Initial LIR");			
			LirFilter* lir = pInitialLister;
			for (currIns = lir->read(); !currIns->isop(LIR_start); currIns = lir->read()) {
			}
			pInitialLister->finish();
			_logc->flush();
		})



		/* Set up the generic text output cache for the assembler */
		verbose_only( StringList asmOutput(_internalAlloc); )
		verbose_only( _outputCache = &asmOutput; )

		beginAssembly(arGroupCount);
		if (hasError())
			return;

		verbose_only( if (anyVerb) {
			_logc->printf("=== Translating LIR fragments into assembly:\n");
		})

		// now the the main trunk

		// Used for debug printing, if needed
		debug_only(ValidateReader *validate = NULL;)
		verbose_only(
		ReverseLister *pp_init = NULL;
		ReverseLister *pp_after_sf = NULL;
		)

		// The LIR passes through these filters as listed in this
		// function, viz, top to bottom.

		// set up backwards pipeline: assembler <- LirReader
		LirFilter* lir = new (_internalAlloc) LirReader(storage);

#ifdef NANOJIT_DEBUG
		// VALIDATION
		validate = new (_internalAlloc) ValidateReader(lir);
		lir = validate;
#endif

		assemble(lir);

		// If we were accumulating debug info in the various ReverseListers,
		// call finish() to emit whatever contents they have accumulated.
		verbose_only(
		if (pp_init)        pp_init->finish();
		if (pp_after_sf)    pp_after_sf->finish();
		)

		endAssembly();

		// Reverse output so that assembly is displayed low-to-high.
		// Up to this point, _outputCache has been non-NULL, and so has been
		// accumulating output.  Now we set it to NULL, traverse the entire
		// list of stored strings, and hand them a second time to output.
		// Since _outputCache is now NULL, outputf just hands these strings
		// directly onwards to _logc->printf.
		verbose_only( if (anyVerb) {
			_logc->printf("\n");
			_logc->printf("=== Aggregated assembly output: BEGIN\n");
			_logc->printf("===\n");
			_outputCache = 0;
			for (Seq<char*>* p = asmOutput.get(); p != NULL; p = p->tail) {
				char *str = p->head;
				outputf("  %s", str);
			}
			_logc->printf("===\n");
			_logc->printf("=== Aggregated assembly output: END\n");
		});

		/* BEGIN decorative postamble */
		verbose_only( if (anyVerb) {
			_logc->printf("\n");
			_logc->printf("===\n");
			_logc->printf("=== END LIR::compile\n");
			_logc->printf("========================================"
						  "========================================\n");
			_logc->printf("\n");
		});

		/* END decorative postamble */
		
		_relativeBranchPatchList.clear();

		verbose_only(_logc->flush());
	}


	void Assembler::beginAssembly(unsigned int arGroupCount) {
		verbose_only( codeBytes = 0; )

		reset(arGroupCount);

		NanoAssert(_codeBlockList.isEmpty());
		NanoAssert(_nIns == NULL);

		setError(None);

		// native code gen buffer setup
		nativePageSetup();

		// make sure we got memory at least one page
		if (hasError()) return;

		verbose_only( _nInsAfter = _nIns; )

		nBeginAssembly(_internalAlloc);
	}

	CodeBlock* Assembler::findOwnerBlock(const NIns* ins) {
		if ((currentCodeBlock != NULL) && currentCodeBlock->isInBlock(ins))
			return currentCodeBlock;

		CodeBlock* finishedCodeBlock = _codeBlockList.head();
		while (finishedCodeBlock != NULL) {
			if (finishedCodeBlock->isInBlock(ins))
				return finishedCodeBlock;

			finishedCodeBlock = finishedCodeBlock->next();
		}

		return NULL;
	}

	ptrdiff_t Assembler::computeLinearOffset(const NIns* ins, CodeBlock const** ownerBlock) const {

		NanoAssert(currentCodeBlock == NULL); // cannot be called during compilation

		const CodeBlock* block = _codeBlockList.head();
		ptrdiff_t offset = 0;

		while (block != NULL) {

			if (block->isInBlock(ins)) {
				*ownerBlock = block;
				return offset + block->offsetFromCodeStartInBytes(ins);
			}

			offset += block->codeSize();
			block = block->next();
		}
		NanoAssert(0);
		return 0;
	}

	void Assembler::assemble(LirFilter* reader) {
		if (hasError()) return;

		verbose_only(
			memset(&compilationStats, 0, sizeof(compilationStats));
		)

		gen(reader);

		if (!hasError()) {

			BranchPatchList::Iter it = _relativeBranchPatchList.begin();

			while (it != _relativeBranchPatchList.end()) {
				nPatchRelativeBranch(it->first, it->second->labelNativeAddress());
				_relativeBranchRelocationList.push_back(makePair(it->first, it->second->labelNativeAddress()));
				++it;
			}
		}
	}

	void Assembler::cleanupAfterError() {
		_codeAlloc->freeList(&_codeBlockList);
		NanoAssert(_codeBlockList.isEmpty());
		_nIns = NULL;
		_codeAlloc->markAllExec(); // expensive but safe, we mark all code pages R-X
	}

	void Assembler::endAssembly() {
		// don't try to patch code if we are in an error state since we might have partially
		// overwritten the code cache already
		if (hasError()) {
			// something went wrong, release all allocated code memory
			cleanupAfterError();
			return;
		}

		nPrologue();

		nEndAssembly(_internalAlloc);


		verbose_only( asm_output("[prologue]"); )
#if defined(NANOJIT_VERBOSE)
		debug_only(_activation.checkForResourceLeaks(_printer, _logc));
#endif

		NanoAssert(_activation.stackTop() == _activation.firstEntryAllocationIndex());

		debug_only( pageValidate(); )

		finalizeCodeAlloc();

#ifdef NANOJIT_IA32
		NanoAssertMsgf(_fpuStkDepth == 0,"_fpuStkDepth %d\n",_fpuStkDepth);
#endif

		verbose_only( _logc->flush() );
	}

#ifdef PERFM
#define countlir_live() _nvprof("lir-live",1)
#define countlir_ret() _nvprof("lir-ret",1)
#define countlir_alloc() _nvprof("lir-alloc",1)
#define countlir_var() _nvprof("lir-var",1)
#define countlir_use() _nvprof("lir-use",1)
#define countlir_def() _nvprof("lir-def",1)
#define countlir_imm() _nvprof("lir-imm",1)
#define countlir_param() _nvprof("lir-param",1)
#define countlir_cmov() _nvprof("lir-cmov",1)
#define countlir_ld() _nvprof("lir-ld",1)
#define countlir_ldq() _nvprof("lir-ldq",1)
#define countlir_alu() _nvprof("lir-alu",1)
#define countlir_qjoin() _nvprof("lir-qjoin",1)
#define countlir_qlo() _nvprof("lir-qlo",1)
#define countlir_qhi() _nvprof("lir-qhi",1)
#define countlir_fpu() _nvprof("lir-fpu",1)
#define countlir_st() _nvprof("lir-st",1)
#define countlir_stq() _nvprof("lir-stq",1)
#define countlir_jmp() _nvprof("lir-jmp",1)
#define countlir_jcc() _nvprof("lir-jcc",1)
#define countlir_label() _nvprof("lir-label",1)
#define countlir_xcc() _nvprof("lir-xcc",1)
#define countlir_x() _nvprof("lir-x",1)
#define countlir_call() _nvprof("lir-call",1)
#define countlir_cvt() _nvprof("lir-cvt",1)
#else
#define countlir_live()
#define countlir_ret()
#define countlir_alloc()
#define countlir_var()
#define countlir_use()
#define countlir_def()
#define countlir_imm()
#define countlir_param()
#define countlir_cmov()
#define countlir_ld()
#define countlir_ldq()
#define countlir_alu()
#define countlir_qjoin()
#define countlir_qlo()
#define countlir_qhi()
#define countlir_fpu()
#define countlir_st()
#define countlir_stq()
#define countlir_jmp()
#define countlir_jcc()
#define countlir_label()
#define countlir_xcc()
#define countlir_x()
#define countlir_call()
#define countlir_cvt()
#endif

	void Assembler::asm_jmp(LIns* ins) {

		NanoAssert((ins->isop(LIR_j) && (ins->branchCondition() == NULL)) ||
				   (ins->isop(LIR_jf) && (ins->branchCondition()->isImmI(0))) ||
				   (ins->isop(LIR_jt) && (ins->branchCondition()->isImmI(1))));

		countlir_jmp();
		LIns* to = ins->branchTarget();

#ifdef NANOJIT_IA32
		// Unreachable, so assume correct stack depth.
		debug_only( _fpuStkDepth = 0; )
#endif
		
		NanoAssert(to->isLInsLabel());

		if (to->labelNativeAddress() != 0) {
#ifdef NANOJIT_IA32
			// Set stack depth according to the register state we just loaded,
			// negating the effect of any unreachable x87 stack pop that might
			// have been emitted by unionRegisterState().
			debug_only( _fpuStkDepth = (_allocator.getActive(FST0) ? -1 : 0); )
#endif
			NIns* branch = nRelativeBranch(to->labelNativeAddress());
			_relativeBranchRelocationList.push_back(makePair(branch, to->labelNativeAddress()));
		} else {
			// Backwards jump
#ifdef NANOJIT_IA32
			debug_only( _fpuStkDepth = (_allocator.getActive(FST0) ? -1 : 0); )
#endif
			verbose_only(
				if (_logc->getLcBits() & LC_Native) {
					setOutputForEOL(" <= backward jump to %s", _printer->lookupName(to)); 
				} )
			
			NIns* branch = nRelativeBranch(0);
			_relativeBranchPatchList.push_back(makePair(branch, to));
		}
	}

	void Assembler::asm_jcc(LIns* ins) {
		bool branchOnFalse = (ins->opcode() == LIR_jf);
		LIns* cond = ins->branchCondition();
		if (cond->isImmI()) {
			if ((!branchOnFalse && !cond->immI()) || (branchOnFalse && cond->immI())) {
				// jmp never taken, not needed
			} else {
				asm_jmp(ins); // jmp always taken
			}
			return;
		}

		countlir_jcc();
		LIns* to = ins->branchTarget();
		NanoAssert(to->isLInsLabel());

		setResultLive(cond);
		setResultLive(cond->oprnd1());
		setResultLive(cond->oprnd2());

		const NIns* destinationAddress = to->labelNativeAddress();

		verbose_only(
			if ((destinationAddress == 0) && (_logc->getLcBits() & LC_Native)) {
				setOutputForEOL(" <= backward jump to %s", _printer->lookupName(to));
			} )

		Branches branches;
		nRelativeBranchCond(branchOnFalse, cond, destinationAddress, true, &branches);

		for (size_t index = 0; index < branches.getCount(); ++index) {
			NIns* branch = branches.getBranch(index);
			const NIns* dest = branches.getDestination(index);

			if (dest == 0) {
				_relativeBranchPatchList.push_back(makePair(branch, to));
			} else {
				_relativeBranchRelocationList.push_back(makePair(branch, dest));
			}
		}
	}

	void Assembler::asm_entry(LIns* ins) {
		evictAllActiveRegs();

		NIns* entry = nEntryPoint(ins->entryGetSaveMask());
		_entryPointList.push_back(entry);
		ins->entrySetAddress(entry);

		debug_only(_activation.checkCustomArAllocations();)
	}

	void Assembler::asm_exit(LIns* ins) {
		countlir_x();
		bool isPatchable = ins->exitIsPatchable();
		NIns* exitPointPatchAddress = nExitPoint(ins->exitGetRestoreMask(), isPatchable);

		if (isPatchable)
			_exitPointPatchList.push_back(exitPointPatchAddress);

		reconcileExitRegs(ins->exitParamData(), ins->exitParamCount());
	}

	// helper function for nop insertion feature that results in no more
	// than 1 no-op instruction insertion every 128-1151 Bytes
	static inline uint32_t noiseForNopInsertion(Noise* n) {
		return n->getValue(1023) + 128;
	}

	void Assembler::gen(LirFilter* reader) {
		NanoAssert(!hasError());

		// compiler hardening setup
		NIns* priorIns = _nIns;
		int32_t nopInsertTrigger = hardenNopInsertion(_config) ? noiseForNopInsertion(_noise): 0;

		// What's going on here: we're visiting all the LIR instructions in
		// the buffer, working strictly backwards in buffer-order, and
		// generating machine instructions for them as we go.
		//
		// For each LIns, we first check if it's live.  If so we mark its
		// operands as also live, and then generate code for it *if
		// necessary*.  It may not be necessary if the instruction is an
		// expression and code has already been generated for all its uses in
		// combination with previously handled instructions (ins->isExtant()
		// will return false if this is so).

		// Note that the backwards code traversal can make register allocation
		// confusing.  (For example, we restore a value before we spill it!)
		// In particular, words like "before" and "after" must be used very
		// carefully -- their meaning at regalloc-time is opposite to their
		// meaning at run-time.  We use the term "pre-regstate" to refer to
		// the register allocation state that occurs prior to an instruction's
		// execution, and "post-regstate" to refer to the state that occurs
		// after an instruction's execution, e.g.:
		//
		//   pre-regstate:  ebx(ins)
		//   instruction:   mov eax, ebx     // mov dst, src
		//   post-regstate: eax(ins)
		//
		// At run-time, the instruction updates the pre-regstate into the
		// post-regstate (and these states are the real machine's regstates).
		// But when allocating registers, because we go backwards, the
		// pre-regstate is constructed from the post-regstate (and these
		// regstates are those stored in RegAlloc).
		//
		// One consequence of generating code backwards is that we tend to
		// both spill and restore registers as early (at run-time) as
		// possible;  this is good for tolerating memory latency.  If we
		// generated code forwards, we would expect to both spill and restore
		// registers as late (at run-time) as possible;  this might be better
		// for reducing register pressure.

		debug_only(unsigned int lirIndex = 0;)

		// The trace must end with one of these opcodes.  Mark it as live.
		for (currIns = reader->read(); !currIns->isop(LIR_start); currIns = reader->read()) {

			LIns* ins = currIns;        // give it a shorter name for local use

			debug_only(lirIndex++;)

			if (!ins->isLive()) {
				// in case of unconditional jumps, the value might have an 
				// ar slot assigned. 
				NanoAssert(!ins->isInReg());
				freeResourcesOf(ins); 
				continue;
			}

			verbose_only(const char* insName = _printer->lookupName(ins); NanoUnusedVariable(insName));

#if defined(NANOJIT_VERBOSE)
			// Output the post-regstate (registers and/or activation).
			// Because asm output comes in reverse order, doing it now means
			// it is printed after the LIR and native code, exactly when the
			// post-regstate should be shown.
			if ((_logc->getLcBits() & LC_Native) && (_logc->getLcBits() & LC_Activation))
				printActivationState();
			if ((_logc->getLcBits() & LC_Native) && (_logc->getLcBits() & LC_RegAlloc))
				printRegState();
#endif

			// compiler hardening technique that inserts no-op instructions in the compiled method when nopInsertTrigger < 0
			if (hardenNopInsertion(_config)) {
				size_t delta = (uintptr_t)priorIns - (uintptr_t)_nIns; // # bytes that have been emitted since last go-around

				// if no codeList then we know priorIns and _nIns are on same page, otherwise make sure priorIns was not in the previous code block
				if (!currentCodeBlock->isInBlock(priorIns)) {
					NanoAssert(delta < _codeAlloc->getChunkSizeInBytes()); // sanity check
					nopInsertTrigger -= (int32_t) delta;
					if (nopInsertTrigger < 0)
					{
						nopInsertTrigger = noiseForNopInsertion(_noise);
						nInsertRandomNop();
					}
				}
				priorIns = _nIns;
			}

			LOpcode op = ins->opcode();
			switch (op) {

				default:
 					NanoAssertMsgf(false, "unsupported LIR instruction: %d\n", op);
					break;

				case LIR_livei:
				CASE64(LIR_liveq:) {
					countlir_live();
					LIns* op1 = ins->oprnd1();
					setResultLive(op1);
					if (!op1->isExtant())
						spawnValue(op1, GpRegs);
					break;
				}

				
				case LIR_lived:
				case LIR_livef: {
					countlir_live();
					LIns* op1 = ins->oprnd1();
					setResultLive(op1);
					if (!op1->isExtant())
						spawnValue(op1, FpRegs);
					break;
				}

				case LIR_immi:
					countlir_imm();
					if (ins->isExtant()) {
						nImmI(ins);
					} else {
						ins->setResultDead();
					}
					break;

#ifdef NANOJIT_64BIT
				case LIR_immq:
					countlir_imm();
					if (ins->isExtant()) {
						nImmQ(ins);
					} else {
						ins->setResultDead();
					}
					break;
#endif
				case LIR_immd:
					countlir_imm();
					if (ins->isExtant()) {
						nImmD(ins);
					} else {
						ins->setResultDead();
					}
					break;

			   case LIR_immf:
					countlir_imm();
					if (ins->isExtant()) {
						nImmF(ins);
					} else {
						ins->setResultDead();
					}
					break;

				case LIR_parami:
				CASE64(LIR_paramq:)
				case LIR_paramf:
				case LIR_paramd:
					countlir_param();
					NanoAssert(!ins->getGenFlags());
					if (ins->isExtant()) {
						asm_param(ins);
					} else {
						ins->setResultDead();
					}
					break;

				case LIR_regfence:
					asm_registerFence(ins);
					break;

				case LIR_regfenceall:
					asm_registerFenceAll(ins);
					break;

				case LIR_spillfence:
					asm_spillFence(ins);
					break;

				case LIR_spillfenceall:
					asm_spillFenceAll(ins);
					break;

				case LIR_killfence:
					asm_killFence(ins);
					break;

				case LIR_killfenceall:
					asm_killFenceAll(ins);
					break;

				case LIR_cmovi:
				CASE64(LIR_cmovq:)
				case LIR_cmovd:
				case LIR_cmovf:
					countlir_cmov();
					setResultLive(ins->oprnd1());
					setResultLive(ins->oprnd2());
					setResultLive(ins->oprnd3());
					if (ins->isExtant()) {
						nCmov(ins);
					} else {
						ins->setResultDead();
					}
					break;

				case LIR_lduc2ui:
				case LIR_ldus2ui:
				case LIR_ldrus2ui:
				case LIR_ldc2i:
				case LIR_lds2i:
				case LIR_ldrs2i:
				case LIR_ldi:
				case LIR_ldri:
					countlir_ld();
					setResultLive(ins->oprnd1());
					if (ins->isExtant()) {
						nLoad32(ins);
					} else {
						ins->setResultDead();
					}
					break;

				CASE64(LIR_ldq:)
				CASE64(LIR_ldrq:)
				case LIR_ldd:
				case LIR_ldf2d:
				case LIR_ldf: // Ok, ldf is not really 64-bits, but it's still more natural to
							   // handle it in load64 than in load32 (which basically deals with ints)
					countlir_ldq();
					setResultLive(ins->oprnd1());
					if (ins->isExtant()) {
						nLoad64(ins);
					} else {
						ins->setResultDead();
					}
					break;

				case LIR_negi:
				case LIR_noti:
				CASE64(LIR_negq:)
				CASE64(LIR_notq:)
					countlir_alu();
					setResultLive(ins->oprnd1());
					if (ins->isExtant()) {
						nNegNot(ins);
					} else {
						ins->setResultDead();
					}
					break;

				case LIR_addi:
				case LIR_addci:
				case LIR_subi:
				case LIR_subbi:
				case LIR_muli:
				case LIR_andi:
				case LIR_andci:
				case LIR_ori:
				case LIR_xori:
				case LIR_lshi:
				case LIR_rshi:
				case LIR_rshui:
				case LIR_rori:
				case LIR_rcri:
				CASE86(LIR_divis:)
				CASE86(LIR_diviu:)
				CASE64(LIR_addq:)
				CASE64(LIR_addcq:)
				CASE64(LIR_subq:)
				CASE64(LIR_subbq:)
				CASE64(LIR_mulq:)
				CASE64(LIR_andq:)
				CASE64(LIR_andcq:)
				CASE64(LIR_lshq:)
				CASE64(LIR_rshuq:)
				CASE64(LIR_rshq:)
				CASE64(LIR_rorq:)
				CASE64(LIR_rcrq:)
				CASE64(LIR_orq:)
				CASE64(LIR_xorq:)
					countlir_alu();
					setResultLive(ins->oprnd1());
					setResultLive(ins->oprnd2());
					if (ins->isExtant()) {
						nArith(ins);
					} else {
						ins->setResultDead();
					}
					break;

#if defined(NANOJIT_HOST_PLATFORM_X64)
				CASE86(LIR_modi:)
					countlir_alu();
					setResultLive(ins->oprnd1());
					if (ins->isExtant()) {
						nArith(ins);
					} else {
						ins->setResultDead();
					}
					break;
#endif

				case LIR_mullui:
				case LIR_mullsi:
				CASE64(LIR_mullsq:)
				CASE64(LIR_mulluq:)
					countlir_alu();
					setResultLive(ins->oprnd1());
					setResultLive(ins->oprnd2());

					if (ins->isExtant()) {
						nMull(ins);
					} else {
						ins->setResultDead();
					}
					break;

				case LIR_mulluhii:
				case LIR_mullshii:
				CASE64(LIR_mullshiq:)
				CASE64(LIR_mulluhiq:)
					setResultLive(ins->oprnd1());

					if (ins->isExtant()) {
						nMullhi(ins);
					} else {
						ins->setResultDead();
					}
					break;

				case LIR_negf:
				case LIR_negd:
				case LIR_absf:
				case LIR_absd:
					countlir_fpu();
					setResultLive(ins->oprnd1());
					if (ins->isExtant()) {
						nNegAbs(ins);
					} else {
						ins->setResultDead();
					}
					break;

				case LIR_recipf:
				case LIR_recipd:
				case LIR_rsqrtf:
				case LIR_rsqrtd:
				case LIR_sqrtf:
				case LIR_sqrtd:
					countlir_fpu();
					setResultLive(ins->oprnd1());
					if (ins->isExtant()) {
						nRecipSqrt(ins);
					} else {
						ins->setResultDead();
					}
					break;

				case LIR_addf:
				case LIR_subf:
				case LIR_mulf:
				case LIR_divf:
				case LIR_addd:
				case LIR_subd:
				case LIR_muld:
				case LIR_divd:
					countlir_fpu();
					setResultLive(ins->oprnd1());
					setResultLive(ins->oprnd2());
					if (ins->isExtant()) {
						nFop(ins);
					} else {
						ins->setResultDead();
					}
					break;

				case LIR_cvtif:
				case LIR_cvtuif:
				case LIR_cvtitf:
				case LIR_cvtuitf:
				case LIR_cvfif:
				case LIR_cvfuif:
				case LIR_cvtid:
				case LIR_cvtuid:
				case LIR_cvtitd:
				case LIR_cvtuitd:
				case LIR_cvfid:
				case LIR_cvfuid:
				CASE64(LIR_cvtqd:)
				CASE64(LIR_cvtuqd:)
				CASE64(LIR_cvtqtd:)
				CASE64(LIR_cvtuqtd:)
				CASE64(LIR_cvfqd:)
				CASE64(LIR_cvfuqd:)
				case LIR_cvffd:
				case LIR_cvtfd:
					countlir_fpu();
					setResultLive(ins->oprnd1());
					if (ins->isExtant()) {
						nFconv(ins);
					} else {
						ins->setResultDead();
					}
					break;

				CASE64(LIR_movq2i:)
				case LIR_movf2i:
				case LIR_movd2i:
					countlir_cvt();
					setResultLive(ins->oprnd1());
					if (ins->isExtant()) {
						nMovI(ins);
					} else {
						ins->setResultDead();
					}
					break;

#if defined(NANOJIT_64BIT)
				case LIR_movsi2sq:
				case LIR_movui2uq:
				case LIR_movd2q:
					countlir_cvt();
					setResultLive(ins->oprnd1());
					if (ins->isExtant()) {
						nMovQ(ins);
					} else {
						ins->setResultDead();
					}
					break;
#endif

				case LIR_movi2f:
				case LIR_movd2f:
					countlir_cvt();
					setResultLive(ins->oprnd1());
					if (ins->isExtant()) {
						nMovF(ins);
					} else {
						ins->setResultDead();
					}
					break;

				case LIR_movi2d:
				CASE64(LIR_movq2d:)
				case LIR_movf2d:
					countlir_cvt();
					setResultLive(ins->oprnd1());
					if (ins->isExtant()) {
						nMovD(ins);
					} else {
						ins->setResultDead();
					}
					break;

				case LIR_sti2c:
				case LIR_sti2s:
				case LIR_stri2s:
				case LIR_sti:
				case LIR_stri:
					countlir_st();
					setResultLive(ins->oprnd1());
					setResultLive(ins->oprnd2());
					nStore32(op, ins->oprnd1(), ins->disp(), ins->oprnd2());
					break;

				CASE64(LIR_stq:)
				CASE64(LIR_strq:)
				case LIR_std:
				case LIR_stf:
				case LIR_std2f: {
					countlir_stq();
					setResultLive(ins->oprnd1());
					setResultLive(ins->oprnd2());
					LIns* value = ins->oprnd1();
					LIns* base = ins->oprnd2();
					int dr = ins->disp();
					nStore64(op, value, dr, base);
					break;
				}

				case LIR_j:
					asm_jmp(ins);
					break;

				case LIR_jt:
				case LIR_jf:
					setResultLive(ins->branchCondition());
					asm_jcc(ins);
					break;

				case LIR_label: {
					countlir_label();
					// see if the backend needs to be notified about this
					// labels: PPC tracks them as an instruction swap fence.
					nLabel();
					ins->labelSetNativeAddress(_nIns);

					verbose_only(
						RefBuf b;
						if (_logc->getLcBits() & LC_Native) {
							asm_output("[%s]", _printer->formatRef(&b, ins));
					})
					
					break;
				}


				case LIR_phicapture:
					asm_phiCapture(ins);
					break;

				case LIR_phireconcile:
					asm_phiReconcile(ins);
					break;

				case LIR_entry:
					asm_entry(ins);
					break;

				case LIR_exit:
					asm_exit(ins);
					break;

				case LIR_eqf:
				case LIR_nef:
				case LIR_lef:
				case LIR_ltf:
				case LIR_gtf:
				case LIR_gef:
				case LIR_unf:
				case LIR_eqd:
				case LIR_ned:
				case LIR_led:
				case LIR_ltd:
				case LIR_gtd:
				case LIR_ged:
				case LIR_und:
					countlir_fpu();
					break;

				case LIR_getflgz:
				case LIR_getflgnz:
					setResultLive(ins->oprnd1());
					if (ins->isExtant()) {
						ins->oprnd1()->setFlagRequested(ArithFlag_Zero);
						nSaveFlag(ins);
					} else {
						ins->setResultDead();
					}
					break;

				case LIR_getflgs:
				case LIR_getflgns:
					setResultLive(ins->oprnd1());
					if (ins->isExtant()) {
						ins->oprnd1()->setFlagRequested(ArithFlag_Sign);
						nSaveFlag(ins);
					} else {
						ins->setResultDead();
					}
					break;

				case LIR_getflgc:
				case LIR_getflgnc:
					setResultLive(ins->oprnd1());
					if (ins->isExtant()) {
						ins->oprnd1()->setFlagRequested(ArithFlag_Carry);
						nSaveFlag(ins);
					} else {
						ins->setResultDead();
					}
					break;

				case LIR_getflgb:
				case LIR_getflgnb:
					setResultLive(ins->oprnd1());
					if (ins->isExtant()) {
						ins->oprnd1()->setFlagRequested(ArithFlag_Borrow);
						nSaveFlag(ins);
					} else {
						ins->setResultDead();
					}
					break;

				case LIR_getflgv:
				case LIR_getflgnv:
					setResultLive(ins->oprnd1());
					if (ins->isExtant()) {
						ins->oprnd1()->setFlagRequested(ArithFlag_Overflow);
						nSaveFlag(ins);
					} else {
						ins->setResultDead();
					}
					break;

				case LIR_getflgfov:
					setResultLive(ins->oprnd1());
					if (ins->isExtant()) {
						ins->oprnd1()->setFlagRequested(ArithFlag_FloatOverflow);
						nSaveFlag(ins);
					} else {
						ins->setResultDead();
					}
					break;

				case LIR_getflgfuf:
					setResultLive(ins->oprnd1());
					if (ins->isExtant()) {
						ins->oprnd1()->setFlagRequested(ArithFlag_FloatUnderflow);
						nSaveFlag(ins);
					} else {
						ins->setResultDead();
					}
					break;

				case LIR_getflgfix:
					setResultLive(ins->oprnd1());
					if (ins->isExtant()) {
						ins->oprnd1()->setFlagRequested(ArithFlag_FloatInvalid);
						nSaveFlag(ins);
					} else {
						ins->setResultDead();
					}
					break;

				case LIR_setflgz:
				case LIR_setflgs:
				case LIR_setflgc:
				case LIR_setflgb:
				case LIR_setflgv:
					setResultLive(ins->oprnd1());
					nRestoreFlag(ins);
					break;

				case LIR_eqi:
				case LIR_nei:
				case LIR_lei:
				case LIR_lti:
				case LIR_gti:
				case LIR_gei:
				case LIR_ltui:
				case LIR_leui:
				case LIR_gtui:
				case LIR_geui:
				CASE64(LIR_eqq:)
				CASE64(LIR_neq:)
				CASE64(LIR_leq:)
				CASE64(LIR_ltq:)
				CASE64(LIR_gtq:)
				CASE64(LIR_geq:)
				CASE64(LIR_ltuq:)
				CASE64(LIR_leuq:)
				CASE64(LIR_gtuq:)
				CASE64(LIR_geuq:)
					countlir_alu();
					break;

				case LIR_callv:
				case LIR_calli:
				CASE64(LIR_callq:)
				case LIR_calld:
				case LIR_callf: {
					countlir_call();
					nCall(ins, reinterpret_cast<const NIns*>(ins->callAddress()));
					break;
				}

				case LIR_pc: {
					// record the current PC
					void **dest = (void **)ins->oprnd1()->immP();
					*dest = _nIns;
					break;
				}

				case LIR_nop:
				case LIR_comment:
					// Do nothing.
					break;

				case LIR_copyi: {
					setResultLive(ins->oprnd1());
					if (ins->isExtant()) {
						if (canReuseStorage(ins)) {
							reuseStorage(ins, GpRegs);
						} else {
							nCopyI(ins);
						}
					} else {
						ins->setResultDead();
					}
					break;
				}

#if defined NANOJIT_64BIT
				case LIR_copyq: {
					setResultLive(ins->oprnd1());

					if (ins->isExtant()) {
						if (canReuseStorage(ins)) {
							reuseStorage(ins, GpRegs);
						} else {
							nCopyQ(ins);
						}
					} else {
						ins->setResultDead();
					}
					break;
				}
#endif

				case LIR_copyf: {
					setResultLive(ins->oprnd1());
					if (ins->isExtant()) {
						if (canReuseStorage(ins)) {
							reuseStorage(ins, FpRegs);
						} else {
							nCopyF(ins);
						}
					} else {
						ins->setResultDead();
					}
					break;
				}

				case LIR_copyd: {
					setResultLive(ins->oprnd1());
					if (ins->isExtant()) {
						if (canReuseStorage(ins)) {
							reuseStorage(ins, FpRegs);
						} else {
							nCopyD(ins);
						}
					} else {
						ins->setResultDead();
					}
					break;
				}

				case LIR_getfp: {
					if (ins->isExtant()) {
						Register r = prepareResultReg(ins, GpRegs);
						nGpCopy(r, FP);
						freeResourcesOf(ins);
					} else {
						ins->setResultDead();
					}
					break;
				}
			}

#if defined(NANOJIT_VERBOSE)
			// We do final LIR printing inside this loop to avoid printing
			// dead LIR instructions.  We print the LIns after generating the
			// code.  This ensures that the LIns will appear in debug output
			// *before* the native code, because Assembler::outputf()
			// prints everything in reverse.
			//
			if (_logc->getLcBits() & LC_AfterDCE) {
				InsBuf b;
				if (ins->isop(LIR_comment))
					outputf("%s", _printer->formatIns(&b, ins));
				else
					outputf("  %s (in reg? %d)", _printer->formatIns(&b, ins), ins->isInReg() ? REGNUM(ins->getReg()) : 0xFFFFFFFF);
			}
#endif

			if (hasError())
				return;

			NanoAssert(ins->isV() || (!ins->isLive()));

			// check that all is well (don't check in exit paths since its more complicated)
			debug_only(pageValidate();)
			debug_only(resourceConsistencyCheck();)
			
		}
	}

	void Assembler::evictCallParamValues(ParameterValue* callParams, unsigned int paramCount, RegisterMask abiEvictMask, RegisterMask resultRegisterMask) {
		RegisterMask callEvictMask = abiEvictMask;


		for (unsigned int i = 0; i < paramCount; i++) {
			const ParameterValue*	param		= &callParams[i];
			LIns*					paramValue	= param->value;

			if (param->byRef && paramValue->isExtant())
			{
				setResultLive(paramValue);
				nRestore(paramValue, param->reg);
			}
		}

		evictSomeActiveRegs(callEvictMask & ~resultRegisterMask);
	}

	void Assembler::makeCallParamExtend(ParameterValue* paramData, unsigned int paramCount) {

		RegisterMask paramRegMask = 0;

		for (unsigned int i = 0; i < paramCount; i++) {

			const ParameterValue*	data		= &paramData[i];
			LIns*					paramValue	= data->value;
			Register				reg			= data->reg;
			int						stackOffset	= data->stackOffset;

			NanoAssert(paramValue != NULL);
			RegisterMask allow = paramValue->isIorQ() ? GpRegs : FpRegs;

			if (reg != UnspecifiedReg) {

				// register must have been spilled after the call
				NanoAssert(_allocator.isFree(reg));
				RegisterMask dstMask = rmask(reg);

				if (data->byRef) {

					paramValue->preventOptimize();
					setResultLive(paramValue);
					findMemFor(paramValue);
					findRegFor(paramValue, GpRegs & ~paramRegMask);
					nSpill(paramValue);

					int d = arDisp(paramValue->getArIndex());
					nLoadEffectiveAddress(reg, FP, d);

				} else if (paramValue->isInReg()) {

					if (paramValue->getReg() == reg)
						continue;

					// it means the value was already used for another parameter
					// copy it to the other register
					Register srcReg = paramValue->getReg();
					if ((rmask(srcReg) & GpRegs) && (dstMask & GpRegs)) {
						nGpCopy(reg, srcReg);
					} else {
						nNonGpCopy(reg, srcReg);
					}
				} else {
					setResultLive(paramValue);
					findRegFor(paramValue, dstMask);
					NanoAssert(paramValue->getReg() == reg);
				}

				paramRegMask |= dstMask;
			} else {
				setResultLive(paramValue);

				if (data->byRef) {
					paramValue->preventOptimize();

					Register tmpReg = allocTempReg(LTy_P, GpRegs & ~paramRegMask);
					findMemFor(paramValue);
					findRegFor(paramValue, GpRegs & ~(paramRegMask | rmask(tmpReg)));
					nSpill(paramValue);
					int d = arDisp(paramValue->getArIndex());
					nStoreToStack(paramValue->retType(), tmpReg, stackOffset);
					nLoadEffectiveAddress(tmpReg, FP, d);

				} else {
					Register r = findRegFor(paramValue, allow & (~paramRegMask));
					nStoreToStack(paramValue->retType(), r, stackOffset);
				}
			}
		}
	}

	void Assembler::reconcileExitRegs(ParameterValue* paramData, unsigned int paramCount) {

		RegisterMask paramRegMask = 0;

		for (unsigned int i = 0; i < paramCount; i++) {

			const ParameterValue*   data        = &paramData[i];
			LIns*                   paramValue  = data->value;
			Register                reg         = data->reg;

			NanoAssert(paramValue != NULL);

			if (reg != UnspecifiedReg) {

				RegisterMask dstMask = rmask(reg);

				if (paramValue->isInReg()) {
					Register srcReg = paramValue->getReg();

					if (srcReg == reg) {
						paramRegMask |= dstMask;
						continue;
					}

					RegisterMask srcRegMask = rmask(srcReg);
					RegisterMask allow = paramValue->isIorQ() ? GpRegs : FpRegs;

					if ((srcRegMask & paramRegMask) != 0) {
						if ((srcRegMask & allow) && (dstMask & allow)) {
							nGpCopy(reg, srcReg);    // move 'ins' from its pre-state reg (r) to its post-state reg (s)
						} else {
							nNonGpCopy(reg, srcReg);
						}

						continue;
					}

					paramValue->clearReg();
					_allocator.retire(srcReg);
				}

				NanoAssert(!paramValue->isInReg());

				// we can't call evict because there is no need to restore the value
				LIns* active = _allocator.getActive(reg);
				if (active != NULL) {
					findMemFor(active);
					active->clearReg();
					_allocator.retire(reg);
				}

				setResultLive(paramValue);
				findRegFor(paramValue, dstMask);

				paramRegMask |= dstMask;

			} else {
				setResultLive(paramValue);
				findMemFor(paramValue);
			}
		}
	}
	
	bool Assembler::canReuseStorage(LIns* insCopy) {
#if defined(NANOJIT_64BIT)
		NanoAssert(
				insCopy->isop(LIR_copyi)
			||	insCopy->isop(LIR_copyq)
			||	insCopy->isop(LIR_copyf)
			||	insCopy->isop(LIR_copyd));
#else
		NanoAssert(
			insCopy->isop(LIR_copyi)
			||	insCopy->isop(LIR_copyf)
			||	insCopy->isop(LIR_copyd));
#endif

		LIns* insToAssign = insCopy->oprnd1();

		if (insToAssign->isExtant())
			return false;

		if (insCopy->isInAr() && (insCopy->getArGroup() != insToAssign->getArGroup()))
			return false;

		if (insCopy->getGenFlags() || insToAssign->getGenFlags())
			return false;

		if (!insCopy->canOptimize() || !insToAssign->canOptimize())
			return false;

		return true;
	}

	void Assembler::reuseStorage(LIns* insCopy, RegisterMask allow) {

		LIns* insSrc = insCopy->oprnd1();

		NanoAssert(!(insCopy->getGenFlags() && insCopy->flagRequested()));
		NanoAssert(!insSrc->isExtant());

		if (RegAlloc::nCanRemat(insSrc)) {
			// rematerialize insSrc into insCopy register
			Register r = prepareResultReg(insCopy, allow);
			nRestore(insSrc, r);
			freeResourcesOf(insCopy);
		} else {
			asm_copy_slot(insCopy, insSrc);
		}
	}


#if defined(NANOJIT_VERBOSE)
	void Assembler::printRegState() {
		char* s = &outline[0];
		VMPI_memset(s, ' ', 26);  s[26] = '\0';
		s += VMPI_strlen(s);
		VMPI_sprintf(s, "RR");
		s += VMPI_strlen(s);

		RegisterMask active = _allocator.activeMask();
		for (Register r = lsReg(active); active != 0 ; r = nextLsReg(active, r)) {
			LIns *ins = _allocator.getActive(r);
			NanoAssertMsg(!_allocator.isFree(r),
						  "Coding error; register is both free and active! " );
			RefBuf b;
			const char* n = _printer->formatRef(&b, ins);

			VMPI_sprintf(s, " %s(%s)", gpn(r), n);
			s += VMPI_strlen(s);
		}
		output();
	}

	void Assembler::printActivationState() {
		char* s = &outline[0];
		VMPI_memset(s, ' ', 26);  s[26] = '\0';
		s += VMPI_strlen(s);
		VMPI_sprintf(s, "AR");
		s += VMPI_strlen(s);

		LIns* ins = 0;
		uint32_t nStackSlots = 0;
		int32_t arIndex = 0;
		for (AR::Iter iter(_activation); iter.next(ins, nStackSlots, arIndex); ) {
			RefBuf b;
			const char* n = _printer->formatRef(&b, ins);
			if (nStackSlots > 1) {
				VMPI_sprintf(s," %d-%d(%s)", 4*arIndex, 4*(arIndex+nStackSlots-1), n);
			} else {
				VMPI_sprintf(s," %d(%s)", 4*arIndex, n);
			}
			s += VMPI_strlen(s);
		}
		output();
	}
#endif

	uint32_t Assembler::arReserve(LIns* ins) {
		return _activation.reserveEntry(ins);
	}

	void Assembler::arRelease(LIns* ins) {
		NanoAssert(ins->isInAr());
		uint32_t arIndex = ins->getArIndex();
#if defined(NANOJIT_DEBUG)
		NanoAssert(_activation.isValidEntry(arIndex, ins));
#endif
		_activation.releaseEntryAt(arIndex); // mark entry as free, but do not allow realloaction
	}

	// Generate code to restore any registers in 'regs' that are currently active,
	void Assembler::evictSomeActiveRegs(RegisterMask regs) {
		RegisterMask evict_set = regs & _allocator.activeMask();
		for (Register r = lsReg(evict_set); evict_set; r = nextLsReg(evict_set, r)) {
			LIns* ins = _allocator.getActive(r);
			Register r1 = ins->getReg();
			NanoAssert((rmask(r) & rmask(r1)) == rmask(r));
			r = r1;
			evict(ins);
		}
	}

#if defined(NANOJIT_VERBOSE)
	char Assembler::outline[8192];
	char Assembler::outlineEOL[512];

	void Assembler::output() {
		// The +1 is for the terminating NUL char.
		VMPI_strncat(outline, outlineEOL, sizeof(outline)-(strlen(outline)+1));

		if (_outputCache) {
			//_logc->printf("%s\n", outline);

			char* str = new (_outputCache->getAllocator()) char[VMPI_strlen(outline)+1];
			VMPI_strcpy(str, outline);
			_outputCache->insert(str);
		} else {
			_logc->printf("%s\n", outline);
		}

		outline[0] = '\0';
		outlineEOL[0] = '\0';
	}

	void Assembler::outputf(const char* format, ...) {
		va_list args;
		va_start(args, format);

		outline[0] = '\0';
		vsprintf(outline, format, args);
		output();
	}

	void Assembler::setOutputForEOL(const char* format, ...) {
		va_list args;
		va_start(args, format);

		outlineEOL[0] = '\0';
		vsprintf(outlineEOL, format, args);
	}

	void Assembler::maybe_disassemble() {
	}
#endif // NANOJIT_VERBOSE


	void Assembler::asm_param(LIns *ins) {
		Register a = ins->paramReg();

		if (a == UnspecifiedReg) {
			if (!ins->isInAr()) {
				NanoAssert(ins->getArGroup() != LINS_DEFAULT_AR_GROUP);
				reallocateArEntry(ins, ins->getArGroup());
			}

			if (ins->isInReg()) {
				nRestore(ins, ins->getReg());
			}

			freeResourcesOf(ins);
		} else {
			// incoming arg in register
			prepareResultReg(ins, rmask(a));
			freeResourcesOf(ins);
		}
	}

	void Assembler::spawnValue(LIns* ins, RegisterMask allow) {
		NanoAssert(!ins->isExtant());

		Register r = _allocator.allocReg(ins, allow);
		if (r == UnspecifiedReg) {
			// no more register available, spawn value in ar
			findMemFor(ins);
		}
	}

	void Assembler::setReconcileDefaultRegAlloc(LIns* reconcileIns, PhiState* phiStateArray, size_t phiStateCount) {
		LIns** valueIns = reconcileIns->phiValues();
		PhiState* phiState = phiStateArray;

		for (size_t i = 0; i < phiStateCount; ++i) {
			if (phiState->isLive && phiState->inReg) {
				Register r = { phiState->regnum };
				(*valueIns)->setCaptureReg(r);
			}
			++phiState;
			++valueIns;
		}
	}

	void Assembler::asm_phiCapture(LIns* ins) {
		NanoAssert(ins->isLInsPhi());

		RegisterMask        usedRegisterMask    = 0;

		PhiData*            phiData             = ins->phiData();
		size_t              phiValueCount       = ins->phiCount();
		LIns**              phiValue            = ins->phiValues();
		PhiState*           phiStates           = (PhiState*)_internalAlloc->alloc(phiValueCount * sizeof(PhiState));


		phiData->capturedStates = phiStates;

		for (unsigned int i = 0; i < phiValueCount; ++i) {
			PhiState*       state       = &phiStates[i];
			LIns*           value       = phiValue[i];
			bool            isLive      = value->isLive();


			if (!isLive) {
				state->isLive = false;
				continue;
			}

			if (!value->isExtant()) {
				// the value was not used so far, assign a register
				RegisterMask allow = 0;
				switch (value->retType()) {
					case LTy_I:
					CASE64(LTy_Q:)
						allow = GpRegs;
						break;

					case LTy_F:
					case LTy_D:
						allow = FpRegs;
						break;

					default:
						NanoAssert(0); // invalid type
				}

				spawnValue(value, allow & ~usedRegisterMask);
			}

			state->isLive = true;


			if (value->isInReg()) {
				Register reg = value->getReg();

				state->inReg    = true;
				state->regnum   = static_cast<uint8_t>(REGNUM(reg));

				usedRegisterMask |= rmask(reg);
			} else {
				state->inReg    = false;
				state->regnum   = static_cast<uint8_t>(REGNUM(UnspecifiedReg));
			}

			if (value->isInAr()) {
				state->inAr     = true;
				state->arIndex  = static_cast<uint16_t>(value->getArIndex());
			} else {
				state->inAr     = false;
				state->arIndex  = 0;
			}

#if defined(NANOJIT_DEBUG)
			state->arGroupIndex = value->getArGroup();
			state->otherValue = value;
#endif
		}

		LIns* reconcileIns = ins->phiReconcileNodeNext();
		while (reconcileIns != NULL){
			NanoAssert(reconcileIns->phiCount() == phiValueCount);
			setReconcileDefaultRegAlloc(reconcileIns, phiStates, phiValueCount);
			reconcileIns = reconcileIns->phiReconcileNodeNext();
		}

		// we only need to consider values in registers
		// values in ar will *always* be stored in ar and do not need to be modified
		RegisterMask activeMask = _allocator.activeMask() & ~usedRegisterMask;
		phiData->activeMask = activeMask;

#if defined(NANOJIT_DEBUG)
		memset(phiData->active, 0, sizeof(phiData->active));
#endif

		unsigned int activeCount = 0;

		while (activeMask != 0) {
			// note: in case of overlapping registers, activeIns might be different activeRegFromMask
			Register activeRegFromMask = lsReg(activeMask);
			LIns* activeIns = _allocator.getActive(activeRegFromMask);
			Register activeReg = activeIns->getReg();

			NanoAssert(activeCount < (LastRegNum + 1));
			phiData->active[activeCount].ins = activeIns;
			phiData->active[activeCount].reg = activeReg;

			activeMask &= ~rmask(activeReg);
			activeCount++;
		}

		phiData->activeCount = activeCount;

	}

	void Assembler::asm_phiReconcileForwards(LIns* ins) {
		NanoAssert(ins->isLInsPhi());

		size_t                  phiValueCount       = ins->phiCount();
		const PhiData*          phiData             = ins->phiData();
		const PhiState*         phiReconcileData    = phiData->capturedStates;
		LIns**                  phiValue            = ins->phiValues();
		RegisterMask            phiValueMask        = 0;

		for (unsigned int i = 0; i < phiValueCount; ++i) {

			const PhiState*     state       = &phiReconcileData[i];
			LIns*               value       = phiValue[i];

			// if the corresponding value was not alive during capture, no need to reconcile it
			if (!state->isLive)
				continue; 

#if defined(NANOJIT_DEBUG)
			NanoAssert((state->otherValue == value) || (!state->otherValue->isLive()));
			NanoAssert(state->arGroupIndex == value->getArGroup());
#endif
			// if the value was live at the bottom of the loop, force it to live at the top as well
			if (!value->isLive()) {
				value->setResultLive();
			}

			if (state->inReg) {
				Register reg = { state->regnum };
				RegisterMask regmask = rmask(reg);

				verbose_only(
					if (value->isInReg() && (rmask(value->getReg()) & regmask) == 0) {
						compilationStats.reconcileForwardMoveToRegCount++;
					}
				)

				findRegFor(value, regmask);
				phiValueMask |= regmask;
			}

			if (state->inAr) {
				findMemFor(value);
				NanoAssert(value->getArIndex() == state->arIndex);
			}
		}

		NanoAssert((phiValueMask & phiData->activeMask) == 0);

		RegisterMask activeMask = phiData->activeMask;
		unsigned int activeCount = phiData->activeCount;
		NanoAssert(activeCount < (LastRegNum + 1));

		for (unsigned int activeIndex = 0; activeIndex < activeCount; ++activeIndex) {

			LIns* previousActiveIns = phiData->active[activeIndex].ins;
			Register previousActiveReg = phiData->active[activeIndex].reg;

			RegisterMask previousActiveRegMask = rmask(previousActiveReg);
			NanoAssert((previousActiveRegMask & activeMask) == previousActiveRegMask);
			activeMask &= ~previousActiveRegMask;

			LIns* active = _allocator.getActive(previousActiveReg);

			if (!previousActiveIns->isLive())
				setResultLive(previousActiveIns);

			if (previousActiveIns != active) {
				findRegFor(previousActiveIns, previousActiveRegMask);
			}
		}

		NanoAssert(activeMask == 0);
	}


	void Assembler::asm_reconcileStateBackward(LIns* value, bool inReg, RegisterMask regmask, bool inAr, unsigned int arIndex) {

		NanoUnusedVariable(arIndex);

		if (inReg) {
			verbose_only(
				if (value->isInReg() && (rmask(value->getReg()) & regmask) == 0) {
					compilationStats.reconcileBackwardMoveToRegCount++;
				}
			)

			findRegFor(value, regmask);

			if (inAr) {
				if (!value->isInAr())
					findMemFor(value);

				NanoAssert(value->getArIndex() == arIndex);
			} else if (value->isInAr() /*&& !state->inAr*/) {

				// the bottom value was not spilled, but the top value top was restored from spill buffer
				// a spill is forced to write the value computed at the bottom 
				verbose_only(
					RefBuf b;
					if (_logc->getLcBits()  & LC_Native) {
						setOutputForEOL(" <= phi force spill %s", _printer->formatRef(&b, value));
					}
					compilationStats.reconcileForceSpillCount++;
				)

				nSpill(value);
			}
		} else {
			NanoAssert(inAr);

			if (!value->isInAr())
				findMemFor(value);

			NanoAssert(value->getArIndex() == arIndex);

			if (value->isInReg()) {
				// bottom value was spilled and was no longer in a register when crossing the bottom edge
				// force a restore to read the bottom value from the spill buffer into the current register
				verbose_only(
					RefBuf b;
					if (_logc->getLcBits() & LC_Native) {
						setOutputForEOL(" <= phi force restore %s", _printer->formatRef(&b, value));
					}
					compilationStats.reconcileForceRestoreCount++;
				)

				nRestore(value, value->getReg());

				// todo: check if this a good idea to release the register or not
				_allocator.retire(value->getReg());
				value->clearReg();
			}
		}
	}

	void Assembler::asm_phiReconcileBackwards(LIns* ins) {
		NanoAssert(ins->isLInsPhi());

		const PhiData*          phiData             = ins->phiData();
		size_t                  phiValueCount       = ins->phiCount();
		const PhiState*         phiStates           = phiData->capturedStates;
		LIns**                  phiValues           = ins->phiValues();
		RegisterMask            phiValueMask        = 0;

		for (unsigned int i = 0; i < phiValueCount; ++i) {
			const PhiState*     state           = &phiStates[i];
			LIns*               value           = phiValues[i];

			// if the corresponding value was not alive during capture, no need to reconcile it
			if (!state->isLive)
				continue;

#if defined(NANOJIT_DEBUG)
			NanoAssert((state->otherValue == value) || (!state->otherValue->isLive()));
			NanoAssert(state->arGroupIndex == value->getArGroup());
#endif

			if (!value->isLive()) {
				// value is not live, don't need to reconciliate it
				// it should not have been added to the phi value list.
				continue;
			}

			Register reg = { state->regnum };
			RegisterMask regmask = rmask(reg);
			if (state->inReg)
				phiValueMask |= regmask;

			asm_reconcileStateBackward(value, state->inReg, regmask, state->inAr, state->arIndex);
		}

		RegisterMask activeMask = phiData->activeMask;
		NanoAssert((phiValueMask & phiData->activeMask) == 0);

		RegisterMask evictMask = ~(activeMask | phiValueMask);

		unsigned int activeCount = phiData->activeCount;
		NanoAssert(activeCount < (LastRegNum + 1));

		for (unsigned int activeIndex = 0; activeIndex < activeCount; ++activeIndex) {

			Register previousActiveReg = phiData->active[activeIndex].reg;
			RegisterMask previousActiveRegMask = rmask(previousActiveReg);

			NanoAssert((previousActiveRegMask & phiValueMask) == 0);

			LIns* previousActive = phiData->active[activeIndex].ins;
			// note : previousActive might not be alive anymore (as opposed to forward jumps)
			// it is valid to let values cross the bottom edge while not alive at the top.

			LIns* active = _allocator.getActive(previousActiveReg);


			if (previousActive != active) {
				if (active != NULL)
					evict(active);

				if (previousActive->isLive())
					findRegFor(previousActive, previousActiveRegMask);
			}

			activeMask &= ~previousActiveRegMask;
		}
		NanoAssert(activeMask == 0);

		evictSomeActiveRegs(evictMask);
	}

	void Assembler::asm_phiReconcile(LIns* ins) {
		NanoAssert(ins->isLInsPhi());

		if (ins->phiData()->forward)
			asm_phiReconcileForwards(ins);
		else
			asm_phiReconcileBackwards(ins);
	}

	void Assembler::asm_spillFence(LIns* ins) {
		LIns** valueArray = ins->spillFenceArray();
		size_t valueCount = ins->spillFenceCount();
		for (size_t i = 0; i < valueCount; ++i) {
			LIns* value = valueArray[i];
			NanoAssert(value->getArGroup() != LINS_DEFAULT_AR_GROUP);
			setResultLive(value);
			findMemFor(value);
		}
	}

	void Assembler::asm_spillFenceAll(LIns* ins) {
		NanoUnusedVariable(ins);
		NanoAssert(ins->isop(LIR_spillfenceall));
		RegisterMask activeMask = _allocator.activeMask();
		for (Register r = lsReg(activeMask); activeMask; r = nextLsReg(activeMask, r)) {
			LIns* value = _allocator.getActive(r);
			setResultLive(value);
			findMemFor(value);
		}
	}

	void Assembler::asm_registerFence(LIns* ins) {
		LIns** valueArray = ins->registerFenceArray();
		size_t valueCount = ins->registerFenceCount();
		for (size_t i = 0; i < valueCount; ++i) {
			LIns* value = valueArray[i];
			NanoAssert(value->getArGroup() != LINS_DEFAULT_AR_GROUP);
			if (value->isInReg()) {
				_allocator.retire(value->getReg());
				value->clearReg();
			}
		}
	}

	void Assembler::asm_registerFenceAll(LIns* ins) {
		NanoUnusedVariable(ins);
		NanoAssert(ins->isop(LIR_regfenceall));

		RegisterMask activeMask = _allocator.activeMask();
		if (activeMask == 0)
			return;

		for (Register r = lsReg(activeMask); activeMask; r = nextLsReg(activeMask, r)) {
			LIns* value = _allocator.getActive(r);
			_allocator.retire(value->getReg());
			value->clearReg();
		}
	}

	void Assembler::asm_killFence(LIns* ins) {
		LIns** valueArray = ins->killFenceArray();
		size_t valueCount = ins->killFenceCount();
		for (size_t i = 0; i < valueCount; ++i) {
			LIns* value = valueArray[i];
			_allocator.retire(value->getReg());
			value->clearReg();
			value->setResultDead();
		}
	}

	void Assembler::asm_killFenceAll(LIns* ins) {
		NanoUnusedVariable(ins);
		NanoAssert(ins->isop(LIR_killfenceall));

		RegisterMask activeMask = _allocator.activeMask();
		if (activeMask == 0)
			return;

		for (Register r = lsReg(activeMask); activeMask; r = nextLsReg(activeMask, r)) {
			LIns* value = _allocator.getActive(r);
			NanoAssert(value != NULL);
			// note: in case of overlapping registers valueReg might be different from r
			Register valueReg = value->getReg(); 
			_allocator.retire(valueReg);
			value->clearReg();
			value->setResultDead();
#if defined(NANOJIT_RA_REGISTERS_OVERLAP)
			// retire() will remove the reference in all overlapping registers (if any)
			RegisterMask regMask = rmask(valueReg);
			activeMask &= ~regMask; 
#endif
		}
	}

	void Assembler::asm_copy_slot(LIns* ins, LIns* insToAssign) {
		NanoAssert(ins->isExtant());
		NanoAssert(!insToAssign->isExtant());

		bool            inAr        = ins->isInAr();
		uint32_t        arSlot      = inAr ? ins->getArIndex() : 0;
		bool            inReg       = ins->isInReg();
		Register        reg         = inReg ? ins->getReg() : UnspecifiedReg;

		freeResourcesOf(ins);

		if (inAr) {
			reallocateArEntry(insToAssign, arSlot);
		}

		if (inReg) {
			if (!_allocator.isFree(reg)) {
				LIns* vic = _allocator.getActive(reg);
				NanoAssert(vic->isInReg());
				NanoAssert(vic->getReg() == reg);
				evict(vic);
			}

			_allocator.allocSpecificReg(insToAssign, reg);
		}

		NanoAssert(insToAssign->isExtant());
	}


	unsigned int Assembler::computeInitialReservedNInsCount(NIns* curIns, CodeBlock* codeBlock) {
		ptrdiff_t remainingCount = curIns - codeBlock->internalBufferStart();
		const unsigned int thresholdLimit = _nCodeBlockSizeThresholdInBytes / sizeof(NIns);

		NanoAssert(remainingCount > 0);
		if ((thresholdLimit > 0) && (static_cast<unsigned int>(remainingCount) >= thresholdLimit)) {
			return static_cast<unsigned int>(remainingCount - thresholdLimit);
		} else {
			return 0;
		}
	}

	void Assembler::nativePageSetup() {
		NanoAssert(_nIns == NULL);

		codeAlloc();
		_nLastIns = _nIns;
		_nReservedNInsCount = computeInitialReservedNInsCount(_nIns, currentCodeBlock);
	}

	void Assembler::nativePageReset() {
	}

	void Assembler::nativePageSplit() {
		NIns *pc = _nIns;

		verbose_only(if (_logc->getLcBits() & LC_Native) outputf("newpage %p:", pc);)

		nNotifyPageJump();

		bool nativeDataAdded = (pc != _nIns);

		bool codeAllocated = false;
		NIns *top = currentCodeBlock->internalBufferStart();
		if (((_nIns - top) * sizeof(NIns)) < CodeAlloc::minAllocSize) {
			codeAlloc();
			codeAllocated = true;
		}

		_nReservedNInsCount = computeInitialReservedNInsCount(_nIns, currentCodeBlock);

		if (codeAllocated || nativeDataAdded)
		{
			NIns* afterBranch = _nIns;

			nRelativeBranch(pc); // cannot create a new page

			if (nativeDataAdded) {
				_relativeBranchRelocationList.push_back(makePair<NIns*, const NIns*>(_nIns, pc));
			}

			if (codeAllocated)
			{
				if (nativeDataAdded)
					_nLastIns = afterBranch;  // include jump in block 
				else
					_nLastIns = _nIns; // jump is not need when block will flattened
			}
		}
	}

	// This should only be called from within emit() et al.
	void Assembler::underrunProtect(size_t count) {
		NanoAssertMsg(count >= NANOJIT_SMALLEST_UNDERRUN_PROT, "underrunProtect called with a bytes too small for the platform");
		NIns *pc = _nIns;
		NIns *top = currentCodeBlock->internalBufferStart() + _nReservedNInsCount;

		if ((pc - count) < top) {
			nativePageSplit();
		}
	}

	bool Assembler::reserveNInsSlotAtCodeStart(unsigned int count) {
		NIns *pc = _nIns;
		NIns *top = currentCodeBlock->internalBufferStart() + _nReservedNInsCount;

		NanoAssert(pc > top);

		if (static_cast<unsigned int>(pc - top) > count) {
			_nReservedNInsCount += count;
			return true;
		}

		return false;
	}

	void Assembler::freeReservedNInsSlot(unsigned int count) {
		NanoAssert(_nReservedNInsCount >= count);
		_nReservedNInsCount -= count;
	}

	bool Assembler::flattenCodeBlockList(void* buffer, size_t bufferSize) const {

		NanoAssert(currentCodeBlock == NULL); // must not be compiling

		const CodeBlock* block = _codeBlockList.head();
		uint8_t* const outputBuffer = reinterpret_cast<uint8_t*>(buffer);
		uint8_t* writeBuffer = outputBuffer;

		while (block != NULL) {

			size_t codeSize = block->codeSize();
			if (((writeBuffer - outputBuffer) + codeSize) > bufferSize)
				return false;

			memcpy(writeBuffer, block->codeStart(), codeSize);
			writeBuffer += codeSize;

			block = block->next();
		}

		for (size_t i = 0, count = _relativeBranchRelocationList.size(); i < count; ++i) {
			NIns* branch = _relativeBranchRelocationList[i].first;
			const NIns* destination = _relativeBranchRelocationList[i].second;

			const CodeBlock* branchOwnerBlock = NULL; //findOwnerBlock(branch);
			const CodeBlock* destinationOwnerBlock = NULL; //findOwnerBlock(destination);

			ptrdiff_t branchLinearOffset = computeLinearOffset(branch, &branchOwnerBlock);
			ptrdiff_t destinationLinearOffset = computeLinearOffset(destination, &destinationOwnerBlock);

			NanoAssert(static_cast<size_t>(branchLinearOffset) < bufferSize);
			NanoAssert(static_cast<size_t>(destinationLinearOffset) < bufferSize);

			if (branchOwnerBlock != destinationOwnerBlock) {

				NIns* linearBranchToPatch = reinterpret_cast<NIns*>(outputBuffer + branchLinearOffset);
				NIns* linearDestination = reinterpret_cast<NIns*>(outputBuffer + destinationLinearOffset);

				NanoAssert(*linearBranchToPatch == *branch);
				NanoAssert(*linearDestination == *destination);

				nPatchRelativeBranch(linearBranchToPatch, linearDestination);
			}
		}

		return true;
	}

	bool Assembler::computeEntryPointLinearOffsetList(ptrdiff_t* linearOffsetList, size_t linearOffsetListSize) const {

		if (linearOffsetListSize < _entryPointList.size()) {
			NanoAssert(0);
			return false;
		}

		for (size_t i = 0, count = _entryPointList.size(); i < count; ++i) {
			const CodeBlock* ownerBlock = NULL;
			*(linearOffsetList + i) = computeLinearOffset(static_cast<const NIns*>(_entryPointList[i]), &ownerBlock);
		}

		return true;
	}

	bool Assembler::computeExitPointPatchLinearOffsetList(ptrdiff_t* linearOffsetList, size_t linearOffsetListSize) const {

		if (linearOffsetListSize < _exitPointPatchList.size()) {
			NanoAssert(0);
			return false;
		}

		for (size_t i = 0, count = _exitPointPatchList.size(); i < count; ++i) {
			const CodeBlock* ownerBlock = NULL;
			*(linearOffsetList + i) = computeLinearOffset(static_cast<const NIns*>(_exitPointPatchList[i]), &ownerBlock);
		}

		return true;
	}

	void Assembler::clearCompilationResults() {
		_codeBlockList.clear();
		_relativeBranchRelocationList.clear();
		_relativeBranchPatchList.clear();
		_entryPointList.clear();
	}

}

