/* -*- Mode: C++; c-basic-offset: 4; indent-tabs-mode: nil; tab-width: 4 -*- */
/* vi: set ts=4 sw=4 expandtab: (add to ~/.vimrc: set modeline modelines=5) */
/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */


#ifndef __nanojit_Assembler__
#define __nanojit_Assembler__

namespace nanojit
{
	/**
	 * Some notes on this Assembler (Emitter).
	 *
	 *   The class RegAlloc is essentially the register allocator from MIR
	 *
	 *   The Assembler class parses the LIR instructions starting at any point and converts
	 *   them to machine code.  It does the translation using expression trees which are simply
	 *   LIR instructions in the stream that have side-effects.  Any other instruction in the
	 *   stream is simply ignored.
	 *   This approach is interesting in that dead code elimination occurs for 'free', strength
	 *   reduction occurs fairly naturally, along with some other optimizations.
	 *
	 *   A negative is that we require state as we 'push' and 'pop' nodes along the tree.
	 *   Also, this is most easily performed using recursion which may not be desirable in
	 *   the mobile environment.
	 *
	 */

	class Noise
	{
		public:
			virtual ~Noise() {}

			// produce a random number from 0-maxValue for the JIT to use in attack mitigation
			virtual uint32_t getValue(uint32_t maxValue) = 0;
	};

	// error codes
	enum AssemblerError
	{
		 None = 0
		,StackFull
		,UnknownBranch
		,BranchTooFar
	};

	struct CompilationStats
	{
		size_t      reconcileBackwardMoveToRegCount;
		size_t      reconcileForwardMoveToRegCount;
		size_t      reconcileForceSpillCount;
		size_t      reconcileForceRestoreCount;
		size_t      spillCount;
		size_t      restoreCount;
	};

	typedef SeqBuilder<NIns*> NInsList;
	typedef HashMap<NIns*, LIns*, 16> NInsMap;
	typedef PodVector<Pair<NIns*, LIns*> > BranchPatchList;
	typedef PodVector<Pair<NIns*, const NIns*> > RelocationList;
	typedef PodVector<const void*> EntryPointList;
	typedef PodVector<const NIns*> RelativeOffsetPatchList;
	typedef PodVector<NIns*> ExitPointPatchList;

#if NANOJIT_USES_IMMD_POOL
	typedef HashMap<uint64_t, uint64_t*> ImmDPoolMap;
#endif

	/**
-    * Some architectures (i386, X64) can emit two branches that need patching
-    * in some situations. This is returned by asm_branch() implementations
-    * with 0, 1 or 2 of these fields set to a non-NULL value. (If only 1 is set, 
-    * it must be patch1, not patch2.)
-    */
	class Branches
	{
		static const unsigned int MAX_BRANCH = 3;

		NIns* branch[MAX_BRANCH];
		const NIns* destination[MAX_BRANCH];
		size_t count;

	public:

		inline Branches()
			: count(0) {
		}

		inline void addBranch(NIns* b, const NIns* d) {
			NanoAssert(count < MAX_BRANCH);
			branch[count] = b;
			destination[count] = d;
			count++;
		}

		inline size_t getCount() const {
			return count;
		}

		inline NIns* getBranch(size_t branchIndex) {
			NanoAssert(branchIndex < count);
			return branch[branchIndex];
		}

		inline const NIns* getDestination(size_t branchIndex) {
			NanoAssert(branchIndex < count);
			return destination[branchIndex];
		}
	};

	/**
	 * Information about the activation record for the method is built up
	 * as we generate machine code.  As part of the prologue, we issue
	 * a stack adjustment instruction and then later patch the adjustment
	 * value.  Temporary values can be placed into the AR as method calls
	 * are issued.
	 */
	class Assembler
	{
#ifdef NANOJIT_VERBOSE
		public:
			// Buffer for holding text as we generate it in reverse order.
			StringList* _outputCache;

			// Outputs the format string and 'outlineEOL', and resets
			// 'outline' and 'outlineEOL'.
			void outputf(const char* format, ...);

		private:
			static int disassembler_printf(void* arg, const char* fmt, ...);
			static void* disassembler_event(void* cookie, const char* event, void* arg);

			// Log controller object.  Contains what-stuff-should-we-print
			// bits, and a sink function for debug printing.
			LogControl* _logc;

			// instruction printer
			LInsPrinter* _printer;

			// Buffer used in most of the output function.  It must big enough
			// to hold both the output line and the 'outlineEOL' buffer, which
			// is concatenated onto 'outline' just before it is printed.
			static char  outline[8192];
			// Buffer used to hold extra text to be printed at the end of some
			// lines.
			static char  outlineEOL[512];

			// Outputs 'outline' and 'outlineEOL', and resets them both.
			// Output goes to '_outputCache' if it's non-NULL, or is printed
			// directly via '_logc'.
			void output();

			void maybe_disassemble();
			void disassemble(void* from, void* to);

			// Sets 'outlineEOL'.
			void setOutputForEOL(const char* format, ...);

			void printRegState();
			void printActivationState();
#endif // NANOJIT_VERBOSE

		public:
			Assembler(CodeAlloc* codeAlloc, Allocator* internalAlloc,
					  const Config& config
					  verbose_only(, LogControl* logc, LInsPrinter* printer = NULL));
			~Assembler();

			void								compile(LirChainWriter* writer, unsigned int arGroupCount);

			const Config&						getConfig() const;

			void								reset(unsigned int ar);
			
			AssemblerError						getError() const;
			bool								hasError() const;

#ifdef NANOJIT_VERBOSE
			const CompilationStats*				getCompilationStats() const;
#endif
			inline CodeBlockList				getCodeBlockList() const;
			// note: entry points are in reverse order
			inline const EntryPointList&			getEntryPointList() const;
			inline const RelativeOffsetPatchList&	getRelativeOffsetPatchList() const;
			inline const ExitPointPatchList&		getExitPointPatchList() const;

			inline const RelocationList&		getBranchRelocationList() const;

			ptrdiff_t							computeLinearOffset(const NIns* ins, CodeBlock const** ownerBlock) const;
			bool								flattenCodeBlockList(void* buffer, size_t bufferSize) const;
			bool								computeEntryPointLinearOffsetList(ptrdiff_t* linearOffsetList, size_t linearOffsetListSize) const;
			bool								computeExitPointPatchLinearOffsetList(ptrdiff_t* linearOffsetList, size_t linearOffsetListSize) const;

			void								clearCompilationResults();

			static inline void					patchRelativeOffset(const NIns *originalAddress, NIns *opcode) { nPatchRelativeOffset(originalAddress, opcode); }
			static inline void					patchExitPointToStaticJump(const NIns* destinationAddress, NIns* patchableExitPoint) { nPatchExitPointToStaticJump(destinationAddress, patchableExitPoint); };

private:
			void        beginAssembly(unsigned int arGroupCount);
			void        assemble(LirFilter* reader);
			void        endAssembly();

			void        setNoiseGenerator(Noise* noise)  { _noise = noise; } // used for attack mitigation; setting to 0 disables all mitigations

			void        setError(AssemblerError e) { _err = e; }
			void        cleanupAfterError();
			void        clearNInsPtrs();

			debug_only ( void       pageValidate(); )

			// support calling out from a fragment ; used to debug the jit
			debug_only( void        resourceConsistencyCheck(); )
			debug_only( void        registerConsistencyCheck(); )

private:
			void        gen(LirFilter* toCompile);
			NIns*       nEntryPoint(RegisterMask saveMask);
			NIns*       nExitPoint(RegisterMask restoreMask, bool patchableToDirectJump);
			void        nPrologue();

			uint32_t    arReserve(LIns* ins);
			void        arRelease(LIns* ins);
			void        arReset(unsigned int arGroupCount);

			void        evictAllActiveRegs() {
				// The evicted set will be be intersected with activeSet(),
				// so use an all-1s mask to avoid an extra load or call.
				evictSomeActiveRegs(~RegisterMask(0));
			}
			void        evictSomeActiveRegs(RegisterMask regs);

			Register    getBaseReg(LIns *ins, int &d, RegisterMask allow);
			void        getBaseReg2(RegisterMask allowValue, LIns* value, Register& rv,
									RegisterMask allowBase, LIns* base, Register& rb, int &d);
#if NANOJIT_USES_IMMD_POOL
			const uint64_t* findImmDFromPool(uint64_t q);
#endif

			int         findMemFor(LIns* ins);
			void        reallocateArEntry(LIns* ins, uint32_t entry);
			Register    findRegFor(LIns* ins, RegisterMask allow);
			void        findRegFor2(RegisterMask allowa, LIns* ia, Register &ra,
									RegisterMask allowb, LIns *ib, Register &rb);
			Register    findSpecificRegFor(LIns* ins, Register r);
			Register    findSpecificRegForUnallocated(LIns* ins, Register r);
			Register    prepareResultReg(LIns *ins, RegisterMask allow);
			void        freeResourcesOf(LIns *ins);
			void        evictIfActive(Register r);
			void        evict(LIns* vic);

			void        setBlockMargins(CodeBlock* block, NIns* start, NIns* end);

			void        codeAlloc();
			void        finalizeCodeAlloc();


			Allocator*					_internalAlloc;     // for items with same lifetime as this Assembler
			CodeAlloc*					_codeAlloc;         // for code we generate

			BranchPatchList				_relativeBranchPatchList;
			RelocationList				_relativeBranchRelocationList;
			EntryPointList				_entryPointList;
			RelativeOffsetPatchList		_relativeOffsetPatchList;
			ExitPointPatchList			_exitPointPatchList;

			Noise*						_noise;             // object to generate random noise used when hardening enabled.

		#if NANOJIT_USES_IMMD_POOL
			ImmDPoolMap         _immDPool;
		#endif

private:
			CodeBlockList	_codeBlockList;         // used blocks of code.
			CodeBlock*		currentCodeBlock;
			NIns*			_nIns;                  // current instruction in current normal code chunk
			NIns*			_nLastIns;

		#ifdef NANOJIT_VERBOSE
			NIns*       _nInsAfter;             // next instruction (ascending) in current normal/exit code chunk (for verbose output)
			size_t      codeBytes;              // bytes allocated in normal code chunks

			CompilationStats	compilationStats;
		#endif

			AssemblerError   _err;           // 0 = means assemble() appears ok, otherwise it failed

			// Holds the current instruction during gen().
			LIns*       currIns;

			AR          _activation;
			RegAlloc    _allocator;

			const Config& _config;


			unsigned int	_nReservedNInsCount;

			void		underrunProtect(size_t count);
			bool		reserveNInsSlotAtCodeStart(unsigned int count);
			void		freeReservedNInsSlot(unsigned int count);


			static unsigned int computeInitialReservedNInsCount(NIns* curIns, CodeBlock* codeBlock);

			void		nativePageReset();
			void		nativePageSetup();
			void		nativePageSplit();

			CodeBlock*	findOwnerBlock(const NIns* ins);

			void        nSpill(Register rr, int d, int8_t nWords);

			void        asm_jmp(LIns* ins);
			void        asm_jcc(LIns* ins);
			void        asm_entry(LIns* ins);
			void        asm_exit(LIns* ins);

			void        nSaveFlag(LIns* ins);
			void        nRestoreFlag(LIns* ins);

			// WARNING: the implementation of nRestore() should emit fast code
			// to rematerialize instructions where nCanRemat() returns true.
			// Otherwise, register allocation decisions will be suboptimal.
			void        nRestore(LIns*, Register);

			bool        asm_maybe_spill(LIns* ins, bool pop);

			void        nSpill(const LIns* insToSpill);

			void        nImmI(LIns* ins);
#ifdef NANOJIT_64BIT
			void        nImmQ(LIns* ins);
#endif
			void        nImmF(LIns* ins);
			void        nImmD(LIns* ins);

			void        nLoad32(LIns* ins);
			void        nLoad64(LIns* ins);

			void        nStore32(LOpcode op, LIns *val, int d, LIns *base);
			void        nStore64(LOpcode op, LIns *val, int d, LIns *base);

			void        nArith(LIns* ins);
			void        nNegNot(LIns* ins);
			void        nCmov(LIns* ins);
			void        asm_param(LIns* ins);
			void        nNegAbs(LIns* ins);        // fpu neg, abs
			void        nRecipSqrt(LIns* ins);     // fpu recip, rsqrt, sqrt
			void        nFop(LIns* ins);            // fpu add, sub, mul, div
			void        nFconv(LIns* ins);          // fpu int <=> float conversions
			void        nMull(LIns* ins);
			void        nMullhi(LIns* ins);

			// reinterpret  - try to keep original value bit pattern
			// only exception is si2sq which will sign extend the i value (added for simplicity)
			void		nMovI(LIns* ins);     // move from double to int
#if defined(NANOJIT_64BIT)
			void		nMovQ(LIns* ins);   // move from signed int to quad  (sign extend)
#endif
			void		nMovF(LIns* ins);     // move from int to float
			void		nMovD(LIns* ins);     // move from int to double


			void        nGpCopy(Register d, Register s);
			void        nNonGpCopy(Register d, Register s);
			void        nCall(LIns* ins, const NIns* address);
			NIns*       nRelativeBranch(const NIns* targ);
			void        nRelativeBranchCond(bool branchOnFalse, LIns* cond, const NIns* targ, bool patchable, Branches* branches);
			void        nInsertRandomNop();
			void        nLabel();
			void        nStoreToStack(LTy type, Register r, int offset);
			void        nLoadEffectiveAddress(Register r, Register b, int offset);

			void        nCopyI(LIns* ins);
			void        nCopyQ(LIns* ins);
			void        nCopyF(LIns* ins);
			void        nCopyD(LIns* ins);

			void        reconcileExitRegs(ParameterValue* paramData, unsigned int paramCount);
			void		evictCallParamValues(ParameterValue* callParams, unsigned int paramCount, RegisterMask abiEvictMask, RegisterMask resultRegisterMask);
			void        makeCallParamExtend(ParameterValue* paramData, unsigned int paramCount);

			bool        canReuseStorage(LIns* insCopy);
			void        reuseStorage(LIns* insCopy, RegisterMask allow);

			// platform specific implementation (see NativeXXX.cpp file)
			void        nCreate();
			void        nDestroy();
			void        nBeginAssembly(Allocator* compilationAllocator);
			void        nEndAssembly(Allocator* compilationAllocator);

			static void nPatchRelativeBranch(NIns* branch, const NIns* target);
			static void nPatchRelativeOffset(const NIns *originalAddress, NIns *opcodeAddress);
			static void nPatchExitPointToStaticJump(const NIns* destinationAddress, NIns* patchableExitPoint);

			void        spawnValue(LIns* ins, RegisterMask allow);

			void        asm_phiReconcile(LIns* ins);
			void        asm_phiReconcileForwards(LIns* ins);
			void        asm_phiReconcileBackwards(LIns* ins);
			void        asm_phiCapture(LIns* ins);
			void		setReconcileDefaultRegAlloc(LIns* reconcileIns, PhiState* phiStateArray, size_t phiStateCount);
			void		asm_reconcileStateBackward(LIns* value, bool inReg, RegisterMask regmask, bool inAr, unsigned int arIndex);

			void        asm_spillFence(LIns* ins);
			void        asm_spillFenceAll(LIns* ins);
			void        asm_registerFence(LIns* ins);
			void        asm_registerFenceAll(LIns* ins);
			void        asm_killFence(LIns* ins);
			void        asm_killFenceAll(LIns* ins);

			Register    allocReg(LIns* ins, RegisterMask allow);
			Register    allocTempReg(LTy insType, RegisterMask allow);
			Register    moveToReg(LIns* ins, RegisterMask allow);
			static void setResultLive(LIns* ins);
			void        asm_copy_slot(LIns* ins, LIns* insToAssign);

			// platform specific methods
		public:
			DECLARE_PLATFORM_ASSEMBLER()
	};

	inline int32_t arDisp(uint32_t arIndex) {
		// even on 64bit cpu's, we allocate stack area in 1byte chunks
		return int32_t(arIndex);
	}

	inline const Config& Assembler::getConfig() const {
		return _config;
	}

	inline AssemblerError Assembler::getError() const {
		return _err;
	}

	inline bool Assembler::hasError() const {
		return _err != None;
	}

#ifdef NANOJIT_VERBOSE
	inline const CompilationStats* Assembler::getCompilationStats() const {
		return &compilationStats;
	}
#endif

	inline CodeBlockList Assembler::getCodeBlockList() const {
		return _codeBlockList;
	}

	inline const EntryPointList& Assembler::getEntryPointList() const {
		return _entryPointList; 
	}
	
	inline const RelativeOffsetPatchList& Assembler::getRelativeOffsetPatchList() const {
		return _relativeOffsetPatchList;
	}

	inline const ExitPointPatchList& Assembler::getExitPointPatchList() const {
		return _exitPointPatchList;
	}

	inline const RelocationList& Assembler::getBranchRelocationList() const {
		return _relativeBranchRelocationList;
	}

	inline void Assembler::setResultLive(LIns* ins) {
		ins->setResultLive();
	}

	inline Register Assembler::findSpecificRegFor(LIns* i, Register w) {
		return findRegFor(i, rmask(w));
	}

	// Like findRegFor(), but called when the LIns is used as a pointer.  It
	// doesn't have to be called, findRegFor() can still be used.
	inline Register Assembler::getBaseReg(LIns* base, int &d, RegisterMask  allow) {

		if (base->isop(LIR_getfp))
			return FP;

		(void) d;
		Register r =  findRegFor(base, allow);
		return r;
	}

	// Like findRegFor2(), but used for stores where the base value has the
	// same type as the stored value, eg. in nStore32() on 32-bit platforms
	// and nStore64() on 64-bit platforms.
	inline void Assembler::getBaseReg2(RegisterMask allowValue, LIns* value, Register&  rv,
		RegisterMask allowBase, LIns* base, Register& rb, int &d) {

		(void) d;
		findRegFor2(allowValue, value, rv, allowBase, base, rb);
	}

}
#endif // __nanojit_Assembler__
