/* -*- Mode: C++; c-basic-offset: 4; indent-tabs-mode: nil; tab-width: 4 -*- */
/* vi: set ts=4 sw=4 expandtab: (add to ~/.vimrc: set modeline modelines=5) */
/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

#ifndef __nanojit_LIR__
#define __nanojit_LIR__

#include <stdlib.h>
#include <string.h>

namespace nanojit
{
	enum LOpcode
	{
#define OP___(op, repKind, retType, isCse) \
		LIR_##op,
#include "LIRopcode.tbl"
		LIR_sentinel,
#undef OP___

#ifdef NANOJIT_64BIT
#  define PTR_SIZE(a,b)  b
#else
#  define PTR_SIZE(a,b)  a
#endif

		// Pointer-sized synonyms.

		LIR_paramp  = PTR_SIZE(LIR_parami,  LIR_paramq),

		LIR_livep   = PTR_SIZE(LIR_livei,   LIR_liveq),

		LIR_ldp     = PTR_SIZE(LIR_ldi,     LIR_ldq),

		LIR_stp     = PTR_SIZE(LIR_sti,     LIR_stq),

		LIR_callp   = PTR_SIZE(LIR_calli,   LIR_callq),

		LIR_eqp     = PTR_SIZE(LIR_eqi,     LIR_eqq),
		LIR_nep     = PTR_SIZE(LIR_nei,     LIR_neq),
		LIR_ltp     = PTR_SIZE(LIR_lti,     LIR_ltq),
		LIR_gtp     = PTR_SIZE(LIR_gti,     LIR_gtq),
		LIR_lep     = PTR_SIZE(LIR_lei,     LIR_leq),
		LIR_gep     = PTR_SIZE(LIR_gei,     LIR_geq),
		LIR_ltup    = PTR_SIZE(LIR_ltui,    LIR_ltuq),
		LIR_gtup    = PTR_SIZE(LIR_gtui,    LIR_gtuq),
		LIR_leup    = PTR_SIZE(LIR_leui,    LIR_leuq),
		LIR_geup    = PTR_SIZE(LIR_geui,    LIR_geuq),

		LIR_addp    = PTR_SIZE(LIR_addi,    LIR_addq),
		LIR_subp    = PTR_SIZE(LIR_subi,    LIR_subq),

		LIR_andp    = PTR_SIZE(LIR_andi,    LIR_andq),
		LIR_andcp   = PTR_SIZE(LIR_andci,   LIR_andcq),
		LIR_orp     = PTR_SIZE(LIR_ori,     LIR_orq),
		LIR_xorp    = PTR_SIZE(LIR_xori,    LIR_xorq),

		LIR_lshp    = PTR_SIZE(LIR_lshi,    LIR_lshq),
		LIR_rshp    = PTR_SIZE(LIR_rshi,    LIR_rshq),
		LIR_rshup   = PTR_SIZE(LIR_rshui,   LIR_rshuq),

		LIR_cmovp   = PTR_SIZE(LIR_cmovi,   LIR_cmovq)
	};

// Check that all opcodes are between 0 and 255.
#define OP___(op, repKind, retType, isCse) \
NanoStaticAssert(LIR_##op >= 0 && LIR_##op < 256);
#include "LIRopcode.tbl"
#undef OP___
NanoStaticAssert(LIR_start == 0 && LIR_sentinel <= 256); // It's ok if LIR_sentinel is 256 since it's not actually used as opcode.
	
	// 32-bit integer comparisons must be contiguous, as must 64-bit integer
	// comparisons and 64-bit float comparisons.
	NanoStaticAssert(LIR_eqi + 1 == LIR_nei  &&
					 LIR_eqi + 2 == LIR_lti  &&
					 LIR_eqi + 3 == LIR_gei  &&
					 LIR_eqi + 4 == LIR_gti  &&
					 LIR_eqi + 5 == LIR_lei  &&
					 LIR_eqi + 6 == LIR_ltui &&
					 LIR_eqi + 7 == LIR_geui &&
					 LIR_eqi + 8 == LIR_gtui &&
					 LIR_eqi + 9 == LIR_leui);
#ifdef NANOJIT_64BIT
	NanoStaticAssert(LIR_eqq + 1 == LIR_neq  &&
					 LIR_eqq + 2 == LIR_ltq  &&
					 LIR_eqq + 3 == LIR_geq  &&
					 LIR_eqq + 4 == LIR_gtq  &&
					 LIR_eqq + 5 == LIR_leq  &&
					 LIR_eqq + 6 == LIR_ltuq &&
					 LIR_eqq + 7 == LIR_geuq &&
					 LIR_eqq + 8 == LIR_gtuq &&
					 LIR_eqq + 9 == LIR_leuq);
#endif
	NanoStaticAssert(LIR_eqd + 1 == LIR_ned &&
					 LIR_eqd + 2 == LIR_ltd &&
					 LIR_eqd + 3 == LIR_ged &&
					 LIR_eqd + 4 == LIR_gtd &&
					 LIR_eqd + 5 == LIR_led);

	NanoStaticAssert(LIR_eqf + 1 == LIR_nef &&
					 LIR_eqf + 2 == LIR_ltf &&
					 LIR_eqf + 3 == LIR_gef &&
					 LIR_eqf + 4 == LIR_gtf &&
					 LIR_eqf + 5 == LIR_lef);

	// Various opcodes must be changeable to their opposite with op^1
	// (although we use invertXyz() when possible, ie. outside static
	// assertions).
	NanoStaticAssert((LIR_jt^1) == LIR_jf && (LIR_jf^1) == LIR_jt);

	NanoStaticAssert((LIR_eqi^1)  == LIR_nei);
	// invert
	NanoStaticAssert((LIR_lti^1)  == LIR_gei  && (LIR_gei^1)  == LIR_lti);
	NanoStaticAssert((LIR_lei^1)  == LIR_gti  && (LIR_gti^1)  == LIR_lei);
	NanoStaticAssert((LIR_ltui^1) == LIR_geui && (LIR_geui^1) == LIR_ltui);
	NanoStaticAssert((LIR_leui^1) == LIR_gtui && (LIR_gtui^1) == LIR_leui);
	// swap
	NanoStaticAssert((LIR_lti^2)  == LIR_gti  && (LIR_gti^2)  == LIR_lti);
	NanoStaticAssert((LIR_lei^2)  == LIR_gei  && (LIR_gei^2)  == LIR_lei);
	NanoStaticAssert((LIR_ltui^2) == LIR_gtui && (LIR_gtui^2) == LIR_ltui);
	NanoStaticAssert((LIR_leui^2) == LIR_geui && (LIR_geui^2) == LIR_leui);

#ifdef NANOJIT_64BIT
	NanoStaticAssert((LIR_eqq^1)  == LIR_neq);
	// invert
	NanoStaticAssert((LIR_ltq^1)  == LIR_geq  && (LIR_geq^1)  == LIR_ltq);
	NanoStaticAssert((LIR_leq^1)  == LIR_gtq  && (LIR_gtq^1)  == LIR_leq);
	NanoStaticAssert((LIR_ltuq^1) == LIR_geuq && (LIR_geuq^1) == LIR_ltuq);
	NanoStaticAssert((LIR_leuq^1) == LIR_gtuq && (LIR_gtuq^1) == LIR_leuq);
	// swap
	NanoStaticAssert((LIR_ltq^2)  == LIR_gtq  && (LIR_gtq^2)  == LIR_ltq);
	NanoStaticAssert((LIR_leq^2)  == LIR_geq  && (LIR_geq^2)  == LIR_leq);
	NanoStaticAssert((LIR_ltuq^2) == LIR_gtuq && (LIR_gtuq^2) == LIR_ltuq);
	NanoStaticAssert((LIR_leuq^2) == LIR_geuq && (LIR_geuq^2) == LIR_leuq);
#endif

	NanoStaticAssert((LIR_eqf^1)  == LIR_nef);
	// invert
	NanoStaticAssert((LIR_ltf^1) == LIR_gef && (LIR_gef^1) == LIR_ltf);
	NanoStaticAssert((LIR_lef^1) == LIR_gtf && (LIR_gtf^1) == LIR_lef);
	// swap
	NanoStaticAssert((LIR_ltf^2) == LIR_gtf && (LIR_gtf^2) == LIR_ltf);
	NanoStaticAssert((LIR_lef^2) == LIR_gef && (LIR_gef^2) == LIR_lef);

	NanoStaticAssert((LIR_eqd^1) == LIR_ned);
	// invert
	NanoStaticAssert((LIR_ltd^1) == LIR_ged && (LIR_ged^1) == LIR_ltd);
	NanoStaticAssert((LIR_led^1) == LIR_gtd && (LIR_gtd^1) == LIR_led);
	// swap
	NanoStaticAssert((LIR_ltd^2) == LIR_gtd && (LIR_gtd^2) == LIR_ltd);
	NanoStaticAssert((LIR_led^2) == LIR_ged && (LIR_ged^2) == LIR_led);


	const unsigned int LIR_CMPF_TO_CMPD_OFFSET = LIR_eqf - LIR_eqd;

	NanoStaticAssert(LIR_eqf==LIR_eqd+LIR_CMPF_TO_CMPD_OFFSET);
	NanoStaticAssert(LIR_nef==LIR_ned+LIR_CMPF_TO_CMPD_OFFSET);
	NanoStaticAssert(LIR_ltf==LIR_ltd+LIR_CMPF_TO_CMPD_OFFSET);
	NanoStaticAssert(LIR_gef==LIR_ged+LIR_CMPF_TO_CMPD_OFFSET);
	NanoStaticAssert(LIR_gtf==LIR_gtd+LIR_CMPF_TO_CMPD_OFFSET);
	NanoStaticAssert(LIR_lef==LIR_led+LIR_CMPF_TO_CMPD_OFFSET);
	NanoStaticAssert(LIR_unf==LIR_und+LIR_CMPF_TO_CMPD_OFFSET);

	// This is much the same as LTy, but we need to distinguish signed and
	// unsigned 32-bit ints so that they will be extended to 64-bits correctly
	// on 64-bit platforms.
	//
	enum ArgType {
		ARGTYPE_V  = 0,     // void
		ARGTYPE_I  = 1,     // int32_t
		ARGTYPE_UI = 2,     // uint32_t
#ifdef NANOJIT_64BIT
		ARGTYPE_Q  = 3,     // uint64_t
#endif
		ARGTYPE_D  = 4,     // double
		ARGTYPE_F  = 5,     // single-precision float;
		// aliases
		ARGTYPE_P = PTR_SIZE(ARGTYPE_I, ARGTYPE_Q), // pointer
		ARGTYPE_B = ARGTYPE_I,                       // bool
		ARGTYPE_FP                                   // will store FP (if not already set)
	};

	enum IndirectCall {
		CALL_INDIRECT = 0
	};

	//-----------------------------------------------------------------------
	// Aliasing
	// --------
	// *Aliasing* occurs when a single memory location can be accessed through
	// multiple names.  For example, consider this code:
	//
	//   ld a[0]
	//   sti b[0]
	//   ld a[0]
	//
	// In general, it's possible that a[0] and b[0] may refer to the same
	// memory location.  This means, for example, that you cannot safely
	// perform CSE on the two loads.  However, if you know that 'a' cannot be
	// an alias of 'b' (ie. the two loads do not alias with the store) then
	// you can safely perform CSE.
	//
	// Access regions
	// --------------
	// Doing alias analysis precisely is difficult.  But it turns out that
	// keeping track of aliasing at a coarse level is enough to help with many
	// optimisations.  So we conceptually divide the memory that is accessible
	// from LIR into a small number of "access regions" (aka. "Acc").  An
	// access region may be non-contiguous.  No two access regions can
	// overlap.  The union of all access regions covers all memory accessible
	// from LIR.
	//
	// In general a (static) load or store may be executed more than once, and
	// thus may access multiple regions;  however, in practice almost all
	// loads and stores will obviously access only a single region.  A
	// function called from LIR may load and/or store multiple access regions
	// (even if executed only once).
	//
	// If two loads/stores/calls are known to not access the same region(s),
	// then they do not alias.
	//
	// All regions are defined by the embedding.  It makes sense to add new
	// embedding-specific access regions when doing so will help with one or
	// more optimisations.
	//
	// Access region sets and instruction markings
	// -------------------------------------------
	// Each load/store is marked with an "access region set" (aka. "AccSet"),
	// which is a set of one or more access regions.  This indicates which
	// parts of LIR-accessible memory the load/store may touch.
	//
	// Each function called from LIR is also marked with an access region set
	// for memory stored to by the function.  (We could also have a marking
	// for memory loads done by the function, but there's no need at the
	// moment.)  These markings apply to the function itself, not the call
	// site, ie. they're not context-sensitive.
	//
	// These load/store/call markings MUST BE ACCURATE -- if not then invalid
	// optimisations might occur that change the meaning of the code.
	// However, they can safely be imprecise (ie. conservative), ie. a
	// load/store/call can be marked with an access region set that is a
	// superset of the actual access region set.  Such imprecision is safe but
	// may reduce optimisation opportunities.
	//
	// Optimisations that use access region info
	// -----------------------------------------
	// Currently only CseFilter uses this, and only for determining whether
	// loads can be CSE'd.  Note that CseFilter treats loads that are marked
	// with a single access region precisely, but all loads marked with
	// multiple access regions get lumped together.  So if you can't mark a
	// load with a single access region, you might as well use ACC_LOAD_ANY.
	//-----------------------------------------------------------------------

	// An access region set is represented as a bitset.  Using a uint32_t
	// restricts us to at most 32 alias regions for the moment.  This could be
	// expanded to a uint64_t easily if needed.
	typedef uint32_t AccSet;
	static const int NUM_ACCS = sizeof(AccSet) * 8;

	// Some common (non-singleton) access region sets.  ACCSET_NONE does not make
	// sense for loads or stores (which must access at least one region), it
	// only makes sense for calls.
	//
	static const AccSet ACCSET_NONE      = 0x0;
	static const AccSet ACCSET_ALL       = 0xffffffff;
	static const AccSet ACCSET_LOAD_ANY  = ACCSET_ALL;      // synonym
	static const AccSet ACCSET_STORE_ANY = ACCSET_ALL;      // synonym

	inline bool isSingletonAccSet(AccSet accSet) {
		// This is a neat way of testing if a value has only one bit set.
		return (accSet & (accSet - 1)) == 0;
	}

	// Full AccSets don't fit into load and store instructions.  But
	// load/store AccSets almost always contain a single access region.  We
	// take advantage of this to create a compressed AccSet, MiniAccSet, that
	// does fit.
	//
	// The 32 single-region AccSets get compressed into a number in the range
	// 0..31 (according to the position of the set bit), and all other
	// (multi-region) AccSets get converted into MINI_ACCSET_MULTIPLE.  So the
	// representation is lossy in the latter case, but that case is rare for
	// loads/stores.  We use a full AccSet for the storeAccSets of calls, for
	// which multi-region AccSets are common.
	//
	// We wrap the uint8_t inside a struct to avoid the possiblity of subtle
	// bugs caused by mixing up AccSet and MiniAccSet, which is easy to do.
	// However, the struct gets padded inside LInsLd in an inconsistent way on
	// Windows, so we actually store a MiniAccSetVal inside LInsLd.  Sigh.
	// But we use MiniAccSet everywhere else.
	//
	typedef uint8_t MiniAccSetVal;
	struct MiniAccSet { MiniAccSetVal val; };
	static const MiniAccSet MINI_ACCSET_MULTIPLE = { 99 };

	static MiniAccSet compressAccSet(AccSet accSet) {
		if (isSingletonAccSet(accSet)) {
			MiniAccSet ret = { uint8_t(msbSet32(accSet)) };
			return ret;
		}

		// If we got here, it must be a multi-region AccSet.
		return MINI_ACCSET_MULTIPLE;
	}

	static AccSet decompressMiniAccSet(MiniAccSet miniAccSet) {
		return (miniAccSet.val == MINI_ACCSET_MULTIPLE.val) ? ACCSET_ALL : (1 << miniAccSet.val);
	}

	// The LoadQual affects how a load can be optimised:
	//
	// - CONST: These loads are guaranteed to always return the same value
	//   during a single execution of a fragment (but the value is allowed to
	//   change between executions of the fragment).  This means that the
	//   location is never stored to by the LIR, and is never modified by an
	//   external entity while the fragment is running.
	//
	// - NORMAL: These loads may be stored to by the LIR, but are never
	//   modified by an external entity while the fragment is running.
	//
	// - VOLATILE: These loads may be stored to by the LIR, and may be
	//   modified by an external entity while the fragment is running.
	//
	// This gives a lattice with the ordering:  CONST < NORMAL < VOLATILE.
	// As usual, it's safe to mark a load with a value higher (less precise)
	// that actual, but it may result in fewer optimisations occurring.
	//
	// Generally CONST loads are highly amenable to optimisation (eg. CSE),
	// VOLATILE loads are entirely unoptimisable, and NORMAL loads are in
	// between and require some alias analysis to optimise.
	//
	// Note that CONST has a stronger meaning to "const" in C and C++;  in C
	// and C++ a "const" variable may be modified by an external entity, such
	// as hardware.  Hence "const volatile" makes sense in C and C++, but
	// CONST+VOLATILE doesn't make sense in LIR.
	//
	// Note also that a 2-bit bitfield in LInsLd is used to hold LoadQual
	// values, so you can one add one more value without expanding it.
	//
	enum LoadQual {
		LOAD_CONST    = 0,
		LOAD_NORMAL   = 1,
		LOAD_VOLATILE = 2
	};

	// Array holding the 'isCse' field from LIRopcode.tbl.
	extern const int8_t isCses[];       // cannot be uint8_t, some values are negative

	inline bool isCseOpcode(LOpcode op) {
		NanoAssert(isCses[op] != -1);   // see LIRopcode.tbl to understand this
		return isCses[op] == 1;
	}
	inline bool isLiveOpcode(LOpcode op) {
		return
#if defined NANOJIT_64BIT
			   op == LIR_liveq ||
#endif
			   op == LIR_livef ||
			   op == LIR_livei || op == LIR_lived;
	}
	inline bool isCmovOpcode(LOpcode op) {
		return
#if defined NANOJIT_64BIT
			op == LIR_cmovq ||
#endif
			op == LIR_cmovf ||
			op == LIR_cmovi ||
			op == LIR_cmovd;
	}
	inline bool isCmpIOpcode(LOpcode op) {
		return LIR_eqi <= op && op <= LIR_leui;
	}
	inline bool isCmpSIOpcode(LOpcode op) {
		return LIR_eqi <= op && op <= LIR_lei;
	}
	inline bool isCmpUIOpcode(LOpcode op) {
		return (LIR_eqi == op) || (LIR_nei == op) || (op >= LIR_ltui && op <= LIR_leui);
	}
#ifdef NANOJIT_64BIT
	inline bool isCmpQOpcode(LOpcode op) {
		return LIR_eqq <= op && op <= LIR_leuq;
	}
	inline bool isCmpSQOpcode(LOpcode op) {
		return LIR_eqq <= op && op <= LIR_leq;
	}
	inline bool isCmpUQOpcode(LOpcode op) {
		return (LIR_eqq == op) || (LIR_neq == op) || (op >= LIR_ltuq && op <= LIR_leuq);
	}
#endif
	inline bool isCmpDOpcode(LOpcode op) {
		return LIR_eqd <= op && op <= LIR_und;
	}
	inline bool isCmpFOpcode(LOpcode op) {
		return LIR_eqf <= op && op <= LIR_unf;
	}
	inline bool isCmpOpcode(LOpcode op) {
		return isCmpIOpcode(op) ||
#if defined NANOJIT_64BIT
			   isCmpQOpcode(op) ||
#endif
			   isCmpFOpcode(op) ||
			   isCmpDOpcode(op);
	}

	inline LOpcode getCmpFOpcode(LOpcode op) {
		NanoAssert(isCmpDOpcode(op));
		return LOpcode(op + LIR_CMPF_TO_CMPD_OFFSET);
	}

	inline LOpcode getCmpDOpcode(LOpcode op) {
		NanoAssert(isCmpFOpcode(op));
		return LOpcode(op - LIR_CMPF_TO_CMPD_OFFSET);
	}

	inline LOpcode invertCondJmpOpcode(LOpcode op) {
		NanoAssert(op == LIR_jt || op == LIR_jf);
		return LOpcode(op ^ 1);
	}

	// invert condition
	inline LOpcode invertCmpOpcode(LOpcode op) {
		NanoAssert(isCmpOpcode(op));
		return LOpcode(op ^ 1);
	}

	// swap operand order
	inline LOpcode swapCmpOpcode(LOpcode op) {
		NanoAssert(isCmpOpcode(op));
		NanoAssert((op != LIR_eqi) && (op != LIR_nei));
#if defined(NANOJIT_64BIT)
		NanoAssert((op != LIR_eqq) && (op != LIR_neq));
#endif
		NanoAssert((op != LIR_eqf) && (op != LIR_nef));
		NanoAssert((op != LIR_eqd) && (op != LIR_ned));

		return LOpcode(op ^ 2);
	}

#ifdef NANOJIT_64BIT
	LOpcode cmpOpcodeI2Q(LOpcode op);
#endif

	// Array holding the 'repKind' field from LIRopcode.tbl.
	extern const uint8_t repKinds[];

	enum LTy {
		LTy_I,		// int:  32-bit integer
#ifdef NANOJIT_64BIT
		LTy_Q,		// quad: 64-bit integer
#endif
		LTy_F,		// float:  32-bit float
		LTy_D,		// double: 64-bit float
		LTy_V,		// void: no value/no type
		LTy_P  = PTR_SIZE(LTy_I, LTy_Q)   // word-sized integer
	};

	// Array holding the 'retType' field from LIRopcode.tbl.
	extern const LTy retTypes[];

	// Array holding the size in bytes of each LIns from LIRopcode.tbl.
	extern const uint8_t insSizes[];


	//-----------------------------------------------------------------------
	// Low-level instructions.  This is a bit complicated, because we have a
	// variable-width representation to minimise space usage.
	//
	// - Instruction size is always an integral multiple of word size.
	//
	// - Every instruction has at least one word, holding the opcode and the
	//   reservation info ("SharedFields").  That word is in class LIns.
	//
	// - Beyond that, most instructions have 1, 2 or 3 extra words.  These
	//   extra words are in classes LInsOp1, LInsOp2, etc (collectively called
	//   "LInsXYZ" in what follows).  Each LInsXYZ class also contains an LIns,
	//   accessible by the 'ins' member, which holds the LIns data.
	//
	// - LIR is written forward, but read backwards.  When reading backwards,
	//   in order to find the opcode, it must be in a predictable place in the
	//   LInsXYZ isn't affected by instruction width.  Therefore, the LIns
	//   word (which contains the opcode) is always the *last* word in an
	//   instruction.
	//
	// - Each instruction is created by casting pre-allocated bytes from a
	//   LirBuffer to the LInsXYZ type.  Therefore there are no constructors
	//   for LIns or LInsXYZ.
	//
	// - The standard handle for an instruction is a LIns*.  This actually
	//   points to the LIns word, ie. to the final word in the instruction.
	//   This is a bit odd, but it allows the instruction's opcode to be
	//   easily accessed.  Once you've looked at the opcode and know what kind
	//   of instruction it is, if you want to access any of the other words,
	//   you need to use toLInsXYZ(), which takes the LIns* and gives you an
	//   LInsXYZ*, ie. the pointer to the actual start of the instruction's
	//   bytes.  From there you can access the instruction-specific extra
	//   words.
	//
	// - However, from outside class LIns, LInsXYZ isn't visible, nor is
	//   toLInsXYZ() -- from outside LIns, all LIR instructions are handled
	//   via LIns pointers and get/set methods are used for all LIns/LInsXYZ
	//   accesses.  In fact, all data members in LInsXYZ are private and can
	//   only be accessed by LIns, which is a friend class.  The only thing
	//   anyone outside LIns can do with a LInsXYZ is call getLIns().
	//
	// - An example Op2 instruction and the likely pointers to it (each line
	//   represents a word, and pointers to a line point to the start of the
	//   word on that line):
	//
	//      [ oprnd_2         <-- LInsOp2* insOp2 == toLInsOp2(ins)
	//        oprnd_1
	//        opcode + resv ] <-- LIns* ins
	//
	//
	// - Various things about the size and layout of LIns and LInsXYZ are
	//   statically checked in staticSanityCheck().  In particular, this is
	//   worthwhile because there's nothing that guarantees that all the
	//   LInsXYZ classes have a size that is a multiple of word size (but in
	//   practice all sane compilers use a layout that results in this).  We
	//   also check that every LInsXYZ is word-aligned in
	//   LirBuffer::makeRoom();  this seems sensible to avoid potential
	//   slowdowns due to misalignment.  It relies on chunks themselves being
	//   word-aligned, which is extremely likely.
	//
	// - There is an enum, LInsRepKind, with one member for each of the
	//   LInsXYZ kinds.  Each opcode is categorised with its LInsRepKind value
	//   in LIRopcode.tbl, and this is used in various places.
	//-----------------------------------------------------------------------

	enum LInsRepKind {
		// LRK_XYZ corresponds to class LInsXYZ.
		LRK_Op0,
		LRK_Op1,
		LRK_Op1b,
		LRK_Op2,
		LRK_Op3,
		LRK_Branch,
		LRK_Ld,
		LRK_St,
		LRK_Sk,
		LRK_Call,
		LRK_P,
		LRK_IorF,
		LRK_QorD,
		LRK_Label,
		LRK_Phi,
		LRK_RegisterFence,
		LRK_SpillFence,
		LRK_KillFence,
		LRK_None    // this one is used for unused opcode numbers
	};

	enum ArithFlag
	{
		ArithFlag_Zero				= 0x01,
		ArithFlag_Sign				= 0x02,
		ArithFlag_Carry				= 0x04,
		ArithFlag_Borrow			= 0x08,
		ArithFlag_Overflow			= 0x10,
		ArithFlag_FloatInvalid		= 0x20,
		ArithFlag_FloatOverflow		= 0x40,
		ArithFlag_FloatUnderflow	= 0x80,
	};

	class LIns;

	struct LiveInfo
	{
		LOpcode     opcode;
		LIns*       ins;
	};

	typedef uint16_t ArIndex;


	typedef SeqBuilder<LiveInfo> LiveList;
	typedef SeqBuilder<char*> StringList;
	typedef HashMap<LIns*,bool, 16> InsSet;


	struct PhiState
	{
		uint16_t        arIndex;
		uint8_t         regnum;
		bool            inReg;
		bool            inAr;
		bool            isLive;
#if defined(NANOJIT_DEBUG)
		unsigned int    arGroupIndex;
		LIns*           otherValue;
#endif
	};

	struct PhiActiveData
	{
		LIns*               ins;
		Register			reg;
	};

	struct PhiData
	{
		PhiActiveData       active[LastRegNum + 1];
		unsigned int        activeCount;
 		RegisterMask        activeMask;
		PhiState*           capturedStates;
		bool                forward;
	};

	struct ParameterValue
	{
		LIns*           value;
		Register        reg;
		int             stackOffset;
		bool			byRef;
	};

	struct CallParameters
	{
		ParameterValue*     values;
		unsigned int        count;
		RegisterMask        volatileMask;
	};

	// 0-operand form.  Used for LIR_start and LIR_label.
	class LInsOp0
	{
	};

	// 1-operand form.  Used for LIR_reti, unary arithmetic/logic ops, etc.
	class LInsOp1
	{
	private:
		friend class LIns;

		LIns*       oprnd_1;
	};

	// 2-operand form.  Used for guards, branches, comparisons, binary
	// arithmetic/logic ops, etc.
	class LInsOp2
	{
	private:
		friend class LIns;

		LIns*       oprnd_1;
		LIns*       oprnd_2;
	};

	// 3-operand form.  Used for conditional moves, jov branches, and xov guards.
	class LInsOp3
	{
	private:
		friend class LIns;

		LIns*       oprnd_1;
		LIns*       oprnd_2;
		LIns*       oprnd_3;
	};

	class LInsBranch
	{
	private:
		friend class LIns;
		LIns*       condition;
		LIns*       target;
	};

	// Used for all loads.
	class LInsLd
	{
	private:
		friend class LIns;

		LIns*       oprnd_1;

		// Nb: the LIR writer pipeline handles things if a displacement
		// exceeds 16 bits.  This is rare, but does happen occasionally.  We
		// could go to 24 bits but then it would happen so rarely that the
		// handler code would be difficult to test and thus untrustworthy.
		//
		// Nb: the types of these bitfields are all 32-bit integers to ensure
		// they are fully packed on Windows, sigh.  Also, 'loadQual' is
		// unsigned to ensure the values 0, 1, and 2 all fit in 2 bits.
		//
		// Nb: explicit signed keyword for bitfield types is required,
		// some compilers may treat them as unsigned without it.
		// See Bugzilla 584219 comment #18
		int16_t     disp;
		uint8_t     miniAccSetVal;
		uint8_t     loadQual;
	};

	// Used for all stores.
	class LInsSt
	{
	private:
		friend class LIns;

		LIns*       oprnd_1;
		LIns*       oprnd_2;

		int16_t     disp;
		uint8_t     miniAccSetVal;
	};

	// Used for LIR_skip.
	class LInsSk
	{
	private:
		friend class LIns;

		LIns*       prevLIns;
	};

	// Used for all variants of LIR_call.
	class LInsCall
	{
	private:
		friend class LIns;

		// parameter mapping
		CallParameters*     parameters;
		const void*         address;
		bool				isPure;
	};

	// Used for LIR_paramp.
	class LInsP
	{
	private:
		friend class LIns;
		Register		reg;
	};

	// Used for LIR_immi, LIR_immf
	class LInsIorF
	{
	private:
		friend class LIns;

		union
		{
			int32_t     asImmI;
			float       asImmF;
		};
	};

	// Used for LIR_immq and LIR_immd.
	class LInsQorD
	{
	private:
		friend class LIns;

		union
		{
			double      asImmD;
			int64_t     asImmQ;
		};
	};

	// Used for LIR_exit
	class LInsExit
	{
	private:
		friend class LIns;
		
		RegisterMask		restoreMask;
		unsigned int        count;
		ParameterValue*     values;
		bool				isPatchable;
	};

	// Used for LIR_entry
	class LInsEntry
	{
	private:
		friend class LIns;

		RegisterMask	saveMask;
		void*           address;
	};

	// Label nodes
	class LInsLabel
	{
	private:
		friend class LIns;
		NIns*       nativeAddress;
		bool        referenced;
	};

	// Phi nodes
	class LInsPhi
	{
	private:
		friend class LIns;
		PhiData*    phiData;
		LIns**		values;
		size_t      count;
		LIns*       reconcileNodeNext;
	};

	// Force spill slot assign
	class LInsSpillFence
	{
	private:
		friend class LIns;
		LIns**          valueArray;
		size_t          valueArraySize;
	};

	// Force spill slot assign and free register
	class LInsRegisterFence
	{
	private:
		friend class LIns;
		LIns**          valueArray;
		size_t          valueArraySize;
	};

	// force kill of values
	class LInsKillFence
	{
	private:
		friend class LIns;
		LIns**          valueArray;
		size_t          valueArraySize;
	};

	// Used only as a placeholder for OP___ macros for unused opcodes in
	// LIRopcode.tbl.
	class LInsNone
	{
	};

	struct LInsCompilationData
	{
		ArIndex     arIndex;                // index into stack frame; displ is 4*arIndex
		uint8_t     regnum;                 // register index

		bool        inReg;                  // if 1, 'reg' is active
		bool        inAr;                   // if 1, 'arIndex' is active

		bool        isResultLive;           // mask for liveliness of all values
		uint8_t     flagsRequested;         // bit set to request flags
	};

	static const ArIndex		INVALID_AR_INDEX = 0xFFFF;
	static const ArIndex		LINS_DEFAULT_AR_GROUP = 0xFFFF;


	union LInsData
	{
		LInsOp0           Op0;
		LInsOp1           Op1;
		LInsOp2           Op2;
		LInsOp3           Op3;
		LInsLd            Ld;
		LInsSt            St;
		LInsSk            Sk;
		LInsCall          Call;
		LInsP             P;
		LInsIorF          IorF;
		LInsQorD          QorD;
		LInsBranch        Branch;
		LInsExit          Exit;
		LInsEntry         Entry;
		LInsLabel         Label;
		LInsPhi           Phi;
		LInsSpillFence    SpillFence;
		LInsRegisterFence RegFence;
		LInsKillFence     KillFence;
	};

	class LIns
	{
	private:
		LInsCompilationData	compilationData;

		ArIndex				arGroup;
		uint8_t				arSize;
#if defined(NANOJIT_DEBUG)
		LOpcode				op;
#else
		uint8_t				op;
#endif
		bool				_allowFolding;
		bool				_genFlags;
		uint8_t				_captureRegIndex;
		LInsData			data;


		inline void initSharedFields(LOpcode opcode) {
			NanoAssert(((int)opcode) >= 0 && opcode <= 255);
#if defined(NANOJIT_DEBUG)
			op = opcode;
#else
			op = static_cast<uint8_t>(opcode);
#endif
			setCaptureReg(UnspecifiedReg);
			if (isV() || isCall())
				compilationData.isResultLive = true;
		}

		// LIns-to-LInsXYZ converters.
		inline LInsOp0*                 getLInsOp0();
		inline LInsOp1*                 getLInsOp1();
		inline LInsOp2*                 getLInsOp2();
		inline LInsOp3*                 getLInsOp3();
		inline LInsLd*                  getLInsLd();
		inline LInsSt*                  getLInsSt();
		inline LInsSk*                  getLInsSk();
		inline LInsCall*                getLInsCall();
		inline LInsP*                   getLInsP();
		inline LInsIorF*                getLInsIorF();
		inline LInsQorD*                getLInsQorD();
		inline LInsBranch*              getLInsBranch();
		inline LInsExit*                getLInsExit();
		inline LInsEntry*               getLInsEntry();
		inline LInsLabel*               getLInsLabel();
		inline LInsPhi*                 getLInsPhi();
		inline LInsSpillFence*          getLInsSpillFence();
		inline LInsRegisterFence*       getLInsRegisterFence();
		inline LInsKillFence*           getLInsKillFence();

		inline const LInsOp0*           getLInsOp0()            const;
		inline const LInsOp1*           getLInsOp1()            const;
		inline const LInsOp2*           getLInsOp2()            const;
		inline const LInsOp3*           getLInsOp3()            const;
		inline const LInsLd*            getLInsLd()             const;
		inline const LInsSt*            getLInsSt()             const;
		inline const LInsSk*            getLInsSk()             const;
		inline const LInsCall*          getLInsCall()           const;
		inline const LInsP*             getLInsP()              const;
		inline const LInsIorF*          getLInsIorF()           const;
		inline const LInsQorD*          getLInsQorD()           const;
		inline const LInsBranch*        getLInsBranch()         const;
		inline const LInsExit*          getLInsExit()           const;
		inline const LInsEntry*         getLInsEntry()          const;
		inline const LInsLabel*         getLInsLabel()          const;
		inline const LInsPhi*           getLInsPhi()            const;
		inline const LInsSpillFence*    getLInsSpillFence()     const;
		inline const LInsRegisterFence* getLInsRegisterFence()  const;
		inline const LInsKillFence*     getLInsKillFence()      const;

	public:
		LIns();
		~LIns();

		NANOJIT_FORCEINLINE void preventOptimize() {
			_allowFolding  = false;
		}

		NANOJIT_FORCEINLINE bool canOptimize() const {
			return _allowFolding;
		}

		NANOJIT_FORCEINLINE void setGenFlags() {
			_genFlags = true;
		}

		NANOJIT_FORCEINLINE bool getGenFlags() {
			return _genFlags;
		}

		NANOJIT_FORCEINLINE Register getCaptureReg() const {
			Register r = { _captureRegIndex };
			return r;
		}

		NANOJIT_FORCEINLINE void setCaptureReg(Register reg) {
			NanoAssert(REGNUM(reg) <= 0xFF);
			_captureRegIndex = static_cast<uint8_t>(REGNUM(reg));
		}

		NANOJIT_FORCEINLINE void setArGroupAndSize(unsigned int group, unsigned int size) {
			NanoAssert(group < LINS_DEFAULT_AR_GROUP);
			NanoAssert((size > 0) && (size < 0xFF));
			arGroup = static_cast<uint16_t>(group);
			arSize = static_cast<uint8_t>(size);
		}

		NANOJIT_FORCEINLINE static unsigned int packArGroupAndSize(unsigned int group, unsigned int size) {
			NanoAssert(group < LINS_DEFAULT_AR_GROUP);
			NanoAssert((size > 0) && (size < 0xFF));
			return (group | (size << 16));
		}

		NANOJIT_FORCEINLINE void setPackedArGroupAndSize(unsigned int packedGroupAndSize) {
			NanoAssert((packedGroupAndSize & 0xFFFF) < LINS_DEFAULT_AR_GROUP);
			NanoAssert(((packedGroupAndSize >> 16) < 255) && ((packedGroupAndSize >> 16) > 0));
			arGroup = static_cast<uint16_t>(packedGroupAndSize);
			arSize = static_cast<uint8_t>(packedGroupAndSize >> 16);
		}

		NANOJIT_FORCEINLINE unsigned int getArGroup() const {
			return arGroup;
		}

		NANOJIT_FORCEINLINE unsigned int getArSize() const {
			return arSize;
		}

		inline void reset() {
			memset(&compilationData, 0, sizeof(compilationData));

			arGroup = LINS_DEFAULT_AR_GROUP;
			arSize = 4;
			_allowFolding = true;
			_genFlags = false;
			setCaptureReg(UnspecifiedReg);
		}

		// LIns initializers.
		inline void         initLInsOp0(LOpcode opcode);
		inline void         initLInsOp1(LOpcode opcode, LIns* oprnd1);
		inline void         initLInsOp2(LOpcode opcode, LIns* oprnd1, LIns* oprnd2);
		inline void         initLInsOp3(LOpcode opcode, LIns* oprnd1, LIns* oprnd2, LIns* oprnd3);
		inline void         initLInsLd(LOpcode opcode, LIns* val, int32_t d, AccSet accSet, LoadQual loadQual);
		inline void         initLInsSt(LOpcode opcode, LIns* val, LIns* base, int32_t d, AccSet accSet);
		// Nb: args[] must be allocated and initialised before being passed in;
		// initLInsC() just copies the pointer into the LInsC.
		inline void         initLInsCall(LOpcode opcode, CallParameters* parameters, const void* address);
		inline void         initLInsP(LOpcode op, Register reg);
		inline void         initLInsI(int32_t immIorF);
		inline void         initLInsF(float immIorF);
#if defined(NANOJIT_64BIT)
		inline void         initLInsQ(int64_t immQorD);
#endif
		inline void         initLInsD(double immIorF);
		inline void         initLInsEntry(RegisterMask saveMask);
		inline void         initLInsExit(ParameterValue* values, unsigned int count, RegisterMask restoreMask, bool isPatchable);
		inline void         initLInsLabel();
		void                initLInsPhi(LOpcode opcode, PhiData* phiData, LIns** values, size_t count, LIns* reconcileNodeNext);
		inline void         initLInsValueFence(LIns** valueArray, size_t valueArraySize);
		inline void         initLInsRegisterFence(LIns** valueArray, size_t valueArraySize);
		inline void         initLInsKillFence(LIns** valueArray, size_t valueArraySize);
		inline void         initLInsBranch(LOpcode op, LIns* condition);

		inline void         duplicateFrom(LIns* insToDuplicate);

		NANOJIT_FORCEINLINE LOpcode opcode() const {
			NanoAssert(op < LIR_sentinel);
			return static_cast<LOpcode>(op);
		}

		// Generally, void instructions (statements) are always live and
		// non-void instructions (expressions) are live if used by another
		// live instruction.  But there are some trickier cases.
		// Any non-void instruction can be marked isResultLive=1 even
		// when it is unreachable, e.g. due to an always-taken branch.
		// The assembler marks it live if it sees any uses, regardless of
		// whether those uses are in reachable code or not.
		NANOJIT_FORCEINLINE bool isLive() const {
			NanoAssert((!isV()) || compilationData.isResultLive);
			return compilationData.isResultLive;
		}

		NANOJIT_FORCEINLINE void setResultLive() {
			compilationData.isResultLive = true;
		}

		NANOJIT_FORCEINLINE void setResultDead() {
			compilationData.isResultLive = false;
		}

		NANOJIT_FORCEINLINE void setFlagRequested(ArithFlag flg) {
			NanoAssert(getGenFlags());
			compilationData.flagsRequested |= flg;
		}

		NANOJIT_FORCEINLINE uint32_t flagRequested() const {
			return compilationData.flagsRequested;
		}

		// Reservation manipulation.
		//
		// "Extant" mean "in existence, still existing, surviving".  In other
		// words, has the value been computed explicitly (not folded into
		// something else) and is it still available (in a register or spill
		// slot) for use?
		NANOJIT_FORCEINLINE bool isExtant() const {
			return (isInReg() || isInAr());
		}
		NANOJIT_FORCEINLINE bool isInReg() const {
			return compilationData.inReg;
		}
		NANOJIT_FORCEINLINE bool isInRegMask(RegisterMask allow) {
			return isInReg() && (rmask(getReg()) & allow);
		}
		NANOJIT_FORCEINLINE Register getReg() const {
			NanoAssert(isInReg());
			Register r = { compilationData.regnum };
			return r;
		}
		NANOJIT_FORCEINLINE void setReg(Register r) {
			NanoAssert(retType() != LTy_V);
			compilationData.inReg = 1;
			compilationData.regnum = static_cast<uint8_t>(REGNUM(r));
		}
		NANOJIT_FORCEINLINE void clearReg() {
			compilationData.inReg = 0;
		}
		NANOJIT_FORCEINLINE bool isInAr() const {
			return compilationData.inAr;
		}
		NANOJIT_FORCEINLINE uint32_t getArIndex() const {
			NanoAssert(isInAr());
			return compilationData.arIndex;
		}
		NANOJIT_FORCEINLINE void setArIndex(uint32_t arIndex) {
			NanoAssert(arIndex < 0x10000);
			compilationData.inAr = 1;
			compilationData.arIndex = static_cast<uint16_t>(arIndex);
		}
		NANOJIT_FORCEINLINE void updateArSlot(uint32_t arIndex) {
			NanoAssert(isInAr());
			NanoAssert(arIndex < 0x10000);
			compilationData.arIndex = static_cast<uint16_t>(arIndex);
		}
		NANOJIT_FORCEINLINE void clearAr() {
			NanoAssert(isInAr());
			compilationData.inAr = 0;
		}

		// For various instruction kinds.
		inline LIns*    oprnd1() const;
		inline LIns*    oprnd2() const;
		inline LIns*    oprnd3() const;


		// For loads.
		inline LoadQual loadQual() const;

		// For loads/stores.
		inline int32_t  disp() const;
		inline MiniAccSet miniAccSet() const;
		inline AccSet   accSet() const;

		// For LInsSk.
		inline LIns*    prevLIns() const;

		// For LInsP.
		inline Register paramReg()  const;

		// For LInsIorF.
		inline int32_t  immI() const;
		inline float    immF() const;
		inline int32_t  immFasI() const;

		// For safepoints.
		inline void*    safePayload() const;

		// For LInsQorD.
#ifdef NANOJIT_64BIT
		inline uint64_t immQ() const;
#endif
		inline double   immD() const;
		inline uint64_t immDasQ() const;

		// For LInsCall.
		inline ParameterValue*  callArgs()          const;  // right-to-left-order: arg(0) is rightmost
		inline uint32_t         callArgc()          const;
		inline const void*      callAddress()       const;
		inline RegisterMask     callVolatileMask()  const;

		// for LInsBranch
		inline LIns*            branchCondition()   const;
		inline LIns*            branchTarget()      const;
		inline void             branchSetTarget(LIns* label);

		//
		void* entryAddress() const;
		void entrySetAddress(void* address);

		void entrySetSaveMask(RegisterMask saveMask);
		RegisterMask entryGetSaveMask() const;

		//
		ParameterValue* exitParamData() const;
		unsigned int exitParamCount() const;

		void exitSetRestoreMask(RegisterMask restoreMask);
		RegisterMask exitGetRestoreMask() const;
		bool exitIsPatchable() const;

		// For LInsLabel
		inline void labelSetReferenced();
		inline bool labelIsReferenced() const;
		inline const NIns* labelNativeAddress() const;
		inline void labelSetNativeAddress(NIns* addr);

		// For LInsPhi
		inline LIns** phiValues() const;
		inline size_t phiCount() const;
		inline LIns* phiReconcileNodeNext() const;
		inline PhiData* phiData() const;
		inline void phiSetData(PhiData* data);
		inline void phiSetValues(LIns** values, size_t count);

		// value fence
		inline LIns** spillFenceArray() const;
		inline size_t spillFenceCount() const;
		inline void spillFenceSetData(LIns** valueArray, size_t valueArraySize);

		// register fence
		inline LIns** registerFenceArray() const;
		inline size_t registerFenceCount() const;
		inline void registerFenceSetData(LIns** valueArray, size_t valueArraySize);

		// kill fence
		inline LIns** killFenceArray() const;
		inline size_t killFenceCount() const;
		inline void killFenceSetData(LIns** valueArray, size_t valueArraySize);


		NANOJIT_FORCEINLINE LInsRepKind repKind() const {
			return static_cast<LInsRepKind>(repKinds[opcode()]);
		}

		// isLInsXYZ() returns true if the instruction has the LInsXYZ form.
		// Note that there is some overlap with other predicates, eg.
		// isStore()==isLInsSt(), isCall()==isLInsC(), but that's ok;  these
		// ones are used mostly to check that opcodes are appropriate for
		// instruction layouts, the others are used for non-debugging
		// purposes.
		NANOJIT_FORCEINLINE bool isLInsOp0() const {
			NanoAssert(LRK_None != repKinds[opcode()]);
			return LRK_Op0 == repKinds[opcode()];
		}
		NANOJIT_FORCEINLINE bool isLInsOp1() const {
			NanoAssert(LRK_None != repKinds[opcode()]);
			return LRK_Op1 == repKinds[opcode()];
		}
		NANOJIT_FORCEINLINE bool isLInsOp1b() const {
			NanoAssert(LRK_None != repKinds[opcode()]);
			return LRK_Op1b == repKinds[opcode()];
		}
		NANOJIT_FORCEINLINE bool isLInsOp2() const {
			NanoAssert(LRK_None != repKinds[opcode()]);
			return LRK_Op2 == repKinds[opcode()];
		}
		NANOJIT_FORCEINLINE bool isLInsOp3() const {
			NanoAssert(LRK_None != repKinds[opcode()]);
			return LRK_Op3 == repKinds[opcode()];
		}
		NANOJIT_FORCEINLINE bool isLInsLd() const {
			NanoAssert(LRK_None != repKinds[opcode()]);
			return LRK_Ld == repKinds[opcode()];
		}
		NANOJIT_FORCEINLINE bool isLInsSt() const {
			NanoAssert(LRK_None != repKinds[opcode()]);
			return LRK_St == repKinds[opcode()];
		}
		NANOJIT_FORCEINLINE bool isLInsSk() const {
			NanoAssert(LRK_None != repKinds[opcode()]);
			return LRK_Sk == repKinds[opcode()];
		}
		NANOJIT_FORCEINLINE bool isLInsC() const {
			NanoAssert(LRK_None != repKinds[opcode()]);
			return LRK_Call ==  repKinds[opcode()];
		}
		NANOJIT_FORCEINLINE bool isLInsP() const {
			NanoAssert(LRK_None != repKinds[opcode()]);
			return LRK_P == repKinds[opcode()];
		}
		NANOJIT_FORCEINLINE bool isLInsIorF() const {
			NanoAssert(LRK_None != repKinds[opcode()]);
			return LRK_IorF == repKinds[opcode()];
		}
		NANOJIT_FORCEINLINE bool isLInsEntry() const    {
			NanoAssert(LRK_None != repKinds[opcode()]);
			return LIR_entry == opcode();
		}
		NANOJIT_FORCEINLINE bool isLInsExit() const {
			NanoAssert(LRK_None != repKinds[opcode()]);
			return LIR_exit == opcode();
		}
		NANOJIT_FORCEINLINE bool isLInsLabel() const {
			NanoAssert(LRK_None != repKinds[opcode()]);
			return LRK_Label == repKinds[opcode()];
		}
		NANOJIT_FORCEINLINE bool isLInsPhi() const {
			NanoAssert(LRK_None != repKinds[opcode()]);
			return LRK_Phi == repKinds[opcode()];
		}
		NANOJIT_FORCEINLINE bool isLInsQorD() const {
			NanoAssert(LRK_None != repKinds[opcode()]);
			return LRK_QorD == repKinds[opcode()];
		}
		NANOJIT_FORCEINLINE bool isLInsBranch() const {
			NanoAssert(LRK_None != repKinds[opcode()]);
			return LRK_Branch == repKinds[opcode()];
		}

		// LIns predicates.
		NANOJIT_FORCEINLINE bool isop(LOpcode o) const {
			return opcode() == o;
		}
		NANOJIT_FORCEINLINE bool isCmp() const {
			return isCmpOpcode(opcode());
		}
		NANOJIT_FORCEINLINE bool isCall() const {
			return isop(LIR_callv) ||
				isop(LIR_calli) ||
#if defined NANOJIT_64BIT
				isop(LIR_callq) ||
#endif
				isop(LIR_callf) ||
				isop(LIR_calld);
		}
		NANOJIT_FORCEINLINE bool isCmov() const {
			return isCmovOpcode(opcode());
		}
		NANOJIT_FORCEINLINE bool isStore() const {
			return isLInsSt();
		}
		NANOJIT_FORCEINLINE bool isLoad() const {
			return isLInsLd();
		}

		NANOJIT_FORCEINLINE bool isCopy() const {
			return 
					isop(LIR_copyi)
#if defined NANOJIT_64BIT
				||	isop(LIR_copyq)
#endif
				||	isop(LIR_copyf) 
				||	isop(LIR_copyd);
		}

		// True if the instruction is a 32-bit integer immediate.
		NANOJIT_FORCEINLINE bool isImmI() const {
			return isop(LIR_immi);
		}
		// True if the instruction is a 32-bit integer immediate and
		// has the value 'val' when treated as a 32-bit signed integer.
		NANOJIT_FORCEINLINE bool isImmI(int32_t val) const {
			return isImmI() && immI()==val;
		}
#ifdef NANOJIT_64BIT
		// True if the instruction is a 64-bit integer immediate.
		NANOJIT_FORCEINLINE bool isImmQ() const {
			return isop(LIR_immq);
		}
#endif
		// True if the instruction is a pointer-sized integer immediate.
		NANOJIT_FORCEINLINE bool isImmP() const
		{
#ifdef NANOJIT_64BIT
			return isImmQ();
#else
			return isImmI();
#endif
		}
		// True if the instruction is a 64-bit float immediate.
		NANOJIT_FORCEINLINE bool isImmD() const {
			return isop(LIR_immd);
		}

		// True if the instruction is a 32-bit float immediate.
		NANOJIT_FORCEINLINE bool isImmF() const {
			return isop(LIR_immf);
		}

		// True if the instruction is a 32-bit integer or float immediate.
		bool isImmIorF() const {
			return isImmI() || isImmF();
		}

		// True if the instruction is a 64-bit integer or double immediate.
		NANOJIT_FORCEINLINE bool isImmQorD() const {
			return
#ifdef NANOJIT_64BIT
				isImmQ() ||
#endif
				isImmD();
		}

		// True if the instruction an any type of immediate.
		bool isImmAny() const {
			return isImmIorF() || isImmQorD();
		}

		NANOJIT_FORCEINLINE bool isConditionalBranch() const {
			return isop(LIR_jt) || isop(LIR_jf);
		}

		NANOJIT_FORCEINLINE bool isUnConditionalBranch() const {
			return isop(LIR_j);
		}

		bool isMullHi() const {
			return 
					isop(LIR_mullshii)
				||  isop(LIR_mulluhii)
#ifdef NANOJIT_64BIT
				|| isop(LIR_mullshiq)
				|| isop(LIR_mullshiq)
#endif
				;
		}

		NANOJIT_FORCEINLINE LTy retType() const {
			return retTypes[opcode()];
		}
		NANOJIT_FORCEINLINE bool isV() const {
			return retType() == LTy_V;
		}
		NANOJIT_FORCEINLINE bool isI() const {
			return (retType() == LTy_I);
		}
#ifdef NANOJIT_64BIT
		NANOJIT_FORCEINLINE bool isQ() const {
			return retType() == LTy_Q;
		}
#endif
		NANOJIT_FORCEINLINE bool isIorQ() const {
			return (
					isI()
#ifdef NANOJIT_64BIT
				||	isQ()
#endif
				);
		}
		NANOJIT_FORCEINLINE bool isD() const {
			return retType() == LTy_D;
		}
		NANOJIT_FORCEINLINE bool isF() const {
			return retType() == LTy_F;
		}
		NANOJIT_FORCEINLINE bool isForD() const {
			return isF() || isD();
		}
		NANOJIT_FORCEINLINE bool isQorD() const {
			return
#ifdef NANOJIT_64BIT
				isQ() ||
#endif
				isD();
		}
		NANOJIT_FORCEINLINE bool isP() const {
#ifdef NANOJIT_64BIT
			return isQ();
#else
			return isI();
#endif
		}

		NANOJIT_FORCEINLINE void* immP() const {
#ifdef NANOJIT_64BIT
			return (void*)immQ();
#else
			return (void*)immI();
#endif
		}

	private:
		// forbid copy
		LIns(const LIns& other);
		LIns& operator=(const LIns& other);
	};

	typedef SeqBuilder<LIns*> InsList;


	NANOJIT_FORCEINLINE LInsOp0*           LIns::getLInsOp0()           { return &data.Op0; }
	NANOJIT_FORCEINLINE LInsOp1*           LIns::getLInsOp1()           { return &data.Op1; }
	NANOJIT_FORCEINLINE LInsOp2*           LIns::getLInsOp2()           { return &data.Op2; }
	NANOJIT_FORCEINLINE LInsOp3*           LIns::getLInsOp3()           { return &data.Op3; }
	NANOJIT_FORCEINLINE LInsLd*            LIns::getLInsLd()            { return &data.Ld; }
	NANOJIT_FORCEINLINE LInsSt*            LIns::getLInsSt()            { return &data.St; }
	NANOJIT_FORCEINLINE LInsSk*            LIns::getLInsSk()            { return &data.Sk; }
	NANOJIT_FORCEINLINE LInsCall*          LIns::getLInsCall()          { return &data.Call; }
	NANOJIT_FORCEINLINE LInsP*             LIns::getLInsP()             { return &data.P; }
	NANOJIT_FORCEINLINE LInsIorF*          LIns::getLInsIorF()          { return &data.IorF; }
	NANOJIT_FORCEINLINE LInsQorD*          LIns::getLInsQorD()          { return &data.QorD; }
	NANOJIT_FORCEINLINE LInsBranch*        LIns::getLInsBranch()        { return &data.Branch; }
	NANOJIT_FORCEINLINE LInsExit*          LIns::getLInsExit()          { return &data.Exit; }
	NANOJIT_FORCEINLINE LInsEntry*         LIns::getLInsEntry()         { return &data.Entry; }
	NANOJIT_FORCEINLINE LInsLabel*         LIns::getLInsLabel()         { return &data.Label; }
	NANOJIT_FORCEINLINE LInsPhi*           LIns::getLInsPhi()           { return &data.Phi; }
	NANOJIT_FORCEINLINE LInsSpillFence*    LIns::getLInsSpillFence()    { return &data.SpillFence; }
	NANOJIT_FORCEINLINE LInsRegisterFence* LIns::getLInsRegisterFence() { return &data.RegFence; }
	NANOJIT_FORCEINLINE LInsKillFence*     LIns::getLInsKillFence()     { return &data.KillFence; }

	NANOJIT_FORCEINLINE const LInsOp0*           LIns::getLInsOp0()           const   { return &data.Op0; }
	NANOJIT_FORCEINLINE const LInsOp1*           LIns::getLInsOp1()           const   { return &data.Op1; }
	NANOJIT_FORCEINLINE const LInsOp2*           LIns::getLInsOp2()           const   { return &data.Op2; }
	NANOJIT_FORCEINLINE const LInsOp3*           LIns::getLInsOp3()           const   { return &data.Op3; }
	NANOJIT_FORCEINLINE const LInsLd*            LIns::getLInsLd()            const   { return &data.Ld; }
	NANOJIT_FORCEINLINE const LInsSt*            LIns::getLInsSt()            const   { return &data.St; }
	NANOJIT_FORCEINLINE const LInsSk*            LIns::getLInsSk()            const   { return &data.Sk; }
	NANOJIT_FORCEINLINE const LInsCall*          LIns::getLInsCall()          const   { return &data.Call; }
	NANOJIT_FORCEINLINE const LInsP*             LIns::getLInsP()             const   { return &data.P; }
	NANOJIT_FORCEINLINE const LInsIorF*          LIns::getLInsIorF()          const   { return &data.IorF; }
	NANOJIT_FORCEINLINE const LInsQorD*          LIns::getLInsQorD()          const   { return &data.QorD; }
	NANOJIT_FORCEINLINE const LInsBranch*        LIns::getLInsBranch()        const   { return &data.Branch; }
	NANOJIT_FORCEINLINE const LInsExit*          LIns::getLInsExit()          const   { return &data.Exit; }
	NANOJIT_FORCEINLINE const LInsEntry*         LIns::getLInsEntry()         const   { return &data.Entry; }
	NANOJIT_FORCEINLINE const LInsLabel*         LIns::getLInsLabel()         const   { return &data.Label; }
	NANOJIT_FORCEINLINE const LInsPhi*           LIns::getLInsPhi()           const   { return &data.Phi; }
	NANOJIT_FORCEINLINE const LInsSpillFence*    LIns::getLInsSpillFence()    const   { return &data.SpillFence; }
	NANOJIT_FORCEINLINE const LInsRegisterFence* LIns::getLInsRegisterFence() const   { return &data.RegFence; }
	NANOJIT_FORCEINLINE const LInsKillFence*     LIns::getLInsKillFence()     const   { return &data.KillFence; }

	void LIns::initLInsOp0(LOpcode opcode) {
		initSharedFields(opcode);
		NanoAssert(isLInsOp0());
	}
	void LIns::initLInsOp1(LOpcode opcode, LIns* oprnd1) {
		initSharedFields(opcode);
		getLInsOp1()->oprnd_1 = oprnd1;
		NanoAssert(isLInsOp1());
	}
	void LIns::initLInsOp2(LOpcode opcode, LIns* oprnd1, LIns* oprnd2) {
		initSharedFields(opcode);
		getLInsOp2()->oprnd_1 = oprnd1;
		getLInsOp2()->oprnd_2 = oprnd2;
		NanoAssert(isLInsOp2());
	}
	void LIns::initLInsOp3(LOpcode opcode, LIns* oprnd1, LIns* oprnd2, LIns* oprnd3) {
		initSharedFields(opcode);
		getLInsOp3()->oprnd_1 = oprnd1;
		getLInsOp3()->oprnd_2 = oprnd2;
		getLInsOp3()->oprnd_3 = oprnd3;
		NanoAssert(isLInsOp3());
	}
	void LIns::initLInsLd(LOpcode opcode, LIns* val, int32_t d, AccSet accSet, LoadQual loadQual) {
		initSharedFields(opcode);
		getLInsLd()->oprnd_1 = val;
		NanoAssert(d == int16_t(d));
		getLInsLd()->disp = int16_t(d);
		getLInsLd()->miniAccSetVal = compressAccSet(accSet).val;
		getLInsLd()->loadQual = static_cast<uint8_t>(loadQual);
		NanoAssert((static_cast<int>(loadQual) < 4) && (static_cast<int>(loadQual) >= 0));
		NanoAssert(isLInsLd());
	}
	void LIns::initLInsSt(LOpcode opcode, LIns* val, LIns* base, int32_t d, AccSet accSet) {
		initSharedFields(opcode);
		getLInsSt()->oprnd_1 = val;
		getLInsSt()->oprnd_2 = base;
		NanoAssert(d == int16_t(d));
		getLInsSt()->disp = int16_t(d);
		getLInsSt()->miniAccSetVal = compressAccSet(accSet).val;
		NanoAssert(isLInsSt());
	}
	void LIns::initLInsCall(LOpcode opcode, CallParameters* parameters, const void* address) {
		initSharedFields(opcode);
		getLInsCall()->parameters = parameters;
		getLInsCall()->address = address;
		NanoAssert(isLInsC());
	}
	void LIns::initLInsP(LOpcode op, Register reg) {
		initSharedFields(op);
		getLInsP()->reg = reg;
		NanoAssert(isLInsP());
	}
	void LIns::initLInsI(int32_t immI) {
		initSharedFields(LIR_immi);
		getLInsIorF()->asImmI = immI;
		NanoAssert(isLInsIorF());
	}
	void LIns::initLInsF(float immF) {
		initSharedFields(LIR_immf);
		getLInsIorF()->asImmF = immF;
		NanoAssert(isLInsIorF());
	}
#if defined(NANOJIT_64BIT)
	void LIns::initLInsQ(int64_t immQ) {
		initSharedFields(LIR_immq);
		getLInsQorD()->asImmQ = immQ;
		NanoAssert(isLInsQorD());
	}
#endif
	void LIns::initLInsD(double immD) {
		initSharedFields(LIR_immd);
		getLInsQorD()->asImmD = immD;
		NanoAssert(isLInsQorD());
	}
	void LIns::initLInsEntry(RegisterMask saveMask) {
		initSharedFields(LIR_entry);
		getLInsEntry()->address = NULL;
		getLInsEntry()->saveMask = saveMask;
		NanoAssert(isLInsEntry());
	}
	void LIns::initLInsExit(ParameterValue* values, unsigned int count, RegisterMask restoreMask, bool isPatchable) {
		initSharedFields(LIR_exit);
		getLInsExit()->values       = values;
		getLInsExit()->count        = count;
		getLInsExit()->restoreMask  = restoreMask;
		getLInsExit()->isPatchable  = isPatchable;
		NanoAssert(isLInsExit());
	}
	void LIns::initLInsLabel() {
		initSharedFields(LIR_label);
		getLInsLabel()->nativeAddress = 0;
		getLInsLabel()->referenced = false;
		NanoAssert(isLInsLabel());
	}
	void LIns::initLInsValueFence(LIns** valueArray, size_t valueArraySize) {
		initSharedFields(LIR_spillfence);
		getLInsSpillFence()->valueArray = valueArray;
		getLInsSpillFence()->valueArraySize = valueArraySize;
	}
	void LIns::initLInsRegisterFence(LIns** valueArray, size_t valueArraySize) {
		initSharedFields(LIR_regfence);
		getLInsRegisterFence()->valueArray = valueArray;
		getLInsRegisterFence()->valueArraySize = valueArraySize;
	}
	void LIns::initLInsKillFence(LIns** valueArray, size_t valueArraySize) {
		initSharedFields(LIR_killfence);
		getLInsKillFence()->valueArray = valueArray;
		getLInsKillFence()->valueArraySize = valueArraySize;
	}
	void LIns::initLInsBranch(LOpcode op, LIns* condition) {
		initSharedFields(op);
		getLInsBranch()->condition = condition;
		getLInsBranch()->target = NULL;
		NanoAssert(isLInsBranch());
	}

	void LIns::duplicateFrom(LIns* insToDuplicate) {
		NanoAssert(!insToDuplicate->isLive());
		NanoAssert(!insToDuplicate->isInAr());
		NanoAssert(!insToDuplicate->isInReg());

		op              = insToDuplicate->op;
		data            = insToDuplicate->data;
	}

	NANOJIT_FORCEINLINE LIns* LIns::oprnd1() const {
		NanoAssert(isLInsOp1() || isLInsOp1b() || isLInsOp2() || isLInsOp3() ||
				   isLInsLd() || isLInsSt() || isLInsExit());
		return getLInsOp1()->oprnd_1;
	}
	NANOJIT_FORCEINLINE LIns* LIns::oprnd2() const {
		NanoAssert(isLInsOp2() || isLInsOp3() || isLInsSt());
		return getLInsOp2()->oprnd_2;
	}
	NANOJIT_FORCEINLINE LIns* LIns::oprnd3() const {
		NanoAssert(isLInsOp3());
		return getLInsOp3()->oprnd_3;
	}
	

	inline void* LIns::entryAddress() const {
		NanoAssert(isLInsEntry());
		return getLInsEntry()->address;
	}

	inline void LIns::entrySetAddress(void* address) {
		NanoAssert(isLInsEntry());
		getLInsEntry()->address = address;
	}

	inline void LIns::entrySetSaveMask(RegisterMask saveMask) {
		NanoAssert(isLInsEntry());
		getLInsEntry()->saveMask = saveMask;
	}

	inline RegisterMask LIns::entryGetSaveMask() const {
		NanoAssert(isLInsEntry());
		return getLInsEntry()->saveMask;
	}

	NANOJIT_FORCEINLINE ParameterValue * LIns::exitParamData() const {
		NanoAssert(isLInsExit());
		return getLInsExit()->values;
	}

	NANOJIT_FORCEINLINE unsigned int LIns::exitParamCount() const {
		NanoAssert(isLInsExit());
		return getLInsExit()->count;
	}


	NANOJIT_FORCEINLINE void LIns::exitSetRestoreMask(RegisterMask restoreMask) {
		NanoAssert(isLInsExit());
		getLInsExit()->restoreMask = restoreMask;
	}

	NANOJIT_FORCEINLINE RegisterMask LIns::exitGetRestoreMask() const {
		NanoAssert(isLInsExit());
		return getLInsExit()->restoreMask;
	}

	NANOJIT_FORCEINLINE bool LIns::exitIsPatchable() const {
		NanoAssert(isLInsExit());
		return getLInsExit()->isPatchable;
	}

	inline void LIns::labelSetReferenced() {
		NanoAssert(isLInsLabel());
		LInsLabel* insLabel = getLInsLabel();
		insLabel->referenced = true;
	}

	NANOJIT_FORCEINLINE bool LIns::labelIsReferenced() const {
		NanoAssert(isLInsLabel());
		const LInsLabel* insLabel = getLInsLabel();
		return insLabel->referenced;
	}

	NANOJIT_FORCEINLINE const NIns* LIns::labelNativeAddress() const {
		NanoAssert(isLInsLabel());
		const LInsLabel* insLabel = getLInsLabel();
		return insLabel->nativeAddress;
	}

	inline void LIns::labelSetNativeAddress(NIns* addr) {
		NanoAssert(isLInsLabel());
		LInsLabel* insLabel = getLInsLabel();
		insLabel->nativeAddress = addr;
	}

	NANOJIT_FORCEINLINE LIns** LIns::phiValues() const {
		NanoAssert(isLInsPhi());
		return getLInsPhi()->values;
	}

	NANOJIT_FORCEINLINE size_t LIns::phiCount() const {
		NanoAssert(isLInsPhi());
		return getLInsPhi()->count;
	}

	NANOJIT_FORCEINLINE LIns* LIns::phiReconcileNodeNext() const {
		NanoAssert(isLInsPhi());
		return getLInsPhi()->reconcileNodeNext;
	}

	NANOJIT_FORCEINLINE PhiData* LIns::phiData() const {
		NanoAssert(isLInsPhi());
		return getLInsPhi()->phiData;
	}

	inline void LIns::phiSetData(PhiData* phiData) {
		NanoAssert(isLInsPhi());
		NanoAssert(getLInsPhi()->phiData == NULL);
		getLInsPhi()->phiData = phiData;

#if defined(NANOJIT_DEBUG)
		if ((phiData != NULL) && (phiCount() > 0)) {
			for (unsigned int i = 0; i < phiCount(); ++i)
				NanoAssert(!phiValues()[i]->canOptimize());
		}
#endif
	}

	inline void LIns::phiSetValues(LIns** values, size_t count) {
		NanoAssert(isLInsPhi());
		NanoAssert(getLInsPhi()->values == NULL);
		NanoAssert(getLInsPhi()->count == 0);
		getLInsPhi()->values = values;
		getLInsPhi()->count = count;
	}

	NANOJIT_FORCEINLINE LIns** LIns::spillFenceArray() const {
		NanoAssert(isop(LIR_spillfence));
		return getLInsSpillFence()->valueArray;
	}

	NANOJIT_FORCEINLINE size_t LIns::spillFenceCount() const {
		NanoAssert(isop(LIR_spillfence));
		return getLInsSpillFence()->valueArraySize;
	}

	inline void LIns::spillFenceSetData(LIns** valueArray, size_t valueArraySize) {
		NanoAssert(isop(LIR_spillfence));
		getLInsSpillFence()->valueArray = valueArray;
		getLInsSpillFence()->valueArraySize = valueArraySize;
	}

	// register fence
	NANOJIT_FORCEINLINE LIns** LIns::registerFenceArray() const {
		NanoAssert(isop(LIR_regfence));
		return getLInsRegisterFence()->valueArray;
	}
	NANOJIT_FORCEINLINE size_t LIns::registerFenceCount() const {
		NanoAssert(isop(LIR_regfence));
		return getLInsRegisterFence()->valueArraySize;
	}

	NANOJIT_FORCEINLINE void LIns::registerFenceSetData(LIns** valueArray, size_t valueArraySize) {
		NanoAssert(isop(LIR_regfence));
		getLInsRegisterFence()->valueArray = valueArray;
		getLInsRegisterFence()->valueArraySize = valueArraySize;
	}

	// kill fence
	NANOJIT_FORCEINLINE LIns** LIns::killFenceArray() const {
		NanoAssert(isop(LIR_killfence));
		return getLInsKillFence()->valueArray;
	}
	NANOJIT_FORCEINLINE size_t LIns::killFenceCount() const {
		NanoAssert(isop(LIR_killfence));
		return getLInsKillFence()->valueArraySize;
	}

	NANOJIT_FORCEINLINE void LIns::killFenceSetData(LIns** valueArray, size_t valueArraySize) {
		NanoAssert(isop(LIR_killfence));
		getLInsKillFence()->valueArray = valueArray;
		getLInsKillFence()->valueArraySize = valueArraySize;
	}


	NANOJIT_FORCEINLINE LoadQual LIns::loadQual() const {
		NanoAssert(isLInsLd());
		return (LoadQual)getLInsLd()->loadQual;
	}

	int32_t LIns::disp() const {
		if (isLInsSt()) {
			return getLInsSt()->disp;
		} else {
			NanoAssert(isLInsLd());
			return getLInsLd()->disp;
		}
	}

	MiniAccSet LIns::miniAccSet() const {
		MiniAccSet miniAccSet;
		if (isLInsSt()) {
			miniAccSet.val = getLInsSt()->miniAccSetVal;
		} else {
			NanoAssert(isLInsLd());
			miniAccSet.val = getLInsLd()->miniAccSetVal;
		}
		return miniAccSet;
	}

	NANOJIT_FORCEINLINE AccSet LIns::accSet() const {
		return decompressMiniAccSet(miniAccSet());
	}

	LIns* LIns::prevLIns() const {
		NanoAssert(isLInsSk());
		return getLInsSk()->prevLIns;
	}

	NANOJIT_FORCEINLINE Register		LIns::paramReg() const			{ NanoAssert(isLInsP()); return getLInsP()->reg; }

	NANOJIT_FORCEINLINE int32_t			LIns::immI() const				{ NanoAssert(isImmI()); return getLInsIorF()->asImmI; }
	NANOJIT_FORCEINLINE int32_t			LIns::immFasI() const			{ NanoAssert(isImmF()); return getLInsIorF()->asImmI; }
	NANOJIT_FORCEINLINE float			LIns::immF() const				{ NanoAssert(isImmF()); return getLInsIorF()->asImmF; }

#ifdef NANOJIT_64BIT
	NANOJIT_FORCEINLINE uint64_t		LIns::immQ() const				{ NanoAssert(isImmQ()); return getLInsQorD()->asImmQ; }
#endif
	NANOJIT_FORCEINLINE double			LIns::immD() const				{ NanoAssert(isImmD()); return getLInsQorD()->asImmD; }
	NANOJIT_FORCEINLINE uint64_t		LIns::immDasQ() const			{ NanoAssert(isImmD()); return getLInsQorD()->asImmQ; }

	NANOJIT_FORCEINLINE ParameterValue* LIns::callArgs() const {
		NanoAssert(isLInsC());
		return getLInsCall()->parameters->values;
	}

	NANOJIT_FORCEINLINE uint32_t LIns::callArgc() const {
		NanoAssert(isLInsC());
		return getLInsCall()->parameters->count;
	}

	NANOJIT_FORCEINLINE const void* LIns::callAddress() const {
		NanoAssert(isLInsC());
		return getLInsCall()->address;
	}

	NANOJIT_FORCEINLINE RegisterMask LIns::callVolatileMask() const {
		NanoAssert(isLInsC());
		return getLInsCall()->parameters->volatileMask;
	}

	// for LInsBranch
	NANOJIT_FORCEINLINE LIns* LIns::branchCondition() const {
		NanoAssert(isLInsBranch());
		return getLInsBranch()->condition;
	}

	NANOJIT_FORCEINLINE LIns* LIns::branchTarget() const {
		NanoAssert(isLInsBranch());
		return getLInsBranch()->target;
	}

	inline void LIns::branchSetTarget(LIns* label) {
		NanoAssert(label && label->isLInsLabel());
		NanoAssert(isLInsBranch());
		NanoAssert(branchTarget() == NULL);
		getLInsBranch()->target = label;

		if (label != NULL)
			label->labelIsReferenced();
	}

#ifdef NANOJIT_VERBOSE
	const char* getLirName(LOpcode opcode);

	// Maps address ranges to meaningful names.
	class AddrNameMap
	{
		Allocator* allocator;
		class Entry
		{
		public:
			Entry(int) : name(0), size(0), align(0) {}
			Entry(char *n, size_t s, size_t a) : name(n), size(s), align(a) {}
			char* name;
			size_t size:29, align:3;
		};
		TreeMap<const void*, Entry*> names;     // maps code regions to names
	public:
		AddrNameMap(Allocator* allocator);
		void clear() { names.clear(); }
		void addAddrRange(const void *p, size_t size, size_t align, const char *name);
		void lookupAddr(void *p, char*& name, int32_t& offset);
	};

#ifdef _MSC_VER
#pragma warning(disable:4996) // disable "strcpy is unsafe" warnings in Visual Studio, they break the build
#endif
	// Maps LIR instructions to meaningful names.
	class LirNameMap
	{
	private:
		Allocator* alloc;

		// A small string-wrapper class, required because we need '==' to
		// compare string contents, not string pointers, when strings are used
		// as keys in CountMap.
		struct Str {
			Allocator* alloc;
			char* s;

			Str(Allocator* alloc_, const char* s_) : alloc(alloc_) {
				s = new (alloc) char[1+strlen(s_)];
				strcpy(s, s_);
			}

			bool operator==(const Str& str) const {
				return (0 == strcmp(this->s, str.s));
			}

			bool operator!=(const Str& str) const {
				return !(*this == str);
			}

			static size_t hash(const Str &k) {
				// (const void*) cast is required by ARM RVCT 2.2
				return murmurhash((const void*)k.s, strlen(k.s));
			}

			// default constructor for compatibiltiy with alignof(), never call.
			Str() : alloc(NULL) {}
		};

		template <class Key, class H=DefaultHash<Key>, class E=DefaultKeysEqual<Key> >
		class CountMap: public HashMap<Key, int, 128, H, E> {
		public:
			CountMap(Allocator* alloc) : HashMap<Key, int, 128, H, E>(alloc) {}
			int add(Key k) {
				int c = 1;
				if (this->containsKey(k)) {
					c = 1 + this->get(k);
				}
				this->put(k,c);
				return c;
			}
		};

		CountMap<int> lircounts;
		CountMap<Str, Str, UseOperEqual<Str> > namecounts;

		void addNameWithSuffix(LIns* i, const char *s, int suffix, bool ignoreOneSuffix);

		class Entry
		{
		public:
			Entry(int) : name(0) {}
			Entry(char* n) : name(n) {}
			char* name;
		};

		HashMap<LIns*, Entry*, 16> names;

	public:
		LirNameMap(Allocator* alloc)
			: alloc(alloc),
			lircounts(alloc),
			namecounts(alloc),
			names(alloc)
		{}

		void        clear() { names.clear(); lircounts.clear(); namecounts.clear(); }
		void        addName(LIns* ins, const char *s);  // gives 'ins' a special name
		const char* createName(LIns* ins);              // gives 'ins' a generic name
		const char* lookupName(LIns* ins);
	};

	// We use big buffers for cases where we need to fit a whole instruction,
	// and smaller buffers for all the others.  These should easily be long
	// enough, but for safety the formatXyz() functions check and won't exceed
	// those limits.
	class InsBuf {
	public:
		static const size_t len = 2048;
		char buf[len];
	};
	class RefBuf {
	public:
		static const size_t len = 512;
		char buf[len];
	};



#endif /* NANOJIT_VERBOSE */


}
#endif // __nanojit_LIR__
