
#include "nanojit.h"

namespace nanojit
{


/*static*/ LIns* const AR::BAD_ENTRY = (LIns*)(0x0); //baadf00ddeadbeefULL);

#ifdef NANOJIT_DEBUG


void AR::validateQuick() {
	NanoAssert((_stackTop == NANOJIT_MAX_STACK_ENTRY) || (_entries[_stackTop] == BAD_ENTRY));
}

void AR::validateFull() {
	validateQuick();

	unsigned int i = 0;

	while (i < _stackTop) {
		LIns* ins = _entries[i];
		if (ins != BAD_ENTRY) {
			unsigned int slotCount = nStackSlotsFor(ins);

			for (unsigned int slot = 1; slot < slotCount; ++slot) {
				NanoAssert(_entries[i] == ins);
			}

			i += slotCount;
		} else {
			++i;
		}
	}
}

void AR::validate() {
	static unsigned int validateCounter = 0;
	if (++validateCounter >= 100) {
		validateFull();
		validateCounter = 0;
	} else {
		validateQuick();
	}
}

bool AR::isValidEntry(unsigned int arIndex, LIns* ins) const {
	NanoAssert(arIndex < _stackTop);

	if (arIndex < _stackTop) {
		return _entries[arIndex] == ins;
	}

	return false;
}

void AR::checkForResourceConsistency(const RegAlloc& regs) {
	validate();
	unsigned int iStackSlot = 0;
	while (iStackSlot < _stackTop) {
		LIns* ins = _entries[iStackSlot];
		if (ins == BAD_ENTRY) {
			iStackSlot++;
			continue;
		}

		unsigned int arIndex = ins->getArIndex();
		unsigned int arSlotCount = nStackSlotsFor(ins);
		NanoAssert(arSlotCount > 0);
		NanoAssert(arIndex < _stackTop);
		NanoAssertMsg(arIndex == iStackSlot, "Stack record index mismatch");

		for (unsigned i = 1; i < arSlotCount; ++i)
			NanoAssert(_entries[iStackSlot + i] == ins);
		NanoAssertMsg(!ins->isInReg() || regs.isConsistent(ins->getReg(), ins),
			"Register   record mismatch");

		iStackSlot += arSlotCount; // skip unused slots
	}
}

void AR::checkCustomArAllocations() const {

	for (unsigned i = _firstEntryAllocationIndex; i < _stackTop; ++i)
		NanoAssert(_entries[i] == BAD_ENTRY);
}


#endif

void AR::reset(unsigned int firstEntryAllocationIndex) {
	_firstEntryAllocationIndex = firstEntryAllocationIndex;
	_stackTop = firstEntryAllocationIndex;
	for (unsigned int i = 0; i < NANOJIT_MAX_STACK_ENTRY; ++i)
		_entries[i] = BAD_ENTRY;
}

bool AR::Iter::next(LIns*& ins, unsigned int& nStackSlots, int& arIndex) {
	while (_i < _ar._stackTop) {
		ins = _ar._entries[_i];
		if (ins != BAD_ENTRY) {
			arIndex = _i;
			nStackSlots = nStackSlotsFor(ins);
			_i += nStackSlots;
			return true;
		}
		_i++;
	}

	ins = NULL;
	nStackSlots = 0;
	arIndex = 0;
	return false;
}

void AR::releaseEntryAt(unsigned int arIndex) {
	NanoAssert(arIndex < _stackTop);

	LIns* i = _entries[arIndex];
	NanoAssert(i != BAD_ENTRY);

#if defined(NANOJIT_DEBUG)	
	unsigned int slotCount = nStackSlotsFor(i);
	NanoAssert(slotCount > 0);
	for (unsigned int slot = 1; slot < slotCount; ++ slot) {
		NanoAssert((arIndex + slot) < _stackTop);
		_entries[arIndex + slot] = BAD_ENTRY;
	}
#endif

	_entries[arIndex] = BAD_ENTRY;

	// spill slot re-use
	if (arIndex >= _firstEntryAllocationIndex) {
		unsigned int top = _stackTop;
		unsigned int slotCount = nStackSlotsFor(i);

		if ((arIndex + slotCount) == top) {
			top = arIndex;

			while (top > _firstEntryAllocationIndex) {
				top--;

				LIns* entry = _entries[top];
				if (entry != BAD_ENTRY) {
					top = entry->getArIndex() + nStackSlotsFor(entry);
					break;
				}
			}
			_stackTop = top;
		}
	}

}

inline bool AR::isEmptyRange(unsigned int start, unsigned int nStackSlots) const {
	NanoAssert((start + nStackSlots) < _stackTop);

	for (unsigned int i = 0; i < nStackSlots; i++) {
		if (_entries[start + i] != BAD_ENTRY)
			return false;
	}
	return true;
}

unsigned int AR::reserveEntry(LIns* ins) {
	NanoAssert(ins->getArGroup() == LINS_DEFAULT_AR_GROUP);

	unsigned int nStackSlots = nStackSlotsFor(ins);
#if defined(NANOJIT_DEBUG)
	NanoAssert(_entries[_stackTop] == BAD_ENTRY);
#endif
	NanoAssert(ins->isInAr() == false);
	NanoAssert(nStackSlots > 0);
	NanoAssert((nStackSlots & (nStackSlots - 1)) == 0); // is power of 2 ?

	unsigned int const		alignMinus1		= nStackSlots - 1;
	unsigned int const		alignedTop		= (_stackTop + alignMinus1) & ~alignMinus1;
	unsigned int const		spaceLeft		= NANOJIT_MAX_STACK_ENTRY - alignedTop;

	if (spaceLeft < nStackSlots) {
		NanoAssert(0);
		return 0xFFFFFFFF; // no more stack space ...
	}

	_stackTop = alignedTop + nStackSlots;

#if defined(NANOJIT_DEBUG)
	NanoAssert(_entries[alignedTop] == BAD_ENTRY);
#endif
	_entries[alignedTop] = ins;

#if defined(NANOJIT_DEBUG)
	for (unsigned int j = 1; j < nStackSlots; j++) {
		NanoAssert(_entries[alignedTop + j] == BAD_ENTRY);
		_entries[alignedTop + j] = ins;
	}
#endif

	NanoAssert(alignedTop >= _firstEntryAllocationIndex);
	return alignedTop;
}

bool AR::reallocateEntry(LIns* ins, unsigned int arIndex)
{
	unsigned int slotCount = nStackSlotsFor(ins);

	NanoAssert(!ins->isInAr());
	NanoAssert((ins->getArGroup() == LINS_DEFAULT_AR_GROUP) || (ins->getArGroup() == arIndex));
	NanoAssert(slotCount > 0);
	NanoAssert(_entries[arIndex] == BAD_ENTRY);

	_entries[arIndex] = ins;

	if (_stackTop < (arIndex + slotCount))
		_stackTop = arIndex + slotCount;

#if defined(NANOJIT_DEBUG)
	unsigned int const nStackSlots = nStackSlotsFor(ins);
	for (unsigned int slotIndex = 1; slotIndex < nStackSlots; ++slotIndex) {
		unsigned int insEntryIndex = arIndex + slotIndex;
		NanoAssert(_entries[insEntryIndex] == BAD_ENTRY);
		_entries[insEntryIndex] = ins;
	}
#endif

	return true;
}

#if defined(NANOJIT_VERBOSE) && defined(NANOJIT_DEBUG)
void AR::checkForResourceLeaks(LInsPrinter* printer, LogControl* logc) const
{
	for (unsigned int i = 0; i < _stackTop; i++) {
		InsBuf buf;
		if (_entries[i] != BAD_ENTRY) {
			printer->formatIns(&buf, _entries[i]);
			logc->printf("frame entry %d wasn't freed %s\n", 4*i, buf.buf);
		}

		NanoAssertMsgf(_entries[i] == BAD_ENTRY, "frame entry %d wasn't freed\n",4*i);
	}
}
#endif


} // namespace nanojit

