
#if !defined(__nanojit_IntrusiveLinkedList__)
#define __nanojit_IntrusiveLinkedList__

// note: can only be used with simple types (no virtual/multiple inheritance)
// because the implementation rely on pointer arithmetic

#include "NanojitPortability.h"
#include <stdint.h>

namespace nanojit
{


struct IntrusiveListNode
{
	IntrusiveListNode*      m_pNext;
	IntrusiveListNode*      m_pPrevious;
};

class IntrusiveLinkedListIterator;
class IntrusiveLinkedListReverseIterator;

class IntrusiveLinkedList
{

	IntrusiveListNode       m_oRoot;

public:
	typedef IntrusiveLinkedListIterator Iterator;
	typedef IntrusiveLinkedListReverseIterator ReverseIterator;

							IntrusiveLinkedList();
							~IntrusiveLinkedList();

	void                    clear();

	void                    insertHead(IntrusiveListNode* a_pNode);
	void                    insertTail(IntrusiveListNode* a_pNode);

	void                    insertBefore(Iterator& a_oIt, IntrusiveListNode* a_pNode);
	void                    insertBefore(ReverseIterator& a_oIt, IntrusiveListNode* a_pNode);


	void                    insertAfter(Iterator& a_oIt, IntrusiveListNode* a_pNode);
	void                    insertAfter(ReverseIterator& a_oIt, IntrusiveListNode* a_pNode);


	void                    remove(const Iterator& a_oIt);

	Iterator                getBegin();
	Iterator                getEnd();
	template<typename T>
	static Iterator         getIteratorAt(T* a_pValue);


	ReverseIterator         getReverseBegin();
	ReverseIterator         getReverseEnd();
	template<typename T>
	static ReverseIterator  getReverseIteratorAt(T* a_pValue);


private:
	void                    insert(IntrusiveListNode* a_pNode, IntrusiveListNode* a_pElementBefore, IntrusiveListNode* a_pElementAfter);
	void                    remove(IntrusiveListNode* a_pNode);

};

/*****************************************************************************************************************************/
/*****************************************************************************************************************************/

class IntrusiveLinkedListIterator
{
private:
	friend class IntrusiveLinkedList;
	IntrusiveListNode*      m_pNode;

private:
					IntrusiveLinkedListIterator(IntrusiveListNode* a_pNode);

public:
					IntrusiveLinkedListIterator(const IntrusiveLinkedListIterator& a_oOther);
					~IntrusiveLinkedListIterator();

	void            moveNext();

	template<typename T>
	T*              getElement();

	bool            operator==(const IntrusiveLinkedListIterator& a_oOther) const;
	bool            operator!=(const IntrusiveLinkedListIterator& a_oOther) const;
};


/*****************************************************************************************************************************/
/*****************************************************************************************************************************/


class IntrusiveLinkedListReverseIterator
{
private:
	friend class IntrusiveLinkedList;
	IntrusiveListNode*      m_pNode;

private:
					IntrusiveLinkedListReverseIterator(IntrusiveListNode* a_pNode);

public:
					IntrusiveLinkedListReverseIterator(const IntrusiveLinkedListReverseIterator& a_oOther);
					~IntrusiveLinkedListReverseIterator();

	void            moveNext();

	template<typename T>
	T*              getElement();

	bool            operator==(const IntrusiveLinkedListReverseIterator& a_oOther) const;
	bool            operator!=(const IntrusiveLinkedListReverseIterator& a_oOther) const;
};

/*****************************************************************************************************************************/
/*****************************************************************************************************************************/


NANOJIT_FORCEINLINE IntrusiveLinkedList::Iterator IntrusiveLinkedList::getBegin()
{
	return Iterator(m_oRoot.m_pNext);
}

NANOJIT_FORCEINLINE IntrusiveLinkedList::Iterator IntrusiveLinkedList::getEnd()
{
	return Iterator(&m_oRoot);
}

template<typename T>
NANOJIT_FORCEINLINE IntrusiveLinkedList::Iterator IntrusiveLinkedList::getIteratorAt(T* a_pValue)
{
	IntrusiveListNode* pNode = reinterpret_cast<IntrusiveListNode*>(reinterpret_cast<uint8_t*>(a_pValue) + T::NODE_OFFSET);
	return Iterator(pNode);
}

NANOJIT_FORCEINLINE IntrusiveLinkedList::ReverseIterator IntrusiveLinkedList::getReverseBegin()
{
	return ReverseIterator(m_oRoot.m_pPrevious);
}

NANOJIT_FORCEINLINE IntrusiveLinkedList::ReverseIterator IntrusiveLinkedList::getReverseEnd()
{
	return ReverseIterator(&m_oRoot);
}

template<typename T>
NANOJIT_FORCEINLINE IntrusiveLinkedList::ReverseIterator IntrusiveLinkedList::getReverseIteratorAt(T* a_pValue)
{
	IntrusiveListNode* pNode = reinterpret_cast<IntrusiveListNode*>(reinterpret_cast<uint8_t*>(a_pValue) + T::NODE_OFFSET);
	return ReverseIterator(pNode);
}

NANOJIT_FORCEINLINE void IntrusiveLinkedList::insertHead(IntrusiveListNode* a_pNode)
{
	insert(a_pNode, &m_oRoot, m_oRoot.m_pNext);
}

NANOJIT_FORCEINLINE void IntrusiveLinkedList::insertTail(IntrusiveListNode* a_pNode)
{
	insert(a_pNode, m_oRoot.m_pPrevious, &m_oRoot);
}

NANOJIT_FORCEINLINE void IntrusiveLinkedList::insertBefore(Iterator& a_oIt, IntrusiveListNode* a_pNode)
{
	insert(a_pNode, a_oIt.m_pNode->m_pPrevious, a_oIt.m_pNode);
}

NANOJIT_FORCEINLINE void IntrusiveLinkedList::insertBefore(ReverseIterator& a_oIt, IntrusiveListNode* a_pNode)
{
	insert(a_pNode, a_oIt.m_pNode->m_pPrevious, a_oIt.m_pNode);
}

NANOJIT_FORCEINLINE void IntrusiveLinkedList::insertAfter(Iterator& a_oIt, IntrusiveListNode* a_pNode)
{
	insert(a_pNode, a_oIt.m_pNode, a_oIt.m_pNode->m_pNext);
}

NANOJIT_FORCEINLINE void IntrusiveLinkedList::insertAfter(ReverseIterator& a_oIt, IntrusiveListNode* a_pNode)
{
	insert(a_pNode, a_oIt.m_pNode, a_oIt.m_pNode->m_pNext);
}

NANOJIT_FORCEINLINE void IntrusiveLinkedList::remove(const Iterator& a_oIt)
{
	remove(a_oIt.m_pNode);
}


NANOJIT_FORCEINLINE IntrusiveLinkedListIterator::IntrusiveLinkedListIterator(IntrusiveListNode* a_pNode)
: m_pNode(a_pNode)
{
}


NANOJIT_FORCEINLINE IntrusiveLinkedListIterator::~IntrusiveLinkedListIterator()
{
}


NANOJIT_FORCEINLINE IntrusiveLinkedListIterator::IntrusiveLinkedListIterator(const IntrusiveLinkedListIterator& a_oOther)
: m_pNode(a_oOther.m_pNode)
{
}

NANOJIT_FORCEINLINE void IntrusiveLinkedListIterator::moveNext()
{
	m_pNode = m_pNode->m_pNext;
}

template<typename T>
NANOJIT_FORCEINLINE T* IntrusiveLinkedListIterator::getElement()
{
	return reinterpret_cast<T*>(reinterpret_cast<uint8_t*>(m_pNode) - T::NODE_OFFSET);
}

NANOJIT_FORCEINLINE bool IntrusiveLinkedListIterator::operator==(const IntrusiveLinkedListIterator& a_oOther) const
{
	return m_pNode == a_oOther.m_pNode;
}

NANOJIT_FORCEINLINE bool IntrusiveLinkedListIterator::operator!=(const IntrusiveLinkedListIterator& a_oOther) const
{
	return m_pNode != a_oOther.m_pNode;
}

NANOJIT_FORCEINLINE void IntrusiveLinkedListReverseIterator::moveNext()
{
	m_pNode = m_pNode->m_pPrevious;
}

NANOJIT_FORCEINLINE IntrusiveLinkedListReverseIterator::IntrusiveLinkedListReverseIterator(IntrusiveListNode* a_pNode)
: m_pNode(a_pNode)
{
}

NANOJIT_FORCEINLINE IntrusiveLinkedListReverseIterator::IntrusiveLinkedListReverseIterator(const IntrusiveLinkedListReverseIterator& a_oOther)
: m_pNode(a_oOther.m_pNode)
{
}

NANOJIT_FORCEINLINE IntrusiveLinkedListReverseIterator::~IntrusiveLinkedListReverseIterator()
{
}

template<typename T>
NANOJIT_FORCEINLINE T* IntrusiveLinkedListReverseIterator::getElement()
{
	return reinterpret_cast<T*>(reinterpret_cast<uint8_t*>(m_pNode) - T::NODE_OFFSET);
}

NANOJIT_FORCEINLINE bool IntrusiveLinkedListReverseIterator::operator==(const IntrusiveLinkedListReverseIterator& a_oOther) const
{
	return m_pNode == a_oOther.m_pNode;
}

NANOJIT_FORCEINLINE bool IntrusiveLinkedListReverseIterator::operator!=(const IntrusiveLinkedListReverseIterator& a_oOther) const
{
	return m_pNode != a_oOther.m_pNode;
}

}

#endif // __nanojit_IntrusiveLinkedList__
